/* This file deals with (you guessed it) dynamic loading.*/

/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */

#include "config.h"
#include <dynamic_loading.h>
#include <log.h>
#include "jni.h"
#include "native-threads.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#if HAVE_DLFCN_H
#include <dlfcn.h>
#endif
#if HAVE_DL_H
#include <dl.h>
#endif

#define MYLOG "ArchDLL"

#if (!defined HAVE_DLL) || defined PROFILING
extern sym_mapping java_lang_mappings[];
extern int num_java_lang_mappings;

extern sym_mapping java_lang_reflect_mappings[];
extern int num_java_lang_reflect_mappings;

extern sym_mapping java_io_mappings[];
extern int num_java_io_mappings;

extern sym_mapping java_net_mappings[];
extern int num_java_net_mappings;

extern sym_mapping java_util_mappings[];
extern int num_java_util_mappings;

extern sym_mapping java_util_zip_mappings[];
extern int num_java_util_zip_mappings;

extern sym_mapping sysres_mappings[];
extern int num_sysres_mappings;

#ifdef HAVE_MOTIF
extern sym_mapping sun_awt_motif_mappings[];
extern int num_sun_awt_motif_mappings;
#endif
#endif

typedef struct library_cache_entry {
  struct library_cache_entry *prev;
  struct library_cache_entry *next;

  char *library_name;
  DLL_LibHandle library_handle;
} library_cache_entry;

static library_cache_entry* cache = NULL;

/* adds a library to the cache. */
static void 
add_to_library_cache(char *lib_name,
		     DLL_LibHandle handle)
{
  library_cache_entry *new_entry = (library_cache_entry*)calloc(1, sizeof(library_cache_entry));

  new_entry->library_name = strdup(lib_name);
  new_entry->library_handle = handle;

  if (cache)
    cache->prev = new_entry;
  new_entry->next = cache;
  cache = new_entry;
}

static void
remove_from_library_cache(DLL_LibHandle handle)
{
  library_cache_entry *cur;

  for (cur = cache;
       cur != NULL;
       cur = cur->next)
    {
      if (cur->library_handle == handle)
	{
	  if (cur->prev)
	    cur->prev->next = cur->next;
	  if (cur->next)
	    cur->next->prev = cur->prev;

	  if (cache == cur)
	    cache = cur->next;

	  free (cur->library_name);
	  free (cur);
	}
    }
}

static DLL_LibHandle
library_in_cache(char *lib_name)
{
  library_cache_entry *cur;

  for (cur = cache;
       cur != NULL;
       cur = cur->next)
    {
      if (!strcmp(lib_name, cur->library_name))
	return cur->library_handle;
    }

  return NULL;
}

/**
 * loads in a shared library and returns a handle to it.
 * this handle is used in all future references to the
 * library.
 *
 * NULL should be returned in the event of failure.
 */
DLL_LibHandle
DLL_load(char *lib_path)
{
  DLL_LibHandle handle = NULL;

  JAVARLOG1(MYLOG, 1, "Loading shared library %s\n", lib_path);

  /* check if it's in our cache. */
  handle = library_in_cache(lib_path);

  /* if so, just return the handle to it. */
  if (handle != NULL)
    return handle;

#if defined HAVE_DLL && !defined PROFILING

#if HAVE_DLFCN_H
  /* if it's not, we attempt to load it. */
  handle = dlopen(lib_path, 1);
#endif

  /* we only add it to the cache if the load succeeded */
  if (handle != NULL)
    add_to_library_cache(lib_path, handle);
#if HAVE_DLFCN_H
  else
    JAVARLOG1(MYLOG, 3, "%s\n", dlerror());
#endif /* HAVE_DLFCN_H */

#else /* !HAVE_DLL || PROFILING */
  handle = (DLL_LibHandle)calloc(1, sizeof(struct DLL_LibHandle));

  if (!strcmp(lib_path, "lang"))
    {
      handle->mappings = java_lang_mappings;
      handle->num_mappings = num_java_lang_mappings;
    }
  else if (!strcmp(lib_path, "io"))
    {
      handle->mappings = java_io_mappings;
      handle->num_mappings = num_java_io_mappings;
    }
  else if (!strcmp(lib_path, "net"))
    {
      handle->mappings = java_net_mappings;
      handle->num_mappings = num_java_net_mappings;
    }
  else if (!strcmp(lib_path, "util"))
    {
      handle->mappings = java_util_mappings;
      handle->num_mappings = num_java_util_mappings;
    }
  else if (!strstr(lib_path, "zip"))
    {
      handle->mappings = java_util_zip_mappings;
      handle->num_mappings = num_java_util_zip_mappings;
    }
  else if (!strstr(lib_path, "sysresource"))
    {
      handle->mappings = sysres_mappings;
      handle->num_mappings = num_sysres_mappings;
    }
#ifdef HAVE_MOTIF
  else if (!strstr(lib_path, "awt"))
    {
      handle->mappings = sun_awt_motif_mappings;
      handle->num_mappings = num_sun_awt_motif_mappings;
    }
#endif

  /* we only add it to the cache if the load succeeded */
  add_to_library_cache(lib_path, handle);

#endif
  return handle;
}

#if !defined( PROFILING )
static DLL_LibHandle*
DLL_findAlongPath(char *lib_name, char *path)
{
  DLL_LibHandle *handle = NULL;
  char *lib_path_entry;
  char *colon;
  char *full_lib_name;

  lib_path_entry = path;
  colon = strchr(lib_path_entry, ':');

  while (lib_path_entry[0] != '\0')
    {
      if (colon)
	*colon = '\0'; /* we'll set this back shortly. */

      full_lib_name = (char*)malloc(strlen(lib_path_entry)
				    + strlen("libjaphar_")
				    + strlen(lib_name)
				    + strlen(".so") + 2);

      sprintf (full_lib_name, "%s/libjaphar_%s.so", lib_path_entry, lib_name);

      handle = DLL_load(full_lib_name);

      free(full_lib_name);

      if (handle != NULL)
	{
	  if (colon)
	    *colon = ':';
	  break;
	}
      else
	{
	  if (colon)
	    {
	      *colon = ':';
	      lib_path_entry = colon + 1;
	      colon = strchr(lib_path_entry, ':');
	    }
	  else
	    break;
	}
    }

  return handle;
}
#endif

DLL_LibHandle
DLL_find(char *lib_name)
{
  DLL_LibHandle handle = NULL;
#if defined HAVE_DLL && !defined PROFILING
  static char *LD_LIBRARY_PATH;

  if (!LD_LIBRARY_PATH)
    LD_LIBRARY_PATH = getenv("LD_LIBRARY_PATH");

  if (LD_LIBRARY_PATH)
    handle = DLL_findAlongPath(lib_name, LD_LIBRARY_PATH);

  if (handle == NULL
      && (!LD_LIBRARY_PATH || strcmp(LD_LIBRARY_PATH, LIBDIR)))
    {
      handle = DLL_findAlongPath(lib_name, LIBDIR);
    }

#else /* !HAVE_DLL || PROFILING */
  handle = DLL_load(lib_name);
#endif

  return handle;
}

void
DLL_unload(DLL_LibHandle handle)
{
#if defined HAVE_DLL && !defined PROFILING
  int close_val;
  
  close_val = dlclose(handle);
  
#ifdef DEBUGGING_ON
  if (close_val != 0) dlerror();
#endif

#else /* !HAVE_DLL || PROFILING */
  free(handle);
#endif

  remove_from_library_cache(handle);
}

static DLL_FuncHandle
handle_for_function_in_lib(char *func_name,
			   DLL_LibHandle handle)
{
#if defined HAVE_DLL && !defined PROFILING
  char buf[1024];

  snprintf(buf, sizeof(buf), "%s%s",
#ifdef NEED_UNDERSCORE
	   "_",
#else
	   "",
#endif
	   func_name);

#if HAVE_DLFCN_H
  return dlsym(handle, buf);
#endif

#else /* !HAVE_DLL || PROFILING */
  int i;

  for (i = 0; i < handle->num_mappings; i ++)
    if (!strcmp(handle->mappings[i].name, func_name))
      return handle->mappings[i].func_ptr;

  return NULL;
#endif
}

static void
get_symbols_from_executable()
{
#if defined HAVE_DLL && !defined PROFILING
  static int only_one_time = 0;

  if (!only_one_time) 
    {
      DLL_LibHandle handle = NULL;

      only_one_time = 1;

#if HAVE_DLFCN_H
      handle = dlopen(NULL, 1);
#endif

      add_to_library_cache ("<exec>", handle);
    }
#endif
}

DLL_FuncHandle
DLL_findFunction(char *func_name)
{
  library_cache_entry *cur;
  DLL_FuncHandle handle = NULL;

  get_symbols_from_executable();

  for (cur = cache;
       cur != NULL;
       cur = cur->next)
    {
      JAVARLOG2(MYLOG, 2, "Is %s() in library %s ?\n", func_name, cur->library_name);
      handle = handle_for_function_in_lib(func_name,
					  cur->library_handle);

      if (handle != NULL)
	{
	  JAVARLOG0(MYLOG, 2, " - yep\n");
	  return handle;
	}

      JAVARLOG0(MYLOG, 2, " - nope\n");
    }

  return NULL;
}

void
DLL_loadSystemLibraries()
{
#if defined HAVE_DLL && !defined PROFILING
  DLL_LibHandle native_handle = DLL_find("lang");
  if (!native_handle)
    {
      JNIEnv *env = THREAD_getEnv();
	  
      jclass cls = (*env)->FindClass(env, "java/lang/LinkageError");
      if (!cls || (*env)->ExceptionOccurred(env))
	(*env)->FatalError(env, "unable to find libjaphar_lang.so and java/lang/LinkageError");
      (*env)->ThrowNew(env, cls, "libjaphar_lang");
      return;
    }


  native_handle = DLL_find("io");
  if (!native_handle)
    {
      JNIEnv *env = THREAD_getEnv();
	  
      jclass cls = (*env)->FindClass(env, "java/lang/LinkageError");
      if (!cls || (*env)->ExceptionOccurred(env))
	(*env)->FatalError(env, "unable to find libjaphar_io.so.");
      (*env)->ThrowNew(env, cls, "libjaphar_io");
      return;
    }

  native_handle = DLL_find("util");
  if (!native_handle)
    {
      JNIEnv *env = THREAD_getEnv();
	  
      jclass cls = (*env)->FindClass(env, "java/lang/LinkageError");
      if (!cls || (*env)->ExceptionOccurred(env))
	(*env)->FatalError(env, "unable to find libjaphar_util.so");
      (*env)->ThrowNew(env, cls, "libjaphar_util");
      return;
    }

  native_handle = DLL_find("reflect");
  if (!native_handle)
    {
      JNIEnv *env = THREAD_getEnv();
	  
      jclass cls = (*env)->FindClass(env, "java/lang/LinkageError");
      if (!cls || (*env)->ExceptionOccurred(env))
	(*env)->FatalError(env, "unable to find libjaphar_reflect.so");
      (*env)->ThrowNew(env, cls, "libjaphar_reflect");
      return;
    }
#else
  /* these take care of adding all the mappings for the various
     native methods when we don't have shared libraries. */
  DLL_load("lang");
  DLL_load("reflect");
  DLL_load("io");
  DLL_load("net");
  DLL_load("util");
  DLL_load("zip");
#endif
}
