/* -*- Mode: C; c-file-style: "gnu" -*-
   monitors.c -- pthread code for monitors.
   Created: Chris Toshok <toshok@hungry.com>, 10-Aug-1997
 */

/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include "native-threads.h"
#include "ClazzFile.h"
#include "objects.h"

#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include "qop.h"
#if defined(USE_PTHREADS)
/* Hack for Linux RedHat 4.1 (libc5) */
#  define _MIT_POSIX_THREADS 1
#  include <pthread.h>
#  define THREAD_EQUAL(a,b) pthread_equal(a,b)
#elif defined(USE_CTHREADS)
#  include <cthreads.h>
#  define THREAD_EQUAL(a,b) (a == b)
#elif defined(USE_NSPR_THREADS)
#  include "prthread.h"
#  include "prmon.h"
#  define THREAD_EQUAL(a,b) (a == b)
#endif

struct MonitorStruct {
#if defined(USE_NSPR_THREADS)
  PRMonitor *_mon;
#else
  HMonitor prev;
  HMonitor next;

  char *name;
  jobject obj;

  int depth;

#  if defined(USE_PTHREADS)
  pthread_t owner;
  pthread_cond_t cv;
  pthread_mutex_t mutex;
#  elif defined(USE_CTHREADS)
  cthread_t owner;
  condition_t cv;
  mutex_t mutex;
#  endif
#endif
};

static HMonitor monitors;

HMonitor
MONITOR_create()
{
  HMonitor mon = (HMonitor)calloc(1, sizeof(struct MonitorStruct));

#if defined(USE_NSPR_THREADS)

  mon->_mon = PR_NewMonitor();

#else

#  if defined(USE_PTHREADS)
  pthread_cond_init(&mon->cv, NULL);
  pthread_mutex_init(&mon->mutex, NULL);
#  elif defined(USE_CTHREADS)
  mon->cv = condition_alloc();
  mon->mutex = mutex_alloc();
  condition_init(mon->cv);
  mutex_init(mon->mutex);
#  endif
  mon->owner = 0;
  mon->depth = 0;

  ENQUEUE(mon, monitors);

  monitors = mon;

#endif

  return mon;
}

HMonitor
MONITOR_createNamed(const char *name)
{
  HMonitor mon = MONITOR_create();

#if !defined(USE_NSPR_THREADS)
  mon->name = strdup(name);
#endif

  return mon;
}

void
MONITOR_destroy(HMonitor monitor)
{
#if defined(USE_NSPR_THREADS)

  PR_DestroyMonitor(monitor->_mon);

#else
  if (monitor->name) free(monitor->name);

#  if defined(USE_PTHREADS)
  pthread_mutex_destroy(&monitor->mutex);
  pthread_cond_destroy(&monitor->cv);
#  elif defined(USE_CTHREADS)
  condition_free(monitor->cv);
  mutex_free(monitor->mutex);
#  endif

  UNQUEUE(monitor, monitors);
#endif

  free(monitor);
}

void
MONITOR_enter(HMonitor monitor)
{
#if defined(USE_NSPR_THREADS)

  PR_EnterMonitor(monitor->_mon);
  
#else
#  if defined(USE_PTHREADS)
  pthread_t current = pthread_self();
  int val = pthread_mutex_trylock(&monitor->mutex);
#  elif defined(USE_CTHREADS)
  cthread_t current = cthread_self();
  int val = mutex_try_lock(monitor->mutex);
#  endif

  if ((THREAD_EQUAL(monitor->owner,current)
       && val == -1)
#  if defined(USE_PTHREADS)
      || val == 0
      || (THREAD_EQUAL(monitor->owner, current) && val == EBUSY))
#  elif defined(USE_CTHREADS)
    || val == 1)
#  endif
{
  monitor->owner = current;
  monitor->depth ++;
}
else /* it is locked by someone else. */
{
  if (THREAD_EQUAL(monitor->owner, current))
    fprintf (stderr, "blocked waiting for me self, owner is %p (val=%d)\n",
	     (void *)monitor->owner, val);
  MONITOR_wait(monitor);
}
#endif
}

void
MONITOR_exit(HMonitor monitor)
{
#if defined(USE_NSPR_THREADS)

PR_ExitMonitor(monitor->_mon);
	
#else

#  if defined(USE_PTHREADS)
  pthread_t current = pthread_self();
#  elif defined(USE_CTHREADS)
  cthread_t current = cthread_self();
#  endif

  if (monitor->owner != current) assert(0);

if (--monitor->depth == 0)
{
  monitor->owner = 0;
	  
#  if defined(USE_PTHREADS)
  pthread_mutex_unlock(&monitor->mutex);
  pthread_cond_broadcast(&monitor->cv);
#  elif defined(USE_CTHREADS)
  mutex_unlock(monitor->mutex);
  condition_broadcast(monitor->cv);
#  endif
}
#endif
}

void
MONITOR_wait(HMonitor monitor)
{
#if defined(USE_NSPR_THREADS)

  PR_Wait(monitor->_mon, PR_INTERVAL_NO_TIMEOUT);

#else
#  if defined(USE_PTHREADS)
  pthread_t current = pthread_self();

  pthread_cond_wait(&monitor->cv, &monitor->mutex);
#  elif defined(USE_CTHREADS)
  cthread_t current = cthread_self();

  condition_wait(monitor->cv, monitor->mutex);
#  endif

  /* we now own the monitor */
  monitor->owner = current;
  monitor->depth = 1;
#endif
}

void
MONITOR_timedWait(HMonitor monitor, jlong time_to_wait)
{
#if defined(USE_NSPR_THREADS)

  PR_Wait(monitor->_mon, (PRUint32)time_to_wait);

#else

  /* XXXX fix me... */
  MONITOR_wait(monitor);

#endif
}

void
MONITOR_notifyOne(HMonitor monitor)
{
#if defined(USE_NSPR_THREADS)

  PR_Notify(monitor->_mon);

#else
  /* this doesn't sound exactly right... */
#  if defined(USE_PTHREADS)
  pthread_cond_signal(&monitor->cv);
#  elif defined(USE_CTHREADS)
  cond_signal(monitor->cv);
#  endif
#endif
}

void
MONITOR_notifyAll(HMonitor monitor)
{
#if defined(USE_NSPR_THREADS)
  PR_NotifyAll(monitor->_mon);
#else
#  if defined(USE_PTHREADS)
  pthread_cond_broadcast(&monitor->cv);
#  elif defined(USE_CTHREADS)
  cond_broadcast(monitor->cv);
#  endif
#endif
}

HMonitor
MONITOR_getForObject(jobject obj)
{
  HMonitor mon = INSTANCE_DATA(obj)->monitor;

  if (!mon)
    {
      mon = INSTANCE_DATA(obj)->monitor = MONITOR_create();
#if !defined(USE_NSPR_THREADS)
      mon->obj = obj;
#endif
    }

  return mon;
}

HMonitor
MONITOR_getNamed(const char *name)
{
#if !defined(USE_NSPR_THREADS)
  HMonitor mon;

  for (mon = monitors;
       mon != NULL;
       mon = mon->next)
    {
      if (mon->name && !strcmp(mon->name, name))
	return mon;
    }
#endif

  return NULL;
}
