/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */

#include "config.h"
#include "jniint.h"
#include "interp.h"
#include "native-threads.h"

#include <stdio.h>
#include <stdlib.h>

#include <sys/time.h>
#include <signal.h>
#if defined(USE_PTHREADS)
/* Hack for Linux RedHat 4.1 (libc5) */
# define _MIT_POSIX_THREADS 1
# include <pthread.h>
# ifdef HAVE_PTHREAD_YIELD
#  define YIELD pthread_yield
# elif defined(HAVE_SCHED_YIELD)
#  define YIELD sched_yield
# elif defined(HAVE_THR_YIELD)
#  define YIELD thr_yield
# elif defined(HAVE_YIELD)
#  define YIELD yield
# else
#  error yield() missing! - port me!
#endif
#elif defined(USE_CTHREADS)
#  include <cthreads.h>
#  define YIELD cthread_yield
#elif defined(USE_NSPR_THREADS)
# include "prthread.h"
# define YIELD() PR_Sleep(PR_INTERVAL_NO_WAIT)
#else
#  error Missing threads. port me!
#endif

#if defined(USE_NSPR_THREADS)
static PRUintn JNIENV_KEY = -1;
static PRUintn JNIVM_KEY = -1;
static PRUintn JTHREADINFO_KEY = -1;
#elif defined(USE_PTHREADS)
static pthread_key_t JNIENV_KEY = (pthread_key_t)-1;
static pthread_key_t JNIVM_KEY = (pthread_key_t)-1;
static pthread_key_t JTHREADINFO_KEY = (pthread_key_t)-1;
#elif defined(USE_CTHREADS)
static cthread_key_t JNIENV_KEY = -1;
static cthread_key_t JNIVM_KEY = -1;
static cthread_key_t JTHREADINFO_KEY = -1;
#endif

static void thread_onetime_init()
{
  static int one_time = 0;

  if (one_time) return;

  one_time = 1;

#if defined(USE_CTHREADS)
  cthread_init();
#endif
}

struct start_arg_cons {
  void *(*start_function)(void*);
  void *arg;
  JavaVM *vm;
};

static void *
thread_start(void *start_arg)
{
  struct start_arg_cons *foo = (struct start_arg_cons*)start_arg;
  void *(*function)(void*) = foo->start_function;
  void *arg = foo->arg;
  JavaVM *vm = foo->vm;
  JNIEnv *env;
  void* result;

  thread_onetime_init();

  if (vm)
    (*vm)->AttachCurrentThread(vm, &env, NULL);

  free(foo);

  result = (*function)(arg);

  if (vm)
    (*vm)->DetachCurrentThread(vm);

  /* free up the thread specific stack. XXX */

  return result;
}

HThread
THREAD_start(void *(*start_function)(void*),
	     void *start_arg,
	     int priority,
	     int initialStackSize)
{
#if defined(USE_NSPR_THREADS)
  PRThread* new_thread;
#elif defined(USE_PTHREADS)
  pthread_t *new_thread = (pthread_t*)malloc(sizeof(pthread_t));
#elif defined(USE_CTHREADS)
  cthread_t new_thread;
#endif
  struct start_arg_cons *cons =
    (struct start_arg_cons*)malloc(sizeof(struct start_arg_cons));

  thread_onetime_init();
  
  cons->start_function = start_function;
  cons->arg = start_arg;
  cons->vm = THREAD_getVM();

#if defined(USE_NSPR_THREADS)
  new_thread = PR_CreateThread(PR_USER_THREAD,
			       thread_start,
			       cons,
			       PR_PRIORITY_NORMAL,
			       PR_LOCAL_THREAD,
			       PR_JOINABLE_THREAD,
			       0);
#elif defined(USE_PTHREADS)
  if (pthread_create(new_thread, NULL, thread_start, cons))
    {
      fprintf (stderr, "pthread_create failed.\n");
      exit(1);
    }
#elif defined(USE_CTHREADS)
  new_thread = cthread_fork(thread_start, cons);
  if (new_thread == NULL)
    {
      fprintf (stderr, "cthread_fork failed.\n");
      exit(1);
    }
#endif

  THREAD_yield();

  return (HThread)new_thread;
}

void
THREAD_exit()
{
  thread_onetime_init();

#if defined(USE_NSPR_THREADS)
  /* nothing here. */
#elif defined(USE_PTHREADS)
  pthread_exit(NULL);
#elif defined(USE_CTHREADS)
  cthread_exit(NULL);
#endif
}

int
THREAD_delete(HThread thread)
{
  thread_onetime_init();

#if defined(USE_NSPR_THREADS)
  /* nothing needed here. */
  return 0;
#elif defined(USE_PTHREADS)
  {
    pthread_t *pthread = (pthread_t*)thread;
    return pthread_detach(*pthread);
  }
#elif defined(USE_CTHREADS)
  {
    cthread_t _thread = (cthread_t)thread;
    
    cthread_detach(_thread);

    return 0;
  }
#endif
}

void
THREAD_yield()
{
  thread_onetime_init();

  YIELD();
}

HThread
THREAD_getCurrent()
{
  thread_onetime_init();

#if defined(USE_NSPR_THREADS)
  return (HThread)PR_GetCurrentThread();
#elif defined(USE_PTHREADS)
  return (HThread)pthread_self();
#elif defined(USE_CTHREADS)
  return (HThread)cthread_self();
#endif
}

static void
thread_setKeyValue(int *key, const void* value)
{
  thread_onetime_init();

#if defined(USE_NSPR_THREADS)
  if (*key == -1)
    if (PR_NewThreadPrivateIndex(key, NULL) != PR_SUCCESS)
      abort();

  if (PR_SetThreadPrivate(*key, value) != PR_SUCCESS)
    abort();
#elif defined(USE_PTHREADS)
  if (*key == (pthread_key_t)-1)
    if (pthread_key_create(key, NULL) != 0)
      abort();

  if (pthread_setspecific(*key, value) != 0)
    abort();
#elif defined(USE_CTHREADS)
  if (*key == (cthread_key_t)-1)
    if (cthread_keycreate(key) != 0)
      abort();

  if (cthread_setspecific(*key, (any_t)value) != 0)
    abort();
#endif
}

static void*
thread_getKeyValue(int *key)
{
  void* value = NULL;

  thread_onetime_init();

#if defined(USE_NSPR_THREADS)
  if (*key == -1)
    if (PR_NewThreadPrivateIndex(key, NULL) != PR_SUCCESS)
      abort();

  value = (void*)PR_GetThreadPrivate(*key);
#elif defined(USE_PTHREADS)
  if (*key == (pthread_key_t)-1)
    if (pthread_key_create(key, NULL) != 0)
      abort();

#ifdef OLD_FREEBSD_PTHREADS
  pthread_getspecific(*key, (void*)&value);
#else
  value = pthread_getspecific(*key);
#endif
#elif defined(USE_CTHREADS)
  if (*key == (cthread_key_t)-1)
    if (cthread_keycreate(key) != 0)
      abort();

  cthread_getspecific(*key, (any_t*)&value);
#endif /* USE_CTHREADS */

  return value;
}

void
THREAD_setEnv(JNIEnv *env)
{
  thread_setKeyValue(&JNIENV_KEY, env);
}

void
THREAD_setVM(JavaVM *vm)
{
  thread_setKeyValue(&JNIVM_KEY, vm);
}

JavaVM *
THREAD_getVM()
{
  return (JavaVM*)thread_getKeyValue(&JNIVM_KEY);
}

JNIEnv *
THREAD_getEnv()
{
  return (JNIEnv*)thread_getKeyValue(&JNIENV_KEY);
}

void
THREAD_setJavaInfo(void *info)
{
  JNIEnv *env;
  HungryJNIEnv *henv;

  thread_setKeyValue(&JTHREADINFO_KEY, info);

  env = THREAD_getEnv();
  henv = (HungryJNIEnv*)env;
  henv->_java_info = info;
}

void*
THREAD_getJavaInfo()
{
  return (JThreadInfo*)thread_getKeyValue(&JTHREADINFO_KEY);
}
