/* -*- Mode: C; c-file-style: "gnu" -*-
   jniarray.c -- Java Native Interface array operations.
   Created: Chris Toshok <toshok@hungry.com>, 26-Jul-1997
*/
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include <stdlib.h>
#include <string.h> /* For strlen() on Alpha/Linux */

#include "jni.h"
#include "jniint.h"
#include "arrays.h"
#include "array-class.h"
#include "interp.h"
#include "qop.h"

jsize
JNIFUNC(GetArrayLength)(JNIEnv *env,
			jarray array)
{
  JavaStackItem length_item;

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  return length_item.value.i;
}

jarray
JNIFUNC(NewObjectArray)(JNIEnv *env,
			jsize length,
			jclass elementClass,
			jobject initialElement)
{
  ClazzFile *cf = (ClazzFile*)elementClass;
  char *element_classname = getClassName(env, cf);
  obj_thunk *array_thunk;

  /* XXX use initialElement? */

  if (element_classname[0] != '[')
    {
      char *array_classname = (char*)malloc(strlen(element_classname) + 4);
      sprintf (array_classname, "[L%s;", element_classname);
      array_thunk = new_array(env, length, createFakeArrayClass(env, array_classname));
      free(array_classname);
    }
  else
    {
      array_thunk = new_array(env, length, createFakeArrayClass(env, element_classname));
    }

  return array_thunk;
}

jobject
JNIFUNC(GetObjectArrayElement)(JNIEnv *env,
			       jobjectArray array,
			       jsize index)
{
  JavaStackItem body_item;

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  return ((jobject*)body_item.value.l)[ index ];
}

void
JNIFUNC(SetObjectArrayElement)(JNIEnv *env,
			       jobjectArray array,
			       jsize index,
			       jobject value)
{
  JavaStackItem body_item;

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  ((jobject*)body_item.value.l)[ index ] = value;
}

jbooleanArray
JNIFUNC(NewBooleanArray)(JNIEnv *env,
			 jsize length)
{
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  ClazzFile *boolean_array_type = createFakeArrayClass(env, "[Z");
  jobject array = new_array(env,length, boolean_array_type);
  
  ENQUEUE((GC_obj*)array, henv->_local_refs);

  return array;
}

jboolean*
JNIFUNC(GetBooleanArrayElements)(JNIEnv *env,
				 jbooleanArray array,
				 jboolean* isCopy)
{
  JavaStackItem body_item, length_item;
  jboolean *the_copy;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  the_copy = (jboolean*)malloc(sizeof(jboolean) * length_item.value.i);

  if (the_copy == NULL)
    {
      if (isCopy) *isCopy = JNI_FALSE;
      return (jboolean*)body_item.value.l;
    }
  else
    {
      memcpy(the_copy, body_item.value.l, sizeof(jboolean) * length_item.value.i);
      if (isCopy) *isCopy = JNI_TRUE;
      return the_copy;
    }
}

void
JNIFUNC(ReleaseBooleanArrayElements)(JNIEnv *env,
				     jbooleanArray array,
				     jboolean* elems,
				     jint mode)
{
  JavaStackItem body_item, length_item;
  jboolean can_free = JNI_FALSE;

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (body_item.value.l != elems)
    can_free = JNI_TRUE;

  switch (mode)
    {
    case 0: /* commit the changes and free elems. */
      memcpy(body_item.value.l, elems, sizeof(jboolean) * length_item.value.i);
      if (can_free) free(elems);
      break;
    case JNI_COMMIT: /* commit the changes but don't free elems. */
      memcpy(body_item.value.l, elems, sizeof(jboolean) * length_item.value.i);
      break;
    case JNI_ABORT:  /* don't commit the changes but free elems. */
      if (can_free) free(elems);
      break;
    }
}

void
JNIFUNC(GetBooleanArrayRegion)(JNIEnv *env,
			       jbooleanArray array,
			       jsize start,
			       jsize len,
			       jboolean* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetBooleanArrayRegion");
      return;
    }

  memcpy(buf, ((jboolean*)body_item.value.l) + start, len * sizeof(jboolean));
}

void
JNIFUNC(SetBooleanArrayRegion)(JNIEnv *env,
			       jbooleanArray array,
			       jsize start,
			       jsize len,
			       jboolean* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetBooleanArrayRegion");
      return;
    }

  memcpy(((jboolean*)body_item.value.l) + start, buf, len * sizeof(jboolean));
}

jbyteArray
JNIFUNC(NewByteArray)(JNIEnv *env,
		      jsize length)
{
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  ClazzFile *byte_array_type = createFakeArrayClass(env, "[B");
  jobject array = new_array(env,length, byte_array_type);
  
  ENQUEUE((GC_obj*)array, henv->_local_refs);

  return array;
}

jbyte*
JNIFUNC(GetByteArrayElements)(JNIEnv *env,
			      jbyteArray array,
			      jboolean* isCopy)
{
  JavaStackItem body_item, length_item;
  jbyte *the_copy;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  the_copy = (jbyte*)malloc(sizeof(jbyte) * length_item.value.i);

  if (the_copy == NULL)
    {
      if (isCopy) *isCopy = JNI_FALSE;
      return (jbyte*)body_item.value.l;
    }
  else
    {
      memcpy(the_copy, body_item.value.l, sizeof(jbyte) * length_item.value.i);
      if (isCopy) *isCopy = JNI_TRUE;
      return the_copy;
    }
}

void
JNIFUNC(ReleaseByteArrayElements)(JNIEnv *env,
				  jbyteArray array,
				  jbyte* elems,
				  jint mode)
{
  JavaStackItem body_item, length_item;
  jboolean can_free = JNI_FALSE;

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (body_item.value.l != elems)
    can_free = JNI_TRUE;

  switch (mode)
    {
    case 0: /* commit the changes and free elems. */
      memcpy(body_item.value.l, elems, sizeof(jbyte) * length_item.value.i);
      if (can_free) free(elems);
      break;
    case JNI_COMMIT: /* commit the changes but don't free elems. */
      memcpy(body_item.value.l, elems, sizeof(jbyte) * length_item.value.i);
      break;
    case JNI_ABORT:  /* don't commit the changes but free elems. */
      if (can_free) free(elems);
      break;
    }
}

void
JNIFUNC(GetByteArrayRegion)(JNIEnv *env,
			    jbyteArray array,
			    jsize start,
			    jsize len,
			    jbyte* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetByteArrayRegion");
      return;
    }

  memcpy(buf, ((jbyte*)body_item.value.l) + start, len * sizeof(jbyte));
}

void
JNIFUNC(SetByteArrayRegion)(JNIEnv *env,
			    jbyteArray array,
			    jsize start,
			    jsize len,
			    jbyte* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetByteArrayRegion");
      return;
    }

  memcpy(((jbyte*)body_item.value.l) + start, buf, len * sizeof(jbyte));
}

jcharArray
JNIFUNC(NewCharArray)(JNIEnv *env,
		      jsize length)
{
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  ClazzFile *char_array_type = createFakeArrayClass(env, "[C");
  jobject array = new_array(env, length, char_array_type);

  ENQUEUE((GC_obj*)array, henv->_local_refs);

  return array;
}

jchar*
JNIFUNC(GetCharArrayElements)(JNIEnv *env,
			      jcharArray array,
			      jboolean* isCopy)
{
  JavaStackItem body_item, length_item;
  jchar *the_copy;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  the_copy = (jchar*)malloc(sizeof(jchar) * length_item.value.i);

  if (the_copy == NULL)
    {
      if (isCopy) *isCopy = JNI_FALSE;
      return (jchar*)body_item.value.l;
    }
  else
    {
      memcpy(the_copy, body_item.value.l, sizeof(jchar) * length_item.value.i);
      if (isCopy) *isCopy = JNI_TRUE;
      return the_copy;
    }
}

void
JNIFUNC(ReleaseCharArrayElements)(JNIEnv *env,
				  jcharArray array,
				  jchar* elems,
				  jint mode)
{
  JavaStackItem body_item, length_item;
  jboolean can_free = JNI_FALSE;

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (body_item.value.l != elems)
    can_free = JNI_TRUE;

  switch (mode)
    {
    case 0: /* commit the changes and free elems. */
      memcpy(body_item.value.l, elems, sizeof(jchar) * length_item.value.i);
      if (can_free) free(elems);
      break;
    case JNI_COMMIT: /* commit the changes but don't free elems. */
      memcpy(body_item.value.l, elems, sizeof(jchar) * length_item.value.i);
      break;
    case JNI_ABORT:  /* don't commit the changes but free elems. */
      if (can_free) free(elems);
      break;
    }
}

void
JNIFUNC(GetCharArrayRegion)(JNIEnv *env,
			    jcharArray array,
			    jsize start,
			    jsize len,
			    jchar* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetCharArrayRegion");
      return;
    }

  memcpy(buf, ((jchar*)body_item.value.l) + start, len * sizeof(jchar));
}

void
JNIFUNC(SetCharArrayRegion)(JNIEnv *env,
			    jcharArray array,
			    jsize start,
			    jsize len,
			    jchar* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetCharArrayRegion");
      return;
    }

  memcpy(((jchar*)body_item.value.l) + start, buf, len * sizeof(jchar));
}

jshortArray
JNIFUNC(NewShortArray)(JNIEnv *env,
		       jsize length)
{
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  ClazzFile *short_array_type = createFakeArrayClass(env, "[S");
  jobject array = new_array(env,length, short_array_type);
  
  ENQUEUE((GC_obj*)array, henv->_local_refs);

  return array;
}

jshort*
JNIFUNC(GetShortArrayElements)(JNIEnv *env,
			       jshortArray array,
			       jboolean* isCopy)
{
  JavaStackItem body_item, length_item;
  jshort *the_copy;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  the_copy = (jshort*)malloc(sizeof(jshort) * length_item.value.i);

  if (the_copy == NULL)
    {
      if (isCopy) *isCopy = JNI_FALSE;
      return (jshort*)body_item.value.l;
    }
  else
    {
      memcpy(the_copy, body_item.value.l, sizeof(jshort) * length_item.value.i);
      if (isCopy) *isCopy = JNI_TRUE;
      return the_copy;
    }
}

void
JNIFUNC(ReleaseShortArrayElements)(JNIEnv *env,
				   jshortArray array,
				   jshort* elems,
				   jint mode)
{
  JavaStackItem body_item, length_item;
  jboolean can_free = JNI_FALSE;

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (body_item.value.l != elems)
    can_free = JNI_TRUE;

  switch (mode)
    {
    case 0: /* commit the changes and free elems. */
      memcpy(body_item.value.l, elems, sizeof(jshort) * length_item.value.i);
      if (can_free) free(elems);
      break;
    case JNI_COMMIT: /* commit the changes but don't free elems. */
      memcpy(body_item.value.l, elems, sizeof(jshort) * length_item.value.i);
      break;
    case JNI_ABORT:  /* don't commit the changes but free elems. */
      if (can_free) free(elems);
      break;
    }
}

void
JNIFUNC(GetShortArrayRegion)(JNIEnv *env,
			     jshortArray array,
			     jsize start,
			     jsize len,
			     jshort* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetShortArrayRegion");
      return;
    }

  memcpy(buf, ((jshort*)body_item.value.l) + start, len * sizeof(jshort));
}

void
JNIFUNC(SetShortArrayRegion)(JNIEnv *env,
			     jshortArray array,
			     jsize start,
			     jsize len,
			     jshort* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetShortArrayRegion");
      return;
    }

  memcpy(((jshort*)body_item.value.l) + start, buf, len * sizeof(jshort));
}

jintArray
JNIFUNC(NewIntArray)(JNIEnv *env,
		     jsize length)
{
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  ClazzFile *int_array_type = createFakeArrayClass(env, "[I");
  jobject array = new_array(env,length, int_array_type);
  
  ENQUEUE((GC_obj*)array, henv->_local_refs);

  return array;
}

jint*
JNIFUNC(GetIntArrayElements)(JNIEnv *env,
			     jintArray array,
			     jboolean* isCopy)
{
  JavaStackItem body_item, length_item;
  jint *the_copy;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  the_copy = (jint*)malloc(sizeof(jint) * length_item.value.i);

  if (the_copy == NULL)
    {
      if (isCopy) *isCopy = JNI_FALSE;
      return (jint*)body_item.value.l;
    }
  else
    {
      memcpy(the_copy, body_item.value.l, sizeof(jint) * length_item.value.i);
      if (isCopy) *isCopy = JNI_TRUE;
      return the_copy;
    }
}

void
JNIFUNC(ReleaseIntArrayElements)(JNIEnv *env,
				 jintArray array,
				 jint* elems,
				 jint mode)
{
  JavaStackItem body_item, length_item;
  jboolean can_free = JNI_FALSE;

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (body_item.value.l != elems)
    can_free = JNI_TRUE;

  switch (mode)
    {
    case 0: /* commit the changes and free elems. */
      memcpy(body_item.value.l, elems, sizeof(jint) * length_item.value.i);
      if (can_free) free(elems);
      break;
    case JNI_COMMIT: /* commit the changes but don't free elems. */
      memcpy(body_item.value.l, elems, sizeof(jint) * length_item.value.i);
      break;
    case JNI_ABORT:  /* don't commit the changes but free elems. */
      if (can_free) free(elems);
      break;
    }
}

void
JNIFUNC(GetIntArrayRegion)(JNIEnv *env,
			   jintArray array,
			   jsize start,
			   jsize len,
			   jint* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetIntArrayRegion");
      return;
    }

  memcpy(buf, ((jint*)body_item.value.l) + start, len * sizeof(jint));
}

void
JNIFUNC(SetIntArrayRegion)(JNIEnv *env,
			   jintArray array,
			   jsize start,
			   jsize len,
			   jint* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetIntArrayRegion");
      return;
    }

  memcpy(((jint*)body_item.value.l) + start, buf, len * sizeof(jint));
}

jlongArray
JNIFUNC(NewLongArray)(JNIEnv *env,
		      jsize length)
{
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  ClazzFile *long_array_type = createFakeArrayClass(env, "[J");
  jobject array = new_array(env,length, long_array_type);
  
  ENQUEUE((GC_obj*)array, henv->_local_refs);

  return array;
}

jlong*
JNIFUNC(GetLongArrayElements)(JNIEnv *env,
			      jlongArray array,
			      jboolean* isCopy)
{
  JavaStackItem body_item, length_item;
  jlong *the_copy;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  the_copy = (jlong*)malloc(sizeof(jlong) * length_item.value.i);

  if (the_copy == NULL)
    {
      if (isCopy) *isCopy = JNI_FALSE;
      return (jlong*)body_item.value.l;
    }
  else
    {
      memcpy(the_copy, body_item.value.l, sizeof(jlong) * length_item.value.i);
      if (isCopy) *isCopy = JNI_TRUE;
      return the_copy;
    }
}

void
JNIFUNC(ReleaseLongArrayElements)(JNIEnv *env,
				  jlongArray array,
				  jlong* elems,
				  jint mode)
{
  JavaStackItem body_item, length_item;
  jboolean can_free = JNI_FALSE;

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (body_item.value.l != elems)
    can_free = JNI_TRUE;

  switch (mode)
    {
    case 0: /* commit the changes and free elems. */
      memcpy(body_item.value.l, elems, sizeof(jlong) * length_item.value.i);
      if (can_free) free(elems);
      break;
    case JNI_COMMIT: /* commit the changes but don't free elems. */
      memcpy(body_item.value.l, elems, sizeof(jlong) * length_item.value.i);
      break;
    case JNI_ABORT:  /* don't commit the changes but free elems. */
      if (can_free) free(elems);
      break;
    }
}

void
JNIFUNC(GetLongArrayRegion)(JNIEnv *env,
			    jlongArray array,
			    jsize start,
			    jsize len,
			    jlong* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetLongArrayRegion");
      return;
    }

  memcpy(buf, ((jlong*)body_item.value.l) + start, len * sizeof(jlong));
}

void
JNIFUNC(SetLongArrayRegion)(JNIEnv *env,
			    jlongArray array,
			    jsize start,
			    jsize len,
			    jlong* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetLongArrayRegion");
      return;
    }

  memcpy(((jlong*)body_item.value.l) + start, buf, len * sizeof(jlong));
}

jfloatArray
JNIFUNC(NewFloatArray)(JNIEnv *env,
		       jsize length)
{
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  ClazzFile *float_array_type = createFakeArrayClass(env, "[F");
  jobject array = new_array(env,length, float_array_type);
  
  ENQUEUE((GC_obj*)array, henv->_local_refs);

  return array;
}

jfloat*
JNIFUNC(GetFloatArrayElements)(JNIEnv *env,
			       jfloatArray array,
			       jboolean* isCopy)
{
  JavaStackItem body_item, length_item;
  jfloat *the_copy;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  the_copy = (jfloat*)malloc(sizeof(jfloat) * length_item.value.i);

  if (the_copy == NULL)
    {
      if (isCopy) *isCopy = JNI_FALSE;
      return (jfloat*)body_item.value.l;
    }
  else
    {
      memcpy(the_copy, body_item.value.l, sizeof(jfloat) * length_item.value.i);
      if (isCopy) *isCopy = JNI_TRUE;
      return the_copy;
    }
}

void
JNIFUNC(ReleaseFloatArrayElements)(JNIEnv *env,
				   jfloatArray array,
				   jfloat* elems,
				   jint mode)
{
  JavaStackItem body_item, length_item;
  jboolean can_free = JNI_FALSE;

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (body_item.value.l != elems)
    can_free = JNI_TRUE;

  switch (mode)
    {
    case 0: /* commit the changes and free elems. */
      memcpy(body_item.value.l, elems, sizeof(jfloat) * length_item.value.i);
      if (can_free) free(elems);
      break;
    case JNI_COMMIT: /* commit the changes but don't free elems. */
      memcpy(body_item.value.l, elems, sizeof(jfloat) * length_item.value.i);
      break;
    case JNI_ABORT:  /* don't commit the changes but free elems. */
      if (can_free) free(elems);
      break;
    }
}

void
JNIFUNC(GetFloatArrayRegion)(JNIEnv *env,
			     jfloatArray array,
			     jsize start,
			     jsize len,
			     jfloat* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetFloatArrayRegion");
      return;
    }

  memcpy(buf, ((jfloat*)body_item.value.l) + start, len * sizeof(jfloat));
}

void
JNIFUNC(SetFloatArrayRegion)(JNIEnv *env,
			     jfloatArray array,
			     jsize start,
			     jsize len,
			     jfloat* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetFloatArrayRegion");
      return;
    }

  memcpy(((jfloat*)body_item.value.l) + start, buf, len * sizeof(jfloat));
}

jdoubleArray
JNIFUNC(NewDoubleArray)(JNIEnv *env,
			jsize length)
{
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  ClazzFile *double_array_type = createFakeArrayClass(env, "[D");
  jobject array = new_array(env,length, double_array_type);
  
  ENQUEUE((GC_obj*)array, henv->_local_refs);

  return array;
}

jdouble*
JNIFUNC(GetDoubleArrayElements)(JNIEnv *env,
				jdoubleArray array,
				jboolean* isCopy)
{
  JavaStackItem body_item, length_item;
  jdouble *the_copy;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  the_copy = (jdouble*)malloc(sizeof(jdouble) * length_item.value.i);

  if (the_copy == NULL)
    {
      if (isCopy) *isCopy = JNI_FALSE;
      return (jdouble*)body_item.value.l;
    }
  else
    {
      memcpy(the_copy, body_item.value.l, sizeof(jdouble) * length_item.value.i);
      if (isCopy) *isCopy = JNI_TRUE;
      return the_copy;
    }
}

void
JNIFUNC(ReleaseDoubleArrayElements)(JNIEnv *env,
				    jdoubleArray array,
				    jdouble* elems,
				    jint mode)
{
  JavaStackItem body_item, length_item;
  jboolean can_free = JNI_FALSE;

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (body_item.value.l != elems)
    can_free = JNI_TRUE;

  switch (mode)
    {
    case 0: /* commit the changes and free elems. */
      memcpy(body_item.value.l, elems, sizeof(jdouble) * length_item.value.i);
      if (can_free) free(elems);
      break;
    case JNI_COMMIT: /* commit the changes but don't free elems. */
      memcpy(body_item.value.l, elems, sizeof(jdouble) * length_item.value.i);
      break;
    case JNI_ABORT:  /* don't commit the changes but free elems. */
      if (can_free) free(elems);
      break;
    }
}

void
JNIFUNC(GetDoubleArrayRegion)(JNIEnv *env,
			      jdoubleArray array,
			      jsize start,
			      jsize len,
			      jdouble* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetDoubleArrayRegion");
      return;
    }

  memcpy(buf, ((jdouble*)body_item.value.l) + start, len * sizeof(jdouble));
}

void
JNIFUNC(SetDoubleArrayRegion)(JNIEnv *env,
			      jdoubleArray array,
			      jsize start,
			      jsize len,
			      jdouble* buf)
{
  JavaStackItem body_item, length_item;
  
  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  get_instance_field(array,
		     STATIC_DATA(array)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (start < 0 || start + len >= length_item.value.i)
    {
      jclass exception = (*env)->FindClass(env, "java/lang/ArrayIndexOutOfBoundsException");
      (*env)->ThrowNew(env, exception, "In JNI GetDoubleArrayRegion");
      return;
    }

  memcpy(((jdouble*)body_item.value.l) + start, buf, len * sizeof(jdouble));
}
