/* -*- Mode: C; c-file-style: "gnu" -*-
   jniinvoc.c -- Java Native Interface Invocation API.
   Created: Chris Toshok <toshok@hungry.com>, 26-Jul-1997
*/
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "jni.h"
#include "jniint.h"
#include "gc.h"
#include "qop.h"
#include "log.h"
#include "native-threads.h"
#include "dynamic_loading.h"
#include <assert.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>

/* XXX should be in some include file [pere] */
void initialize_system_class(JNIEnv *env);
void initialize_system_threads(JNIEnv *env);

/* our list of vms, and the monitor to lock when mucking with the list. */
static HungryJavaVM *vms;
static HMonitor vms_monitor = NULL;

/* the sun vtable */
extern struct JNINativeInterface _hungryJNIInterface;

static void
free_java_vm(JavaVM *vm)
{
  HungryJavaVM *hvm = (HungryJavaVM*)vm;

  /* free the vm's classpath */
  CLASSPATH_destroy(hvm->_cp_entries, hvm->_num_cp_entries);

  /* it monitor */
  MONITOR_destroy(hvm->_mon);

  /* and it initial env */
  /* XXX does something else need to happen here? */
  free(hvm->_initial_env);
}

static jint
JNIFUNC(DestroyJavaVM)(JavaVM *vm)
{
  HungryJavaVM *hvm = (HungryJavaVM*)vm;

  MONITOR_enter(hvm->_mon);

  if (THREAD_getCurrent() != hvm->_initial_env->_thr)
    {
      MONITOR_exit(hvm->_mon);
  
      return -1;
    }
  else
    {
#if 0
      /* loop over the env's, looking for a non-daemon thread. */
      HungryJNIEnv *env;

      for (env = hvm->_envs; env != NULL; env = env->next)
	{
	  if (env_is_not_daemon(env))
	    return -1;
	}
#else
      /* just return -1 if there is more than one thread. */
      if (hvm->_envs->next)
	return -1;
      else
	{
#endif
	  UNQUEUE(hvm, vms);
		  
#if the_garbage_collector_worked
	  /* Since we're deleting this vm, collect its garbage. */
	  GC_runCollector(vm);

	  if (hvm->_args.enableVerboseGC)
	    GC_printStats(vm);
#endif
		  
	  MONITOR_exit(hvm->_mon);

	  free_java_vm(vm);
	}
    }

  return 0;
}

static jint
JNIFUNC(AttachCurrentThread)(JavaVM *vm,
			     JNIEnv **p_env,
			     void *thr_args)
{
  JNIEnv *env = THREAD_getEnv();
  HungryJavaVM *hvm = (HungryJavaVM*)vm;
  HungryJNIEnv *henv;
  
  MONITOR_enter(hvm->_mon);

  /* if the thread has an env, it's been attached already, so just quit */
  if (env != NULL) goto err;

  /* otherwise, we allocate an env */
  henv = (HungryJNIEnv*)calloc(1, sizeof(HungryJNIEnv));
  if (!henv) goto err;

  env = (JNIEnv*)henv;
  henv->_env = &_hungryJNIInterface;

  henv->_vm = (HungryJavaVM*)vm;
  henv->_thr = THREAD_getCurrent();
  THREAD_setEnv(env);

  ENQUEUE(henv, ((HungryJavaVM*)vm)->_envs);

  MONITOR_exit(hvm->_mon);
  return 0;

err:
  MONITOR_exit(hvm->_mon);
  return -1;
}

static jint
JNIFUNC(DetachCurrentThread)(JavaVM *vm)
{
  JNIEnv *env = THREAD_getEnv();
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  HungryJavaVM *hvm = (HungryJavaVM*)vm;

  MONITOR_enter(hvm->_mon);

  /* if it isn't an attached thread */
  if (!henv)
    goto err;    
  
  /* if it's the initial thread -- the one that created the java vm */
  if (henv == hvm->_initial_env)
    goto err;    
  
  /* if it's an attached thread, but attached to a different vm */
  if (henv->_vm != hvm)
    goto err;    
  
  UNQUEUE(henv, hvm->_envs);

  free(henv);

  MONITOR_exit(hvm->_mon);
  return 0;

err:
  MONITOR_exit(hvm->_mon);
  return -1;
}

const struct JNIInvokeInterface _hungryJNIInvokeInterface = {
  NULL,
  NULL,
  NULL,
  JNIFUNC(DestroyJavaVM),
  JNIFUNC(AttachCurrentThread),
  JNIFUNC(DetachCurrentThread),
};

jint
JNI_GetDefaultJavaVMInitArgs(void *vm_args)
{
  JDK1_1InitArgs *args = (JDK1_1InitArgs*)vm_args;
  char *classpath_env;

  if (args->version != JNIVERS)
    return -1;

  args->version = JNIVERS;

  args->properties = NULL;

  args->checkSource = 0;
  args->nativeStackSize = 0; /* should be 16 megs? */
  args->javaStackSize = 0;
  args->minHeapSize = 0; /* should be 16 megs? */
  args->maxHeapSize = 0; /* should be what? */
  args->verifyMode = VERIFY_ALL;

  /*
  ** setup the classpath.  semantics are as follows:
  ** if CLASSPATH is not set, we use the system classpath.
  ** if CLASSPATH is set, we use the value of CLASSPATH, and then append
  **   the system classpath.
  */
  classpath_env = getenv("CLASSPATH");
  if (!classpath_env)
    args->classpath = CLASSPATH_getSystemClasspath();
  else 
    {
      char *sys_path = CLASSPATH_getSystemClasspath();
      args->classpath = (char *) malloc(strlen(classpath_env) +
					strlen(sys_path) + 2);
      strcpy((char*)args->classpath, classpath_env);
      strcat((char*)args->classpath, ":");
      strcat((char*)args->classpath, sys_path);
      free(sys_path);
    }

  args->vfprintf = NULL;

  args->exit = NULL;
  args->abort = NULL;
  args->enableClassGC = JNI_FALSE; /* don't enable this yet... it's buggy. */
  args->enableVerboseGC = JNI_FALSE;
  /* XXX hungry added field... */
  args->enableVerboseClassLoading = JNI_FALSE;
  args->disableAsyncGC = JNI_TRUE;

  return 0;
}

jint
JNI_GetCreatedJavaVMs(JavaVM **vmBuf,
		      jsize bufLen,
		      jsize *nVMs)

{
  HungryJavaVM *cur;
  int count = 0;

  MONITOR_enter(vms_monitor);

  /* we loop through the entire list of vm's to get the count
     right, but we only assign things into vmBuf if it's not
     null and if its long enough (bufLen is big enough.) */
  for (cur = (HungryJavaVM*)vms;
       cur != NULL;
       cur = cur->next)
    {
      if (vmBuf && bufLen > count)
	vmBuf[count] = (JavaVM*)cur;
      count ++;
    }
  
  if (nVMs)
    *nVMs = count;

  MONITOR_exit(vms_monitor);

  return 0;
}

void fatal_signal_handler();

jint
JNI_CreateJavaVM(JavaVM **p_vm,
		 JNIEnv **p_env,
		 void *vm_args)
{
  HungryJavaVM *new_vm;
  HungryJNIEnv *new_env;
  JavaVM *vm;
  JNIEnv *env;
  struct sigaction sigact;

  LOG_init();

  if (vms_monitor == NULL)
    vms_monitor = MONITOR_createNamed("Java VM's Monitor");

  /* allocate our larger vm struct, and memcpy the sun vtable into
     new_vm->_vm */
  new_vm = (HungryJavaVM*)calloc(1, sizeof(HungryJavaVM));
  vm = (JavaVM*)new_vm;

  new_vm->_mon = MONITOR_create();
  MONITOR_enter(new_vm->_mon);

  if (new_vm == NULL)
    {
      MONITOR_exit(new_vm->_mon);
      return -1; /* leak */
    }
  new_vm->_vm = &_hungryJNIInvokeInterface;

  /* allocate an env to tie to this thread, and memcpy the sun vtable into
     new_env->_env */
  new_env = (HungryJNIEnv *)calloc(1, sizeof(HungryJNIEnv));
  if (new_env == NULL)
    {
      MONITOR_exit(new_vm->_mon);
      return -1; /* XXX leaks the new_vm */
    }
  new_env->_env = &_hungryJNIInterface;
  env = (JNIEnv*)new_env;

  /* add the env to the vm's list of envs, and make it the special, "initial"
     env */
  new_vm->_initial_env = new_env;
  ENQUEUE(new_env, new_vm->_envs);

  /* add the backpointer to the env's vm, and let the env know what its
     native thread is */
  new_env->_vm = new_vm;
  new_env->_thr = THREAD_getCurrent();

  if (NULL == new_env->_thr)
    fprintf(stderr, "Warning: THREAD_getCurrent() == NULL\n");

  /* save off the env in the native thread's per thread info so we can get
     it back later */
  THREAD_setEnv(env);
  THREAD_setVM(vm);

  /* create our GC_state. */

  new_vm->_gc_state = GC_createGCState((JavaVM*)new_vm);

  /* make sure our native symbols are loading before we go trying to execute java code. */
  DLL_loadSystemLibraries();

  /* make sure that the env is actually there. */
  assert(THREAD_getEnv() != NULL);

  /* save off the args with which this vm was created */
  memcpy(&new_vm->_args, vm_args, sizeof(JDK1_1InitArgs));

  /* create our classpath */
  CLASSPATH_create(new_vm->_args.classpath, &new_vm->_cp_entries, &new_vm->_num_cp_entries);

  /* create our class repository (hash table of ClazzFile*'s and class descriptors. */
  initialize_class_repository((JavaVM*)new_vm);

  /* get the thread machinery going. */
  initialize_system_threads((JNIEnv*)new_env);

  if ((*env)->ExceptionOccurred(env))
    {
      (*env)->ExceptionDescribe(env);
      (*env)->FatalError(env, "Exception occurred creating a java vm.");;
    }

  sigact.sa_handler = fatal_signal_handler;
  sigemptyset(&sigact.sa_mask);
  sigact.sa_flags = 0;

  sigaction(SIGBUS, &sigact, NULL);
  sigaction(SIGSEGV, &sigact, NULL);

  /* return the vm and env to the caller */
  *p_vm = vm;
  *p_env = env;

  MONITOR_exit(new_vm->_mon);
  return 0;
}
