/* -*- Mode: C; c-file-style: "gnu" -*-
   jnimeth.c -- Java Native Interface methods relating to instance methods.
   Created: Chris Toshok <toshok@hungry.com>, 26-Jul-1997
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "jni.h"
#include "jniint.h"
#include "interp.h"
#include "ClazzFile.h"
#include "objects.h"
#include <stdlib.h>

jmethodID
JNIFUNC(GetMethodID)(JNIEnv *env,
		     jclass clazz,
		     const char *name,
		     const char *sig)
{
  MethodStruct* method = NULL;
  ClazzFile* cf = (ClazzFile*)clazz;
  Signature *method_sig;
  int method_num;

  if (clazz == NULL)
    {
      /* XXX throw something */
      return NULL;
    }

  /* XXX initialize the class, if needed, throwing ExceptionInInitializerError
     if there was an error. */

  method_sig = SIG_parseFromJavaSig(env, (char*)sig);
  if (method_sig == NULL)
    {
      /* throw OutOfMemoryError */
      return NULL;
    }

  for (method_num = 0; method_num < cf->num_methods; method_num++)
    {
      MethodStruct *cur_method = cf->methods[method_num];

      if (!strcmp(cur_method->name, name)
	  && SIG_equal(env, method_sig, cur_method->sig)
	  && (cur_method->access_flags & ACC_STATIC) == 0)
	{
	  method = cur_method;
	  break;
	}
    }

  SIG_free(env, method_sig);

  if (method == NULL)
    /* XXX throw NoSuchMethodError */;

  return method;
}

jboolean
JNIFUNC(CallBooleanMethod)(JNIEnv *env,
			   jobject obj,
			   jmethodID methodID,
			   ...)
{
  va_list varargs;
  jboolean result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallBooleanMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jboolean
JNIFUNC(CallBooleanMethodA)(JNIEnv *env,
			    jobject obj,
			    jmethodID methodID,
			    jvalue *args)
{
  jvalue result = CallJavaMethod((MethodStruct*)methodID,
				 obj, args);
  

  return result.z;
}

jboolean
JNIFUNC(CallBooleanMethodV)(JNIEnv *env,
			    jobject obj,
			    jmethodID methodID,
			    va_list args)
{
  jvalue *params;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  int i;
  jboolean result;

  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  result = JNIFUNC(CallBooleanMethodA)(env, obj, methodID, params);
  
  free(params);

  return result;
}

jbyte
JNIFUNC(CallByteMethod)(JNIEnv *env,
			jobject obj,
			jmethodID methodID,
			...)
{
  va_list varargs;
  jbyte result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallByteMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jbyte
JNIFUNC(CallByteMethodA)(JNIEnv *env,
			 jobject obj,
			 jmethodID methodID,
			 jvalue *args)
{
  jvalue result = CallJavaMethod((MethodStruct*)methodID,
				 obj, args);
  

  return result.b;
}

jbyte
JNIFUNC(CallByteMethodV)(JNIEnv *env,
			 jobject obj,
			 jmethodID methodID,
			 va_list args)
{
  jvalue *params;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  int i;
  jbyte result;

  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  result = JNIFUNC(CallByteMethodA)(env, obj, methodID, params);
  
  free(params);

  return result;
}

jchar
JNIFUNC(CallCharMethod)(JNIEnv *env,
			jobject obj,
			jmethodID methodID,
			...)
{ 
  va_list varargs;
  jchar result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallCharMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jchar
JNIFUNC(CallCharMethodA)(JNIEnv *env,
			 jobject obj,
			 jmethodID methodID,
			 jvalue *args)
{
  jvalue result = CallJavaMethod((MethodStruct*)methodID,
				 obj, args);
  

  return result.c;
}

jchar
JNIFUNC(CallCharMethodV)(JNIEnv *env,
			 jobject obj,
			 jmethodID methodID,
			 va_list args)
{
  jvalue *params;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  int i;
  jchar result;

  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  result = JNIFUNC(CallCharMethodA)(env, obj, methodID, params);
  
  free(params);

  return result;
}

jshort
JNIFUNC(CallShortMethod)(JNIEnv *env,
			 jobject obj,
			 jmethodID methodID,
			 ...)
{
  va_list varargs;
  jshort result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallShortMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jshort
JNIFUNC(CallShortMethodA)(JNIEnv *env,
			  jobject obj,
			  jmethodID methodID,
			  jvalue *args)
{
  jvalue result = CallJavaMethod((MethodStruct*)methodID,
				 obj, args);
  

  return result.s;
}

jshort
JNIFUNC(CallShortMethodV)(JNIEnv *env,
			  jobject obj,
			  jmethodID methodID,
			  va_list args)
{
  jvalue *params;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  int i;
  jshort result;

  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  result = JNIFUNC(CallShortMethodA)(env, obj, methodID, params);
  
  free(params);

  return result;
}

jint
JNIFUNC(CallIntMethod)(JNIEnv *env,
		       jobject obj,
		       jmethodID methodID,
		       ...)
{
  va_list varargs;
  jint result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallIntMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jint
JNIFUNC(CallIntMethodA)(JNIEnv *env,
			jobject obj,
			jmethodID methodID,
			jvalue *args)
{
  jvalue result = CallJavaMethod((MethodStruct*)methodID,
				 obj, args);
  

  return result.i;
}

jint
JNIFUNC(CallIntMethodV)(JNIEnv *env,
			jobject obj,
			jmethodID methodID,
			va_list args)
{
  jvalue *params;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  int i;
  jint result;

  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  result = JNIFUNC(CallIntMethodA)(env, obj, methodID, params);
  
  free(params);

  return result;
}

jlong
JNIFUNC(CallLongMethod)(JNIEnv *env,
			jobject obj,
			jmethodID methodID,
			...)
{
  va_list varargs;
  jlong result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallLongMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jlong
JNIFUNC(CallLongMethodA)(JNIEnv *env,
			 jobject obj,
			 jmethodID methodID,
			 jvalue *args)
{
  jvalue result = CallJavaMethod((MethodStruct*)methodID,
				 obj, args);
  

  return result.j;
}

jlong
JNIFUNC(CallLongMethodV)(JNIEnv *env,
			 jobject obj,
			 jmethodID methodID,
			 va_list args)
{
  jvalue *params;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  int i;
  jlong result;

  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  result = JNIFUNC(CallLongMethodA)(env, obj, methodID, params);
  
  free(params);

  return result;
}

jfloat
JNIFUNC(CallFloatMethod)(JNIEnv *env,
			 jobject obj,
			 jmethodID methodID,
			 ...)
{
  va_list varargs;
  jfloat result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallFloatMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jfloat
JNIFUNC(CallFloatMethodA)(JNIEnv *env,
			  jobject obj,
			  jmethodID methodID,
			  jvalue *args)
{
  jvalue result = CallJavaMethod((MethodStruct*)methodID,
				 obj, args);
  

  return result.f;
}

jfloat
JNIFUNC(CallFloatMethodV)(JNIEnv *env,
			  jobject obj,
			  jmethodID methodID,
			  va_list args)
{
  jvalue *params;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  int i;
  jfloat result;

  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  result = JNIFUNC(CallFloatMethodA)(env, obj, methodID, params);
  
  free(params);

  return result;
}

jdouble
JNIFUNC(CallDoubleMethod)(JNIEnv *env,
			  jobject obj,
			  jmethodID methodID,
			  ...)
{
  va_list varargs;
  jdouble result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallDoubleMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jdouble
JNIFUNC(CallDoubleMethodA)(JNIEnv *env,
			   jobject obj,
			   jmethodID methodID,
			   jvalue *args)
{
  jvalue result = CallJavaMethod((MethodStruct*)methodID,
				 obj, args);
  

  return result.d;
}

jdouble
JNIFUNC(CallDoubleMethodV)(JNIEnv *env,
			   jobject obj,
			   jmethodID methodID,
			   va_list args)
{
  jvalue *params;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  int i;
  jdouble result;

  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  result = JNIFUNC(CallDoubleMethodA)(env, obj, methodID, params);
  
  free(params);

  return result;
}

void
JNIFUNC(CallVoidMethod)(JNIEnv *env,
			jobject obj,
			jmethodID methodID,
			...)
{
  va_list varargs;

  va_start(varargs, methodID);
  JNIFUNC(CallVoidMethodV)(env, obj, methodID, varargs);
  va_end(varargs);
}

void
JNIFUNC(CallVoidMethodA)(JNIEnv *env,
			 jobject obj,
			 jmethodID methodID,
			 jvalue *args)
{
  CallJavaMethod((MethodStruct*)methodID,
		 obj,
		 args);
}

void
JNIFUNC(CallVoidMethodV)(JNIEnv *env,
			 jobject obj,
			 jmethodID methodID,
			 va_list args)
{
  jvalue *params;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  int i;

  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  JNIFUNC(CallVoidMethodA)(env, obj, methodID, params);
  
  free(params);
}


jobject
JNIFUNC(CallObjectMethod)(JNIEnv *env,
			  jobject obj,
			  jmethodID methodID,
			  ...)
{
  va_list varargs;
  jobject result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallObjectMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jobject
JNIFUNC(CallObjectMethodA)(JNIEnv *env,
			   jobject obj,
			   jmethodID methodID,
			   jvalue *args)
{
  jvalue value = CallJavaMethod((MethodStruct*)methodID,
				obj,
				args);

  return value.l;
}

jobject
JNIFUNC(CallObjectMethodV)(JNIEnv *env,
			   jobject obj,
			   jmethodID methodID,
			   va_list args)
{
  jvalue *params;
  int num_params = SIG_numParams(env, ((MethodStruct*)methodID)->sig);
  int i;
  jobject result;

  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    params[i] = va_arg(args, jvalue);

  result = JNIFUNC(CallObjectMethodA)(env, obj, methodID, params);
  
  free(params);

  return result;
}


jboolean
JNIFUNC(CallNonvirtualBooleanMethod)(JNIEnv *env,
				     jobject obj,
				     jmethodID methodID,
				     ...)
{
  va_list varargs;
  jboolean result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallNonvirtualBooleanMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jboolean
JNIFUNC(CallNonvirtualBooleanMethodA)(JNIEnv *env,
				      jobject obj,
				      jmethodID methodID,
				      jvalue *args)
{
  abort();
}

jboolean
JNIFUNC(CallNonvirtualBooleanMethodV)(JNIEnv *env,
				      jobject obj,
				      jmethodID methodID,
				      va_list args)
{
  abort();
}

jbyte
JNIFUNC(CallNonvirtualByteMethod)(JNIEnv *env,
				  jobject obj,
				  jmethodID methodID,
				  ...)
{
  va_list varargs;
  jbyte result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallNonvirtualByteMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jbyte
JNIFUNC(CallNonvirtualByteMethodA)(JNIEnv *env,
				   jobject obj,
				   jmethodID methodID,
				   jvalue *args)
{
  abort();
}

jbyte
JNIFUNC(CallNonvirtualByteMethodV)(JNIEnv *env,
				   jobject obj,
				   jmethodID methodID,
				   va_list args)
{
  abort();
}

jchar
JNIFUNC(CallNonvirtualCharMethod)(JNIEnv *env,
				  jobject obj,
				  jmethodID methodID,
				  ...)
{
  va_list varargs;
  jchar result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallNonvirtualCharMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jchar
JNIFUNC(CallNonvirtualCharMethodA)(JNIEnv *env,
				   jobject obj,
				   jmethodID methodID,
				   jvalue *args)
{
  abort();
}

jchar
JNIFUNC(CallNonvirtualCharMethodV)(JNIEnv *env,
				   jobject obj,
				   jmethodID methodID,
				   va_list args)
{
  abort();
}

jshort
JNIFUNC(CallNonvirtualShortMethod)(JNIEnv *env,
				   jobject obj,
				   jmethodID methodID,
				   ...)
{
  va_list varargs;
  jshort result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallNonvirtualShortMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jshort
JNIFUNC(CallNonvirtualShortMethodA)(JNIEnv *env,
				    jobject obj,
				    jmethodID methodID,
				    jvalue *args)
{
  abort();
}

jshort
JNIFUNC(CallNonvirtualShortMethodV)(JNIEnv *env,
				    jobject obj,
				    jmethodID methodID,
				    va_list args)
{
  abort();
}

jint
JNIFUNC(CallNonvirtualIntMethod)(JNIEnv *env,
				 jobject obj,
				 jmethodID methodID,
				 ...)
{
  va_list varargs;
  jint result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallNonvirtualIntMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jint
JNIFUNC(CallNonvirtualIntMethodA)(JNIEnv *env,
				  jobject obj,
				  jmethodID methodID,
				  jvalue *args)
{
  abort();
}

jint
JNIFUNC(CallNonvirtualIntMethodV)(JNIEnv *env,
				  jobject obj,
				  jmethodID methodID,
				  va_list args)
{
  abort();
}

jlong
JNIFUNC(CallNonvirtualLongMethod)(JNIEnv *env,
				  jobject obj,
				  jmethodID methodID,
				  ...)
{
  va_list varargs;
  jlong result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallNonvirtualLongMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jlong
JNIFUNC(CallNonvirtualLongMethodA)(JNIEnv *env,
				   jobject obj,
				   jmethodID methodID,
				   jvalue *args)
{
  abort();
}

jlong
JNIFUNC(CallNonvirtualLongMethodV)(JNIEnv *env,
				   jobject obj,
				   jmethodID methodID,
				   va_list args)
{
  abort();
}

jfloat
JNIFUNC(CallNonvirtualFloatMethod)(JNIEnv *env,
				   jobject obj,
				   jmethodID methodID,
				   ...)
{
  va_list varargs;
  jfloat result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallNonvirtualFloatMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jfloat
JNIFUNC(CallNonvirtualFloatMethodA)(JNIEnv *env,
				    jobject obj,
				    jmethodID methodID,
				    jvalue *args)
{
  abort();
}

jfloat
JNIFUNC(CallNonvirtualFloatMethodV)(JNIEnv *env,
				    jobject obj,
				    jmethodID methodID,
				    va_list args)
{
  abort();
}

jdouble
JNIFUNC(CallNonvirtualDoubleMethod)(JNIEnv *env,
				    jobject obj,
				    jmethodID methodID,
				    ...)
{
  va_list varargs;
  jdouble result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallNonvirtualDoubleMethodV)(env, obj, methodID, varargs);
  va_end(varargs);

  return result;
}

jdouble
JNIFUNC(CallNonvirtualDoubleMethodA)(JNIEnv *env,
				     jobject obj,
				     jmethodID methodID,
				     jvalue *args)
{
  abort();
}

jdouble
JNIFUNC(CallNonvirtualDoubleMethodV)(JNIEnv *env,
				     jobject obj,
				     jmethodID methodID,
				     va_list args)
{
  abort();
}

void
JNIFUNC(CallNonvirtualVoidMethod)(JNIEnv *env,
				  jobject obj,
				  jmethodID methodID,
				  ...)
{
  va_list varargs;

  va_start(varargs, methodID);
  JNIFUNC(CallNonvirtualVoidMethodV)(env, obj, methodID, varargs);
  va_end(varargs);
}

void
JNIFUNC(CallNonvirtualVoidMethodA)(JNIEnv *env,
				   jobject obj,
				   jmethodID methodID,
				   jvalue *args)
{
  CallJavaMethod((MethodStruct*)methodID,
		 obj,
		 args);
}

void
JNIFUNC(CallNonvirtualVoidMethodV)(JNIEnv *env,
				   jobject obj,
				   jmethodID methodID,
				   va_list args)
{
  jvalue *params;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  int i;

  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  JNIFUNC(CallNonvirtualVoidMethodA)(env, obj, methodID, params);
  
  free(params);
}

jobject
JNIFUNC(CallNonvirtualObjectMethod)(JNIEnv *env,
				    jobject obj,
				    jmethodID methodID,
				    ...)
{
  abort();
}

jobject
JNIFUNC(CallNonvirtualObjectMethodA)(JNIEnv *env,
				     jobject obj,
				     jmethodID methodID,
				     jvalue *args)
{
  abort();
}

jobject
JNIFUNC(CallNonvirtualObjectMethodV)(JNIEnv *env,
				     jobject obj,
				     jmethodID methodID,
				     va_list args)
{
  abort();
}

