/* -*- Mode: C; c-file-style: "gnu" -*-
   jnismeth.c -- Java Native Interface methods relating to static methods.
   Created: Chris Toshok <toshok@hungry.com>, 26-Jul-1997
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "jni.h"
#include "jniint.h"
#include "interp.h"
#include <stdlib.h>

jmethodID
JNIFUNC(GetStaticMethodID)(JNIEnv *env,
			   jclass clazz,
			   const char *name,
			   const char *sig)
{
  MethodStruct* method = NULL;
  ClazzFile *cf = (ClazzFile*)clazz;
  Signature *method_sig;
  int method_num;

  /* XXX initialize the class, if needed, throwing ExceptionInInitializerError
     if there was an error. */

  method_sig = SIG_parseFromJavaSig(env, (char*)sig);
  if (method_sig == NULL)
    {
      /* throw OutOfMemoryError */
      return NULL;
    }

  for (method_num = 0; method_num < cf->num_methods; method_num++)
    {
      MethodStruct *cur_method = cf->methods[method_num];

      if (!strcmp(cur_method->name, name)
	  && SIG_equal(env, method_sig, cur_method->sig)
	  && cur_method->access_flags & ACC_STATIC)
	{
	  method = cur_method;
	  break;
	}
    }

  SIG_free(env, method_sig);

  if (method == NULL)
    {
      jclass no_such_method = (*env)->FindClass(env, "java/lang/NoSuchMethodError");
      if (no_such_method)
	(*env)->ThrowNew(env, no_such_method, name);
      else
	(*env)->FatalError(env, "Could not load java/lang/NoSuchMethodError");
    }

  return method;
}

jboolean
JNIFUNC(CallStaticBooleanMethod)(JNIEnv *env,
				 jclass clazz,
				 jmethodID methodID,
				 ...)
{
  va_list varargs;
  jboolean result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallStaticBooleanMethodV)(env, clazz, methodID, varargs);
  va_end(varargs);

  return result;
}

jboolean
JNIFUNC(CallStaticBooleanMethodA)(JNIEnv *env,
				  jclass clazz,
				  jmethodID methodID,
				  jvalue *args)
{
  jvalue value = CallStaticJavaMethod((MethodStruct*)methodID,
				      clazz,
				      args);

  return value.z;
}

jboolean
JNIFUNC(CallStaticBooleanMethodV)(JNIEnv *env,
				  jclass clazz,
				  jmethodID methodID,
				  va_list args)
{
  jvalue *params;
  int i;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  jboolean result;
  
  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  result = JNIFUNC(CallStaticBooleanMethodA)(env, clazz, methodID, params);
  
  free(params);

  return result;
}

jbyte
JNIFUNC(CallStaticByteMethod)(JNIEnv *env,
			      jclass clazz,
			      jmethodID methodID,
			      ...)
{
  va_list varargs;
  jbyte result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallStaticByteMethodV)(env, clazz, methodID, varargs);
  va_end(varargs);

  return result;
}

jbyte
JNIFUNC(CallStaticByteMethodA)(JNIEnv *env,
			       jclass clazz,
			       jmethodID methodID,
			       jvalue *args)
{
  jvalue value = CallStaticJavaMethod((MethodStruct*)methodID,
				      clazz,
				      args);

  return value.b;
}

jbyte
JNIFUNC(CallStaticByteMethodV)(JNIEnv *env,
			       jclass clazz,
			       jmethodID methodID,
			       va_list args)
{
  jvalue *params;
  int i;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  jbyte result;
  
  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  result = JNIFUNC(CallStaticByteMethodA)(env, clazz, methodID, params);
  
  free(params);

  return result;
}

jchar
JNIFUNC(CallStaticCharMethod)(JNIEnv *env,
			      jclass clazz,
			      jmethodID methodID,
			      ...)
{
  va_list varargs;
  jchar result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallStaticCharMethodV)(env, clazz, methodID, varargs);
  va_end(varargs);

  return result;
}

jchar
JNIFUNC(CallStaticCharMethodA)(JNIEnv *env,
			       jclass clazz,
			       jmethodID methodID,
			       jvalue *args)
{
  jvalue value = CallStaticJavaMethod((MethodStruct*)methodID,
				      clazz,
				      args);

  return value.c;
}

jchar
JNIFUNC(CallStaticCharMethodV)(JNIEnv *env,
			       jclass clazz,
			       jmethodID methodID,
			       va_list args)
{
  jvalue *params;
  int i;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  jchar result;
  
  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  result = JNIFUNC(CallStaticCharMethodA)(env, clazz, methodID, params);
  
  free(params);

  return result;
}

jshort
JNIFUNC(CallStaticShortMethod)(JNIEnv *env,
			       jclass clazz,
			       jmethodID methodID,
			       ...)
{
  va_list varargs;
  jshort result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallStaticShortMethodV)(env, clazz, methodID, varargs);
  va_end(varargs);

  return result;
}

jshort
JNIFUNC(CallStaticShortMethodA)(JNIEnv *env,
				jclass clazz,
				jmethodID methodID,
				jvalue *args)
{
  jvalue value = CallStaticJavaMethod((MethodStruct*)methodID,
				      clazz,
				      args);

  return value.s;
}

jshort
JNIFUNC(CallStaticShortMethodV)(JNIEnv *env,
				jclass clazz,
				jmethodID methodID,
				va_list args)
{
  jvalue *params;
  int i;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  jshort result;
  
  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  result = JNIFUNC(CallStaticShortMethodA)(env, clazz, methodID, params);
  
  free(params);

  return result;
}

jint
JNIFUNC(CallStaticIntMethod)(JNIEnv *env,
			     jclass clazz,
			     jmethodID methodID,
			     ...)
{
  va_list varargs;
  jint result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallStaticIntMethodV)(env, clazz, methodID, varargs);
  va_end(varargs);

  return result;
}

jint
JNIFUNC(CallStaticIntMethodA)(JNIEnv *env,
			      jclass clazz,
			      jmethodID methodID,
			      jvalue *args)
{
  jvalue value = CallStaticJavaMethod((MethodStruct*)methodID,
				      clazz,
				      args);

  return value.i;
}

jint
JNIFUNC(CallStaticIntMethodV)(JNIEnv *env,
			      jclass clazz,
			      jmethodID methodID,
			      va_list args)
{
  jvalue *params;
  int i;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  jint result;
  
  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  result = JNIFUNC(CallStaticIntMethodA)(env, clazz, methodID, params);
  
  free(params);

  return result;
}

jlong
JNIFUNC(CallStaticLongMethod)(JNIEnv *env,
			      jclass clazz,
			      jmethodID methodID,
			      ...)
{
  va_list varargs;
  jlong result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallStaticLongMethodV)(env, clazz, methodID, varargs);
  va_end(varargs);

  return result;
}

jlong
JNIFUNC(CallStaticLongMethodA)(JNIEnv *env,
			       jclass clazz,
			       jmethodID methodID,
			       jvalue *args)
{
  jvalue value = CallStaticJavaMethod((MethodStruct*)methodID,
				      clazz,
				      args);

  return value.j;
}

jlong
JNIFUNC(CallStaticLongMethodV)(JNIEnv *env,
			       jclass clazz,
			       jmethodID methodID,
			       va_list args)
{
  jvalue *params;
  int num_params = SIG_numParams(env, ((MethodStruct*)methodID)->sig);
  int i;
  jlong result;
  
  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    params[i] = va_arg(args, jvalue);

  result = JNIFUNC(CallStaticLongMethodA)(env, clazz, methodID, params);
  
  free(params);

  return result;
}

jfloat
JNIFUNC(CallStaticFloatMethod)(JNIEnv *env,
			       jclass clazz,
			       jmethodID methodID,
			       ...)
{
  va_list varargs;
  jfloat result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallStaticFloatMethodV)(env, clazz, methodID, varargs);
  va_end(varargs);

  return result;
}

jfloat
JNIFUNC(CallStaticFloatMethodA)(JNIEnv *env,
				jclass clazz,
				jmethodID methodID,
				jvalue *args)
{
  jvalue value = CallStaticJavaMethod((MethodStruct*)methodID,
				      clazz,
				      args);

  return value.f;
}

jfloat
JNIFUNC(CallStaticFloatMethodV)(JNIEnv *env,
				jclass clazz,
				jmethodID methodID,
				va_list args)
{
  jvalue *params;
  int num_params = SIG_numParams(env, ((MethodStruct*)methodID)->sig);
  int i;
  jfloat result;
  
  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    params[i] = va_arg(args, jvalue);

  result = JNIFUNC(CallStaticFloatMethodA)(env, clazz, methodID, params);
  
  free(params);

  return result;
}

jdouble
JNIFUNC(CallStaticDoubleMethod)(JNIEnv *env,
				jclass clazz,
				jmethodID methodID,
				...)
{
  va_list varargs;
  jdouble result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallStaticDoubleMethodV)(env, clazz, methodID, varargs);
  va_end(varargs);

  return result;
}

jdouble
JNIFUNC(CallStaticDoubleMethodA)(JNIEnv *env,
				 jclass clazz,
				 jmethodID methodID,
				 jvalue *args)
{
  jvalue value = CallStaticJavaMethod((MethodStruct*)methodID,
				      clazz,
				      args);

  return value.d;
}

jdouble
JNIFUNC(CallStaticDoubleMethodV)(JNIEnv *env,
				 jclass clazz,
				 jmethodID methodID,
				 va_list args)
{
  jvalue *params;
  int num_params = SIG_numParams(env, ((MethodStruct*)methodID)->sig);
  int i;
  jdouble result;
  
  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    params[i] = va_arg(args, jvalue);

  result = JNIFUNC(CallStaticDoubleMethodA)(env, clazz, methodID, params);
  
  free(params);

  return result;
}

jobject
JNIFUNC(CallStaticObjectMethod)(JNIEnv *env,
				jclass clazz,
				jmethodID methodID,
				...)
{
  va_list varargs;
  jobject result;

  va_start(varargs, methodID);
  result = JNIFUNC(CallStaticObjectMethodV)(env, clazz, methodID, varargs);
  va_end(varargs);

  return result;
}

jobject
JNIFUNC(CallStaticObjectMethodA)(JNIEnv *env,
				 jclass clazz,
				 jmethodID methodID,
				 jvalue *args)
{
  jvalue value = CallStaticJavaMethod((MethodStruct*)methodID,
				      clazz,
				      args);

  return value.l;
}

jobject
JNIFUNC(CallStaticObjectMethodV)(JNIEnv *env,
				 jclass clazz,
				 jmethodID methodID,
				 va_list args)
{
  jvalue *params;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  int i;
  jobject result;

  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  result = JNIFUNC(CallStaticObjectMethodA)(env, clazz, methodID, params);
  
  free(params);

  return result;
}

void
JNIFUNC(CallStaticVoidMethod)(JNIEnv *env,
			      jclass clazz,
			      jmethodID methodID, 
			      ...)
{
  va_list varargs;

  va_start(varargs, methodID);
  JNIFUNC(CallStaticVoidMethodV)(env, clazz, methodID, varargs);
  va_end(varargs);
}

void
JNIFUNC(CallStaticVoidMethodA)(JNIEnv *env, 
			       jclass clazz,
			       jmethodID methodID, 
			       jvalue *args)
{
  CallStaticJavaMethod((MethodStruct*)methodID,
		       clazz,
		       args);
}

void
JNIFUNC(CallStaticVoidMethodV)(JNIEnv *env, 
			       jclass clazz,
			       jmethodID methodID, 
			       va_list args)
{
  jvalue *params;
  MethodStruct *method = (MethodStruct*)methodID;
  int num_params = SIG_numParams(env, method->sig);
  int i;

  params = (jvalue*)calloc(num_params, sizeof(jvalue*));

  for (i = 0; i < num_params; i ++)
    {
      if (method->sig->method.params[i]->any.tag != SIG_PRIM) /* XXX punt... */
	{
	  params[i].l = va_arg(args, void*);
	}
      else
	{
	  switch (method->sig->method.params[i]->prim.type)
	    {
	    case SIG_JBOOLEAN:
	    case SIG_JBYTE:
	    case SIG_JCHAR:
	    case SIG_JSHORT:
	    case SIG_JINT:
	    case SIG_JOBJECT:
	    case SIG_JFLOAT:
	      params[i].i = va_arg(args, jint);
	      break;
	    case SIG_JLONG:
	    case SIG_JDOUBLE:
	      params[i].j = va_arg(args, jlong);
	      break;
	    default:
	      abort();
	    }
	}
    }

  JNIFUNC(CallStaticVoidMethodA)(env, clazz, methodID, params);

  free(params);
}
