/* -*- Mode: C; c-file-style: "gnu" -*-
   jvmdi.c -- Java Virtual Machine Debugger Interface (JDK 1.2)
   Created: Chris Toshok <toshok@hungry.com>, 9-Feb-1998.
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */

#include "jvmdiint.h"
#include "interp.h"
#include "objects.h"

#include <stdlib.h>

static JThreadInfo*
thread_obj_to_jthread(JNIEnv *env,
		      obj_thunk *thread_obj)
{
  jclass thread_clazz = (*env)->GetObjectClass(env, (jobject)thread_obj);
  jfieldID privateInfo = (*env)->GetFieldID(env, thread_clazz, "PrivateInfo", "I");
  JThreadInfo *info;

  info = (JThreadInfo*)(*env)->GetIntField(env, thread_obj, privateInfo);

  return info;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_GetThreadStatus(JNIEnv *env,
		      jthread thread,
		      jint *statusPtr)
{
  obj_thunk *thread_obj = (obj_thunk*)thread;
  JThreadInfo *thread_info;

  if (env == NULL
      || thread_obj == NULL
      || statusPtr == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  thread_info = thread_obj_to_jthread(env, thread_obj);
  if (thread_info == NULL)
    return JVMDI_ERROR_INVALID_THREAD;
  
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_SuspendThread(JNIEnv *env,
		    jthread thread)
{
  obj_thunk *thread_obj = (obj_thunk*)thread;
  JThreadInfo *thread_info;

  if (env == NULL
      || thread_obj == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  thread_info = thread_obj_to_jthread(env, thread_obj);
  if (thread_info == NULL)
    return JVMDI_ERROR_INVALID_THREAD;

  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_ResumeThread(JNIEnv *env,
		   jthread thread)
{
  obj_thunk *thread_obj = (obj_thunk*)thread;
  JThreadInfo *thread_info;

  if (env == NULL
      || thread_obj == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  thread_info = thread_obj_to_jthread(env, thread_obj);
  if (thread_info == NULL)
    return JVMDI_ERROR_INVALID_THREAD;

#if notyet
  /* the thread must be suspended for this operation. */
  if (STATE(thread_info) != THREAD_SUSPENDED)
    return JVMDI_ERROR_THREAD_NOT_SUSPENDED;
#endif

  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_SetSingleStep(JNIEnv *env,
		    jthread thread,
		    jboolean shouldStep)
{

  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_GetCurrentFrame(JNIEnv *env,
		      jthread thread,
		      jframeID *framePtr)
{
  obj_thunk *thread_obj = (obj_thunk*)thread;
  JThreadInfo *thread_info;

  /* usual NULL checks */
  if (env == NULL
      || thread_obj == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  /* make sure the object refers to a thread (and a valid one at that.) */
  thread_info = thread_obj_to_jthread(env, thread_obj);
  if (thread_info == NULL)
    return JVMDI_ERROR_INVALID_THREAD;

#if notyet
  /* the thread must be suspended for this operation. */
  if (STATE(thread_info) != THREAD_SUSPENDED)
    return JVMDI_ERROR_THREAD_NOT_SUSPENDED;
#endif

  *framePtr = (jframeID)TOPFRAME(thread_info);
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_GetCallerFrame(JNIEnv *env,
		     jframeID called,
		     jframeID *framePtr)
{
  /* usual NULL checks */
  if (env == NULL
      || framePtr == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  /* sanity check for a valid frame id */
  if (called == NULL)
    return JVMDI_ERROR_INVALID_FRAMEID;

  *framePtr = (jframeID)get_frame_parent((StackFrame*)called);
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_GetFrameMethod(JNIEnv *env,
		     jframeID frame, 
		     jclass *classPtr,
		     jmethodID *methodPtr)
{
  
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_GetFrameBCI(JNIEnv *env,
		  jframeID frame,
		  jint *bciPtr)
{
  StackFrame *sframe = (StackFrame*)frame;

  /* usual NULL checks */
  if (env == NULL
      || bciPtr == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  /* sanity check for a valid frame id */
  if (frame == NULL)
    return JVMDI_ERROR_INVALID_FRAMEID;

  *bciPtr = PC(sframe);

  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_NotifyFramePop(JNIEnv *env,
		     jframeID frame)
{

  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_GetLocalObject(JNIEnv *env,
		     jframeID frame,
		     jint slot,
		     jobject *valuePtr)
{
  StackFrame *sframe = (StackFrame*)frame;

  /* usual NULL checks. */
  if (env == NULL
      || sframe == NULL
      || valuePtr == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  /* make sure the slot number is possible in this sframe. */
  if (slot < 0 || slot > sframe->method->max_locals)
    return JVMDI_ERROR_INVALID_SLOT;

  /* check the type of the slot */
  if (sframe->vars[slot].tag != SIG_JOBJECT)
    return JVMDI_ERROR_TYPE_MISMATCH;

  *valuePtr = sframe->vars[slot].value.l;
  return JVMDI_ERROR_NONE;
}


JNIEXPORT jvmdiError JNICALL
JVMDI_GetLocalInt(JNIEnv *env,
		  jframeID frame,
		  jint slot,
		  jint *valuePtr)
{
  StackFrame *sframe = (StackFrame*)frame;

  /* usual NULL checks. */
  if (env == NULL
      || sframe == NULL
      || valuePtr == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  /* make sure the slot number is possible in this sframe. */
  if (slot < 0 || slot > sframe->method->max_locals)
    return JVMDI_ERROR_INVALID_SLOT;

  /* check the type of the slot */
  if (sframe->vars[slot].tag != SIG_JINT)
    return JVMDI_ERROR_TYPE_MISMATCH;

  *valuePtr = sframe->vars[slot].value.i;
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_GetLocalLong(JNIEnv *env,
		   jframeID frame,
		   jint slot,
		   jlong *valuePtr)
{
  StackFrame *sframe = (StackFrame*)frame;

  /* usual NULL checks. */
  if (env == NULL
      || sframe == NULL
      || valuePtr == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  /* make sure the slot number is possible in this sframe. */
  if (slot < 0 || slot > sframe->method->max_locals - 1)
    return JVMDI_ERROR_INVALID_SLOT;

  /* check the type of the slot */
  if (sframe->vars[slot].tag != SIG_JLONG1
      || sframe->vars[slot + 1].tag != SIG_JLONG2)
    return JVMDI_ERROR_TYPE_MISMATCH;

  *valuePtr = (jlong)sframe->vars[slot].value.i << 32;
  *valuePtr |= sframe->vars[slot + 1].value.j & 0xffffffff;
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_GetLocalFloat(JNIEnv *env,
		    jframeID frame,
		    jint slot,
		    jfloat *valuePtr)
{
  StackFrame *sframe = (StackFrame*)frame;

  /* usual NULL checks. */
  if (env == NULL
      || sframe == NULL
      || valuePtr == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  /* make sure the slot number is possible in this sframe. */
  if (slot < 0 || slot > sframe->method->max_locals)
    return JVMDI_ERROR_INVALID_SLOT;

  /* check the type of the slot */
  if (sframe->vars[slot].tag != SIG_JFLOAT)
    return JVMDI_ERROR_TYPE_MISMATCH;

  *valuePtr = sframe->vars[slot].value.f;
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_GetLocalDouble(JNIEnv *env,
		     jframeID frame,
		     jint slot,
		     jdouble *valuePtr)
{
  StackFrame *sframe = (StackFrame*)frame;
  jvalue value;

  /* usual NULL checks. */
  if (env == NULL
      || sframe == NULL
      || valuePtr == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  /* make sure the slot number is possible in this sframe. */
  if (slot < 0 || slot > sframe->method->max_locals - 1)
    return JVMDI_ERROR_INVALID_SLOT;

  /* check the type of the slot */
  if (sframe->vars[slot].tag != SIG_JDOUBLE1
      || sframe->vars[slot + 1].tag != SIG_JDOUBLE2)
    return JVMDI_ERROR_TYPE_MISMATCH;

  value.j = (jlong)sframe->vars[slot].value.i << 32;
  value.j |= sframe->vars[slot + 1].value.j & 0xffffffff;

  *valuePtr = value.d;
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_SetLocalObject(JNIEnv *env,
		     jframeID frame,
		     jint slot,
		     jobject value)
{
  StackFrame *sframe = (StackFrame*)frame;

  /* usual NULL checks. */
  if (env == NULL
      || sframe == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  /* make sure the slot number is possible in this sframe. */
  if (slot < 0 || slot > sframe->method->max_locals)
    return JVMDI_ERROR_INVALID_SLOT;

  /* check the type of the slot */
  if (sframe->vars[slot].tag != SIG_JOBJECT)
    return JVMDI_ERROR_TYPE_MISMATCH;

  sframe->vars[slot].value.l = value;
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_SetLocalInt(JNIEnv *env,
		  jframeID frame,
		  jint slot,
		  jint value)
{
  StackFrame *sframe = (StackFrame*)frame;

  /* usual NULL checks. */
  if (env == NULL
      || sframe == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  /* make sure the slot number is possible in this sframe. */
  if (slot < 0 || slot > sframe->method->max_locals)
    return JVMDI_ERROR_INVALID_SLOT;

  /* check the type of the slot */
  if (sframe->vars[slot].tag != SIG_JINT)
    return JVMDI_ERROR_TYPE_MISMATCH;

  sframe->vars[slot].value.i = value;
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_SetLocalFloat(JNIEnv *env,
		    jframeID frame,
		    jint slot,
		    jfloat value)
{
  StackFrame *sframe = (StackFrame*)frame;

  /* usual NULL checks. */
  if (env == NULL
      || sframe == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  /* make sure the slot number is possible in this sframe. */
  if (slot < 0 || slot > sframe->method->max_locals)
    return JVMDI_ERROR_INVALID_SLOT;

  /* check the type of the slot */
  if (sframe->vars[slot].tag != SIG_JFLOAT)
    return JVMDI_ERROR_TYPE_MISMATCH;

  sframe->vars[slot].value.f = value;
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_SetLocalLong(JNIEnv *env,
		   jframeID frame,
		   jint slot,
		   jlong value)
{
  StackFrame *sframe = (StackFrame*)frame;

  /* usual NULL checks. */
  if (env == NULL
      || sframe == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  /* make sure the slot number is possible in this sframe. */
  if (slot < 0 || slot > sframe->method->max_locals - 1)
    return JVMDI_ERROR_INVALID_SLOT;

  /* check the type of the slot */
  if (sframe->vars[slot].tag != SIG_JLONG1
      || sframe->vars[slot + 1].tag != SIG_JLONG2)
    return JVMDI_ERROR_TYPE_MISMATCH;

  sframe->vars[slot].value.i = (value >> 32) & 0xffffffff;
  sframe->vars[slot + 1].value.i = value & 0xffffffff;
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_SetLocalDouble(JNIEnv *env,
		     jframeID frame,
		     jint slot,
		     jdouble value)
{
  StackFrame *sframe = (StackFrame*)frame;
  jvalue tmp_value;

  /* usual NULL checks. */
  if (env == NULL
      || sframe == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  /* make sure the slot number is possible in this sframe. */
  if (slot < 0 || slot > sframe->method->max_locals - 1)
    return JVMDI_ERROR_INVALID_SLOT;

  /* check the type of the slot */
  if (sframe->vars[slot].tag != SIG_JLONG1
      || sframe->vars[slot + 1].tag != SIG_JLONG2)
    return JVMDI_ERROR_TYPE_MISMATCH;

  tmp_value.d = value;

  sframe->vars[slot].value.i = (tmp_value.j >> 32) & 0xffffffff;
  sframe->vars[slot + 1].value.i = tmp_value.j & 0xffffffff;
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_SetBreakpoint(JNIEnv *env,
		    jclass clazz,
		    jmethodID method,
		    jint bci)
{

  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_ClearBreakpoint(JNIEnv *env,
		      jclass clazz,
		      jmethodID method,
		      jint bci)
{

  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_ClearAllBreakpoints(JNIEnv *env)
{

  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_GetMethodName(JNIEnv *env,
		    jclass clazz,
		    jmethodID method, 
		    jstring *namePtr,
		    jstring *signaturePtr)
{
  jstring name = NULL, signature = NULL;
  char *java_sig;

  /* the usual NULL check. */
  if (env == NULL
      || namePtr == NULL
      || signaturePtr == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  if (clazz == NULL)
    return JVMDI_ERROR_INVALID_CLASS;

  if (method == NULL)
    return JVMDI_ERROR_INVALID_METHODID;

  name = (*env)->NewStringUTF(env, method->name);
  java_sig = SIG_formatToJavaSig(env, method->sig);
  if (java_sig)
    {
      signature = (*env)->NewStringUTF(env, java_sig);
      free(java_sig);
    }

  if (name == NULL
      || signature == NULL)
    {
      /* how do we free things?  it doesn't make sense to have the
	 same thread local objects for jvmdi methods like the JNI spec
	 does for native methods, does it?  I suppose, since these
	 objects were not changed into global refs, then they should
	 be collected the next time around, so we shouldn't worry. */
      return JVMDI_ERROR_OUT_OF_MEMORY;
    }

  *namePtr = name;
  *signaturePtr = signature;
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_GetMethodDefiningClass(JNIEnv *env,
			     jclass clazz,
			     jmethodID method,
			     jclass *definingClassPtr)
{

  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_IsMethodNative(JNIEnv *env,
		     jclass clazz,
		     jmethodID method,
		     jboolean *isNativePtr)
{
  /* the usual NULL checks. */
  if (env == NULL
      || isNativePtr == NULL)
    return JVMDI_ERROR_NULL_POINTER;

  /* check for invalid class. */
  if (clazz == NULL)
    return JVMDI_ERROR_INVALID_CLASS;

  /* check for invalid method id */
  if (method == NULL)
    return JVMDI_ERROR_INVALID_METHODID;

  *isNativePtr = (method->access_flags & ACC_NATIVE) != 0;
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_SetEventHook(JNIEnv *env,
		   JVMDI_EventHook hook)
{

  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_GetLoadedClasses(JNIEnv *env,
		       jobjectArray *classesPtr)
{
  return JVMDI_ERROR_NONE;
}

JNIEXPORT jvmdiError JNICALL
JVMDI_GetVersionNumber(JNIEnv *env,
		       jint *versionPtr)
{
  *versionPtr = JVMDI_VERSION_NUMBER;
  return JVMDI_ERROR_NONE;
}

static JVMDI_EventHook jvmdi_hook;

JNIEXPORT void JNICALL
jvmdi_SingleStepEvent(JNIEnv *env,
		      jthread thread,
		      jclass clazz,
		      jmethodID method,
		      jint bci)
{
  if (jvmdi_hook)
    {
      JVMDI_Event event;
      event.kind = JVMDI_EVENT_SINGLE_STEP;
      event.u.single_step.thread = thread;
      event.u.single_step.clazz = clazz;
      event.u.single_step.method = method;
      event.u.single_step.bci = bci;

      (*jvmdi_hook)(env, &event);
    }
}

JNIEXPORT void JNICALL
jvmdi_BreakpointEvent(JNIEnv *env,
		      jthread thread,
		      jclass clazz,
		      jmethodID method,
		      jint bci)
{
  if (jvmdi_hook)
    {
      JVMDI_Event event;
      event.kind = JVMDI_EVENT_BREAKPOINT;
      event.u.breakpoint.thread = thread;
      event.u.breakpoint.clazz = clazz;
      event.u.breakpoint.method = method;
      event.u.breakpoint.bci = bci;

      (*jvmdi_hook)(env, &event);
    }
}

JNIEXPORT void JNICALL
jvmdi_FramePopEvent(JNIEnv *env,
		    jthread thread,
		    jframeID frame)
{
  if (jvmdi_hook)
    {
      JVMDI_Event event;
      event.kind = JVMDI_EVENT_FRAME_POP;
      event.u.frame_pop.thread = thread;
      event.u.frame_pop.frame = frame;

      (*jvmdi_hook)(env, &event);
    }
}

JNIEXPORT void JNICALL
jvmdi_ExceptionEvent(JNIEnv *env,
		     jthread thread,
		     jclass clazz,
		     jmethodID method,
		     jint bci,
		     jobject exception,
		     jclass catch_clazz,
		     jmethodID catch_method,
		     jint catch_bci)
{
  if (jvmdi_hook)
    {
      JVMDI_Event event;
      event.kind = JVMDI_EVENT_EXCEPTION;
      event.u.exception.thread = thread;
      event.u.exception.clazz = clazz;
      event.u.exception.method = method;
      event.u.exception.bci = bci;
      event.u.exception.exception = exception;
      event.u.exception.catch_clazz = catch_clazz;
      event.u.exception.catch_method = catch_method;
      event.u.exception.catch_bci = catch_bci;

      (*jvmdi_hook)(env, &event);
    }
}

JNIEXPORT void JNICALL
jvmdi_UserEvent(JNIEnv *env,
		jobject object,
		jint key)
{
  if (jvmdi_hook)
    {
      JVMDI_Event event;
      event.kind = JVMDI_EVENT_USER_DEFINED;
      event.u.user.object = object;
      event.u.user.key = key;

      (*jvmdi_hook)(env, &event);
    }
}

JNIEXPORT void JNICALL
jvmdi_ThreadEndEvent(JNIEnv *env,
		     jthread thread)
{
  if (jvmdi_hook)
    {
      JVMDI_Event event;
      event.kind = JVMDI_EVENT_THREAD_END;
      event.u.thread_change.thread = thread;

      (*jvmdi_hook)(env, &event);
    }
}

JNIEXPORT void JNICALL
jvmdi_ThreadStartEvent(JNIEnv *env,
		       jthread thread)
{
  if (jvmdi_hook)
    {
      JVMDI_Event event;
      event.kind = JVMDI_EVENT_THREAD_START;
      event.u.thread_change.thread = thread;

      (*jvmdi_hook)(env, &event);
    }
}
