/* -*- Mode: C; c-file-style: "gnu" -*-
   class.c -- native methods for java/lang/Class.
   Created: Chris Toshok <toshok@hungry.com>, 28-Jul-1997
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "jni.h"
#include "interp.h"
#include "objects.h"
#include "sig.h"
#include "primitive-class.h"
#include <string.h>
#include <stdlib.h>

static ClazzFile *
get_classfile(JNIEnv *env, jobject obj)
{
  ClazzFile *class_cf;
  obj_thunk *o;

  class_cf = (*env)->FindClass(env, "java/lang/Class");
  o = find_instance_data_for_class(obj, class_cf);

  return (ClazzFile*)(INSTANCE_DATA(o)->fields[0].value.l);
}

JNIEXPORT jobject JNICALL
Java_java_lang_Class_forName(JNIEnv *env,
			     jclass cls,
			     jstring class_name)
{
  jobject clazz_loaded;
  char *cls_name;
  int cls_name_length;
  const jbyte *cls_name_bytes;

  cls_name_bytes = (*env)->GetStringUTFChars(env, class_name, NULL);

  cls_name_length = (*env)->GetStringUTFLength(env, class_name) + 1;
  cls_name = (char*)malloc(cls_name_length * sizeof(char));

  strncpy(cls_name, (char*)cls_name_bytes, cls_name_length);

  (*env)->ReleaseStringUTFChars(env, class_name, cls_name_bytes);

  slashes_to_dots(cls_name);

  clazz_loaded = (jobject)(*env)->FindClass(env, cls_name);

  free(cls_name);

  if ((*env)->ExceptionOccurred(env))
    return NULL;
  else
    return new_classobject(env, clazz_loaded);
}

JNIEXPORT jobject JNICALL
Java_java_lang_Class_newInstance(JNIEnv *env,
				 jobject obj)
{
  ClazzFile *clazz = get_classfile(env, obj);
  jmethodID constructor = (*env)->GetMethodID(env, clazz, "<init>", "()V");

  if (clazz->access_flags & ACC_INTERFACE
      || clazz->access_flags & ACC_ABSTRACT
      || !constructor)
    {
      char buf[1024];
      jclass exception = (*env)->FindClass(env, "java/lang/InstantiationException");
      
      snprintf(buf, 1024, "in java.lang.Class.newInstance, for class %s", getClassName(env, clazz));

      (*env)->ThrowNew(env, exception, buf);

      return NULL;
    }
  
  /* probably need more checks here... */
  if (constructor->access_flags & ACC_PRIVATE)
    {
      char buf[1024];
      jclass exception = (*env)->FindClass(env, "java/lang/IllegalAccessException");
      
      snprintf(buf, 1024, "in java.lang.Class.newInstance, inaccessible constructor for class %s", getClassName(env, clazz));

      (*env)->ThrowNew(env, exception, buf);

      return NULL;
    }

  return (*env)->NewObject(env, clazz, constructor, NULL);
}

/* is this right? */
JNIEXPORT jboolean JNICALL
Java_java_lang_Class_isInstance(JNIEnv *env,
				jobject cls,
				jobject obj)
{
  ClazzFile *clazz = get_classfile(env, obj);

  return (*env)->IsInstanceOf(env, obj, clazz);
}

JNIEXPORT jboolean JNICALL
Java_java_lang_Class_isAssignableFrom(JNIEnv *env,
				      jobject clazz,
				      jobject clazz2)
{
  ClazzFile *cls1 = (ClazzFile*)(((obj_thunk*)clazz)->_instance->fields[0].value.l);
  ClazzFile *cls2 = (ClazzFile*)(((obj_thunk*)clazz2)->_instance->fields[0].value.l);
  return (*env)->IsAssignableFrom(env, cls1, cls2);
}

JNIEXPORT jboolean JNICALL
Java_java_lang_Class_isInterface(JNIEnv *env,
				 jobject obj)
{
  ClazzFile *clazz = get_classfile(env, obj);

  return (clazz->access_flags & ACC_INTERFACE) != 0;
}

JNIEXPORT jboolean JNICALL
Java_java_lang_Class_isArray(JNIEnv *env,
			     jobject obj)
{
  ClazzFile *clazz = get_classfile(env, obj);

  return (clazz->access_flags & ACC_ARRAY) != 0;
}

JNIEXPORT jboolean JNICALL
Java_java_lang_Class_isPrimitive(JNIEnv *env,
				 jobject obj)
{
  ClazzFile *clazz;

  if (obj == NULL) return JNI_FALSE;

  clazz = get_classfile(env, obj);

  return (clazz->access_flags & ACC_PRIMITIVE) != 0;
}

JNIEXPORT jstring JNICALL
Java_java_lang_Class_getName(JNIEnv *env,
			     jobject obj)
{
  ClazzFile *cf = get_classfile(env, obj);
  char *cls_name;
  jstring str;

  if (obj == 0) return 0;

  cls_name = slashes_to_dots(strdup(getClassName(env, cf)));

  str = (*env)->NewStringUTF(env, cls_name);
  str = (*env)->NewGlobalRef(env, str);

  free(cls_name);

  return str;
}

JNIEXPORT jobject JNICALL
Java_java_lang_Class_getClassLoader(JNIEnv *env,
				    jobject obj)
{
  /*  (*env)->FatalError(env, "java.lang.Class.getClassLoader unimplemented");*/
  return NULL;
}

JNIEXPORT jobject JNICALL
Java_java_lang_Class_getSuperclass(JNIEnv *env,
				   jobject obj)
{
  ClazzFile *clazz = get_classfile(env, obj);
  ClazzFile *super;

  super = getSuperClass(env, clazz);

  if (super)
    return (jobject)new_classobject(env, super);
  else
    return NULL;
}

JNIEXPORT jobjectArray JNICALL
Java_java_lang_Class_getInterfaces(JNIEnv *env,
				   jobject obj)
{
  ClazzFile *clazz_cf = (ClazzFile*)(*env)->GetObjectClass(env, obj);
  jclass array_class = (*env)->FindClass(env, "[Ljava/lang/Class;");
  jobjectArray array;
  int array_length = 0;
  int array_index;
  ClazzFile *cur_class;

  cur_class = clazz_cf;
  while (cur_class)
    {
      array_length += cur_class->num_interfaces;
      cur_class = getSuperClass(env, cur_class);
    }

  array = (*env)->NewObjectArray(env, array_length,
				 array_class, NULL);

  cur_class = clazz_cf;
  array_index = 0;
  while (cur_class)
    {
      int i;

      for (i = 0; i < clazz_cf->num_interfaces; i++)
	{
	  ClazzFile *interface_class = getInterface(env, clazz_cf, i);
	  jobject interface_obj = new_classobject(env, interface_class);
		  
	  (*env)->SetObjectArrayElement(env, array,
					array_index,
					interface_obj);
	  array_index ++;
	}
    }

  array = (*env)->NewGlobalRef(env, array);

  return array;
}

JNIEXPORT jobject JNICALL
Java_java_lang_Class_getComponentType(JNIEnv *env,
				      jobject obj)
{
  ClazzFile *clazz = get_classfile(env, obj);
  jstring str;

  if ((clazz->access_flags & ACC_ARRAY) == 0)
    return NULL;

  str = (*env)->NewStringUTF(env, getClassName(env, clazz) + 1 /* to skip past the "[" */);
  str = (*env)->NewGlobalRef(env, str);

  return str;
}

JNIEXPORT jlong JNICALL
Java_java_lang_Class_getModifiers(JNIEnv *env,
				  jobject obj)
{
  ClazzFile *clazz = get_classfile(env, obj);

  return clazz->access_flags;
}

JNIEXPORT jobjectArray JNICALL
Java_java_lang_Class_getSigners(JNIEnv *env,
				jobject obj)
{
  (*env)->FatalError(env, "java.lang.Class.getSigners unimplemented");
  return NULL;
}

JNIEXPORT void JNICALL
Java_java_lang_Class_setSigners(JNIEnv *env,
				jobject obj,
				jobjectArray signers)
{
  (*env)->FatalError(env, "java.lang.Class.setSigners unimplemented");
}

JNIEXPORT jobject JNICALL
Java_java_lang_Class_getPrimitiveClass(JNIEnv *env,
				       jclass cls,
				       jobject obj)
{
  const jbyte *bytes = (jbyte*)(*env)->GetStringUTFChars(env, obj, NULL);
  jobject result;
  jclass clazz;

  clazz = createFakePrimitiveClass(env, (char*)bytes);

  result = (jobject)new_classobject(env, clazz);

  (*env)->ReleaseStringUTFChars(env, obj, bytes);

  return result;
}

static jclass
sig_to_jclass(JNIEnv *env, Signature *sig)
{
  jclass sig_cls = NULL;

  if (sig->any.tag == SIG_CLASS
      || sig->any.tag == SIG_ARRAY)
    {
      char *java_sig = SIG_formatToJavaSig(env, sig);
	  
      sig_cls = (*env)->FindClass(env, java_sig);
	  
      free(java_sig);
    }
  else /* SIG_PRIM */
    {
      switch (sig->prim.type)
	{
	case SIG_JBOOLEAN:
	  sig_cls = (*env)->FindClass(env, "java/lang/Boolean");
	  break;
	case SIG_JBYTE:
	  sig_cls = (*env)->FindClass(env, "java/lang/Byte");
	  break;
	case SIG_JCHAR:
	  sig_cls = (*env)->FindClass(env, "java/lang/Character");
	  break;
	case SIG_JSHORT:
	  sig_cls = (*env)->FindClass(env, "java/lang/Short");
	  break;
	case SIG_JINT:
	  sig_cls = (*env)->FindClass(env, "java/lang/Integer");
	  break;
	case SIG_JLONG:
	  sig_cls = (*env)->FindClass(env, "java/lang/Long");
	  break;
	case SIG_JFLOAT:
	  sig_cls = (*env)->FindClass(env, "java/lang/Float");
	  break;
	case SIG_JDOUBLE:
	  sig_cls = (*env)->FindClass(env, "java/lang/Double");
	  break;
	case SIG_JVOID:
	  sig_cls = (*env)->FindClass(env, "java/lang/Void");
	  break;
	default:
	  (*env)->FatalError(env, "Illegal primitive type.");
	  break;
	}
    }

  return sig_cls;
}

JNIEXPORT jobjectArray JNICALL
Java_java_lang_Class_getFields0(JNIEnv *env,
				jobject obj,
				jint which)
{
  ClazzFile *clazz = get_classfile(env, obj);
  jclass field_class = (*env)->FindClass(env, "java/lang/reflect/Field");
  jmethodID field_ctor;
  jobjectArray array;
  jfieldID slot_field = (*env)->GetFieldID(env, field_class, "slot", "I");
  jfieldID type_field = (*env)->GetFieldID(env, field_class, "type", "Ljava/lang/Class;");
  jfieldID name_field = (*env)->GetFieldID(env, field_class, "name", "Ljava/lang/String;");
  jfieldID clazz_field = (*env)->GetFieldID(env, field_class, "clazz", "Ljava/lang/Class;");
  int i;

  array = (*env)->NewObjectArray(env, clazz->num_fields,
				 field_class, NULL);

  field_ctor = (*env)->GetMethodID(env, field_class, "<init>", "()V");

  for (i = 0; i < clazz->num_fields; i ++)
    {
      jobject field = (*env)->NewObject(env, field_class, field_ctor);
      jstring name = (*env)->NewStringUTF(env, clazz->fields[i]->name);
      jobject type;

      type = new_classobject(env, sig_to_jclass(env, clazz->fields[i]->sig));

      name = (*env)->NewGlobalRef(env, name);
      (*env)->SetObjectField(env, field, name_field, name);
      (*env)->SetIntField(env, field, slot_field, i);
      (*env)->SetObjectField(env, field, clazz_field, new_classobject(env, clazz));
      (*env)->SetObjectField(env, field, type_field, type);

      field = (*env)->NewGlobalRef(env, field);

      (*env)->SetObjectArrayElement(env, array, i, field);
    }

  array = (*env)->NewGlobalRef(env, array);

  return array;
}

JNIEXPORT jobjectArray JNICALL
Java_java_lang_Class_getMethods0(JNIEnv *env,
				 jobject obj,
				 jint which)
{
  ClazzFile *clazz = get_classfile(env, obj);
  jclass class_class = (*env)->FindClass(env, "java/lang/Class");
  jclass method_class = (*env)->FindClass(env, "java/lang/reflect/Method");
  jmethodID method_ctor;
  jobjectArray array;
  jfieldID slot_field = (*env)->GetFieldID(env, method_class, "slot", "I");
  jfieldID name_field = (*env)->GetFieldID(env, method_class, "name", "Ljava/lang/String;");
  jfieldID clazz_field = (*env)->GetFieldID(env, method_class, "clazz", "Ljava/lang/Class;");
  jfieldID returntype_field = (*env)->GetFieldID(env, method_class, "returnType", "Ljava/lang/Class;");
  jfieldID parameters_field = (*env)->GetFieldID(env, method_class, "parameterTypes", "[Ljava/lang/Class;");
  jfieldID exceptions_field = (*env)->GetFieldID(env, method_class, "exceptionTypes", "[Ljava/lang/Class;");
  int i;

  array = (*env)->NewObjectArray(env, clazz->num_methods,
				 method_class, NULL);

  method_ctor = (*env)->GetMethodID(env, method_class, "<init>", "()V");

  for (i = 0; i < clazz->num_methods; i ++)
    {
      jobject method = (*env)->NewObject(env, method_class, method_ctor);
      jstring name = (*env)->NewStringUTF(env, clazz->methods[i]->name);
      jclass returntype = sig_to_jclass(env, clazz->methods[i]->sig->method.return_type);
      int num_params = clazz->methods[i]->sig->method.num_params;
      jobjectArray exception_array;
      jobjectArray param_array;
      int p;

      /* XXX we currently don't know what a method throws. wow. */
      exception_array = (*env)->NewObjectArray(env, 0,
					       class_class, NULL);
      exception_array = (*env)->NewGlobalRef(env, exception_array);

      param_array = (*env)->NewObjectArray(env, num_params,
					   class_class, NULL);
      for (p = 0; p < num_params; p ++)
	{
	  jobject param;
	  jclass param_type;

	  param_type = sig_to_jclass(env, clazz->methods[i]->sig->method.params[p]);

	  param = new_classobject(env, param_type);

	  (*env)->SetObjectArrayElement(env, param_array, p, param);
	}
      param_array = (*env)->NewGlobalRef(env, param_array);

      name = (*env)->NewGlobalRef(env, name);
      (*env)->SetObjectField(env, method, name_field, name);
      (*env)->SetIntField(env, method, slot_field, i);
      (*env)->SetObjectField(env, method, clazz_field, new_classobject(env, clazz));
      (*env)->SetObjectField(env, method, returntype_field, new_classobject(env, returntype));
      (*env)->SetObjectField(env, method, exceptions_field, exception_array);
      (*env)->SetObjectField(env, method, parameters_field, param_array);

      method = (*env)->NewGlobalRef(env, method);

      (*env)->SetObjectArrayElement(env, array, i, method);
    }

  array = (*env)->NewGlobalRef(env, array);

  return array;
}

JNIEXPORT jobjectArray JNICALL
Java_java_lang_Class_getConstructors0(JNIEnv *env,
				      jobject obj,
				      jint which)
{
  (*env)->FatalError(env, "java.lang.Class.getConstructors0 unimplemented");
  return NULL;
}

JNIEXPORT jobject JNICALL
Java_java_lang_Class_getField0(JNIEnv *env,
			       jobject cls,
			       jstring name,
			       jint which)
{
  
  (*env)->FatalError(env, "java.lang.Class.getField0 unimplemented");
  return NULL;
}

JNIEXPORT jobject JNICALL
Java_java_lang_Class_getMethod0(JNIEnv *env,
				jobject obj,
				jstring name,
				jobjectArray parameterTypes,
				jint which)
{
  (*env)->FatalError(env, "java.lang.Class.getMethod0 unimplemented");
  return NULL;
}

JNIEXPORT jobject JNICALL
Java_java_lang_Class_getConstructor0(JNIEnv *env,
				     jobject obj,
				     jobjectArray parameterTypes,
				     jint which)
{
  (*env)->FatalError(env, "java.lang.Class.getConstructor0 unimplemented");
  return NULL;
}

