/* -*- Mode: C; c-file-style: "gnu" -*-
   classloader.c -- native methods for java/lang/ClassLoader.
   Created: Chris Toshok <toshok@hungry.com>, 25-Aug-1997.
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "jniint.h"
#include "objects.h"
#include "classpath.h"
#include "log.h"

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>

#define MYLOG "NativeClassLoader"

JNIEXPORT void JNICALL
Java_java_lang_ClassLoader_init(JNIEnv *env,
				jobject obj)
{
  /* anything to do here? */
}

JNIEXPORT jobject JNICALL
Java_java_lang_ClassLoader_defineClass0(JNIEnv *env,
					jobject obj,
					jobject name_str,
					jbyteArray data,
					jint offset,
					jint length)
{
  jclass cls;
  jbyte *elements;

  elements = (*env)->GetByteArrayElements(env, data, NULL);

  cls = (*env)->DefineClass(env, obj, elements + offset, length);

  (*env)->ReleaseByteArrayElements(env, data, elements, 0);

  return (jobject)new_classobject(env, cls);
}

JNIEXPORT void JNICALL
Java_java_lang_ClassLoader_resolveClass0(JNIEnv *env,
					 jobject obj,
					 jobject cls)
{
  /* nothing to do here, since we resolve classes when we define them (I think...). */
}

JNIEXPORT jobject JNICALL
Java_java_lang_ClassLoader_findSystemClass0(JNIEnv *env,
					    jobject obj,
					    jobject name_str)
{
  jclass cls;
  const jbyte *name_bytes = (*env)->GetStringUTFChars(env, name_str, NULL);

  cls = (*env)->FindClass(env, (char*)name_bytes);

  if (cls == NULL)
    {
      jclass class_not_found;

      (*env)->ExceptionClear(env); /* eat the previous exception */

      class_not_found = (*env)->FindClass(env, 
					  "java/lang/ClassNotFoundException");
      (*env)->ThrowNew(env, class_not_found, name_bytes);

      (*env)->ReleaseStringUTFChars(env, name_str, name_bytes);

      return NULL;
    }
  else
    {
      (*env)->ReleaseStringUTFChars(env, name_str, name_bytes);

      return (jobject)new_classobject(env, cls);
    }
}

JNIEXPORT jobject JNICALL
Java_java_lang_ClassLoader_getSystemResourceAsStream0(JNIEnv *env,
						      jclass cls,
						      jobject name_str)
{
  jmethodID getSystemResource;
  jobject url;
  jclass url_class = (*env)->FindClass(env, "java/net/URL");

  assert(NULL != url_class);
  getSystemResource = (*env)->GetStaticMethodID(env, cls, "getSystemResource",
						"(Ljava/lang/String;)Ljava/net/URL;");
  assert(NULL != getSystemResource);
  url = (*env)->CallStaticObjectMethod(env, cls, getSystemResource,
				       name_str);
  
  if (url == NULL)
    {
      return NULL;
    }
  else
    {
      jobject stream;
      jmethodID openStream;

      openStream = (*env)->GetMethodID(env, url_class, "openStream",
				       "()Ljava/io/InputStream;");

      stream = (*env)->CallObjectMethod(env, url, openStream);
	  
      stream = (*env)->NewGlobalRef(env, stream);

      return stream;
    }
}

char *
locate_resource_on_classpath(const char *resource_name,
			     ClasspathEntry *entries, int num_entries)
{
  char *full_resource_name;
  int i;

  for (i = 0; i < num_entries; i ++)
    {
      if (entries[i].type == CP_ZIP ||
	  entries[i].type == CP_JAR)
	{
	  ZipDirectory *dir;
	  int j;

	  /* if there was an error reading the archive, pass over it. */
	  if (entries[i].zip.fd == -1)
	    continue;

	  dir = (ZipDirectory*)entries[i].zip.central_directory;

	  for (j = 0; j < entries[i].zip.count; j ++, dir = ZIPDIR_NEXT(dir))
	    {
	      if (!strcmp(resource_name, ZIPDIR_FILENAME(dir)))
		{
		  full_resource_name = (char*)malloc(strlen("systemresource:/ZIP/")
						     + strlen(resource_name) 
						     + strlen("/+/")
						     + strlen(entries[i].path) + 2);
				  
		  sprintf (full_resource_name, "systemresource:/ZIP%d/+/%s",
			   i,
			   resource_name);
				  
		  return full_resource_name;
		}
	    }
	}
      else if (entries[i].type == CP_DIR)
	{
	  struct stat stat_buf;

	  full_resource_name = (char*)malloc(strlen("systemresource:/FILE/")
					     + strlen(resource_name) 
					     + strlen(entries[i].path) + 2);

	  sprintf (full_resource_name, "systemresource:/FILE/%s/%s",
		   entries[i].path,
		   resource_name);

	  if (stat(full_resource_name + strlen("systemresource:/FILE"), &stat_buf) == -1)
	    {
	      free(full_resource_name);
	      full_resource_name = NULL;
	      continue;
	    }
	  else
	    return full_resource_name;
	}
    }

  return NULL;
}

JNIEXPORT jstring JNICALL
Java_java_lang_ClassLoader_getSystemResourceAsName0(JNIEnv *env,
						    jclass cls,
						    jobject name_str)
{
  const jbyte *name_bytes = (*env)->GetStringUTFChars(env, name_str, NULL);
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  HungryJavaVM *hvm = henv->_vm;
  ClasspathEntry *entries = hvm->_cp_entries;
  int num_entries = hvm->_num_cp_entries;
  char *resource_location;
  jstring resource_str;

  JAVARLOG1(MYLOG, 2, "java.lang.ClassLoader.getSystemResourceAsName0(%s)\n", name_bytes);

  resource_location = locate_resource_on_classpath((char*)name_bytes,
						   entries, num_entries);

  (*env)->ReleaseStringUTFChars(env, name_str, name_bytes);

  if (resource_location)
    {
      JAVARLOG1 (MYLOG, 2, "Found %s\n", resource_location);
      resource_str = (*env)->NewStringUTF(env, resource_location);
    }
  else
    resource_str = (*env)->NewStringUTF(env, "");

  resource_str = (*env)->NewGlobalRef(env, resource_str);

  return resource_str;
}
