/* -*- Mode: C; c-file-style: "gnu" -*-
   socket.c -- native methods for the plain socket implementation.
   Created: Chris Toshok <toshok@hungry.com>, 2-Aug-97
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "jni.h"
#include "exceptions.h"
#include "log.h"

#include <netinet/in.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/fcntl.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>

#define MYLOG "Native"

#ifndef HAVE_SYS_ERRLIST
#define errorstr(n) ""
#else
#define errorstr(n) sys_errlist[n]
#endif

const char *const JAVA_NET_BINDEXCEPTION = "java/net/BindException";
const char *const JAVA_NET_SOCKETEXCEPTION = "java/net/SocketException";
const char *const JAVA_IO_IOEXCEPTION = "java/io/IOException";

/* Debug code */
static void
log_sockname(int sock_fd)
{
  struct sockaddr_in sock_addr;
  int addr_len = sizeof(sock_addr);
  int retval;
  memset(&sock_addr, 0, sizeof(sock_addr));
  retval = getsockname(sock_fd, &sock_addr, &addr_len);

  JAVARLOG5(MYLOG, 2, "%d = getsockname(%d, addr{addr=0x%08x port=0x%04x}, %d)\n",
	    retval, sock_fd, sock_addr.sin_addr.s_addr,
	    sock_addr.sin_port, sizeof(sock_addr));
}

/* Debug code */
static void
log_peername(int sock_fd)
{
  struct sockaddr_in sock_addr;
  int addr_len = sizeof(sock_addr);
  int retval;
  memset(&sock_addr, 0, sizeof(sock_addr));
  retval = getpeername(sock_fd, &sock_addr, &addr_len);

  JAVARLOG5(MYLOG, 2, "%d = getpeername(%d, addr{addr=0x%08x port=0x%04x}, %d)\n",
	    retval, sock_fd, sock_addr.sin_addr.s_addr,
	    sock_addr.sin_port, sizeof(sock_addr));
}

/*
 * Return classname.fd.fd when classname.fd is java.io.FileDescriptor
 */
static jint
get_object_fd(JNIEnv *env, jobject obj, char *classname)
{
  jobject file_desc;
  jfieldID field;
  jclass sock_class, fd_class;

  /* first we need to get the FileDescriptor object. */
  sock_class = (*env)->FindClass(env, classname);
  field = (*env)->GetFieldID(env, sock_class, "fd", "Ljava/io/FileDescriptor;");
  file_desc = (*env)->GetObjectField(env, obj, field);

  /* now we get the fd field of *that* object, and return it. */
  fd_class = (*env)->FindClass(env, "java/io/FileDescriptor");
  field = (*env)->GetFieldID(env, fd_class, "fd", "I");

  return (*env)->GetIntField(env, file_desc, field);
}

/*
 * Return the value of classname.fieldname when it is integer
 */
static jint
get_object_int(JNIEnv *env, jobject obj,
	       const char *classname, const char *fieldname)
{
  jfieldID field;
  jclass sock_class;

  sock_class = (*env)->FindClass(env, classname);
  assert(NULL != sock_class);

  field = (*env)->GetFieldID(env, sock_class, fieldname, "I");
  assert(NULL != field);

  return (*env)->GetIntField(env, obj, field);
}

/*
 * Store value in classname.fieldname
 */
static void
set_object_int(JNIEnv *env, jobject obj,
	       const char *classname, const char *fieldname, const jint value)
{
  jfieldID field;
  jclass obj_class;

  obj_class = (*env)->FindClass(env, classname);
  assert(NULL != obj_class);
  field = (*env)->GetFieldID(env, obj_class, fieldname, "I");
  assert(NULL != field);
  (*env)->SetIntField(env, obj, field, value);
}

/*
 * Return the object(classname).address object
 * If it doesn't exist, create it and store it in object first.
 */
static jobject
get_object_inetaddress(JNIEnv *env, jobject object, const char *classname)
{
  jobject addr_obj;
  jclass packet_class = (*env)->FindClass(env, classname);
  jfieldID address_obj_field = (*env)->GetFieldID(env, packet_class, "address",
						  "Ljava/net/InetAddress;");
  assert(NULL != address_obj_field);
  assert(NULL != object);
  addr_obj = (*env)->GetObjectField(env, object, address_obj_field);

  if (NULL == addr_obj)
    { /* address == null, create one */
      jclass inetaddress_class = (*env)->FindClass(env, "java/net/InetAddress");
      jmethodID ctor = (*env)->GetMethodID(env, inetaddress_class,
					   "<init>", "()V");
      JAVARLOG1(MYLOG, 5, "get_object_inetaddress(..., \"%s\") "
		"address == null, creating\n", classname);
      addr_obj = (*env)->NewObject(env, inetaddress_class, ctor, NULL);
      (*env)->SetObjectField(env, object, address_obj_field, addr_obj);
    }

  return addr_obj;
}

/*
 * Fills in InetAddress address and family.
 * address is given in host order
 */
static void
set_inetaddress(JNIEnv *env, jobject address_obj, jint address)
{
  JAVARLOG1(MYLOG, 3, "set_inetaddress(.., addr=0x%08x[host order]);\n",
	    address);
  set_object_int(env, address_obj, "java/net/InetAddress", "address", address);
}

static void
set_object_fd(JNIEnv *env, jobject obj, int sock_fd, const char *classname)
{
  jobject sock_class;
  jobject file_desc;
  jclass fd_class;
  jfieldID field;

  /* first we need to get the FileDescriptor object. */
  sock_class = (*env)->FindClass(env, classname);
  field = (*env)->GetFieldID(env, sock_class, "fd", "Ljava/io/FileDescriptor;");
  file_desc = (*env)->GetObjectField(env, obj, field);

  /* now we get the fd field of *that* object, and return it. */
  fd_class = (*env)->FindClass(env, "java/io/FileDescriptor");
  field = (*env)->GetFieldID(env, fd_class, "fd", "I");
  (*env)->SetIntField(env, file_desc, field, sock_fd);
}

static void
fill_in_sockaddr(JNIEnv *env,
		 jobject address,
		 int port,
		 struct sockaddr_in *sock_addr)
{
  jclass address_class = (*env)->FindClass(env, "java/net/InetAddress");
  jmethodID getAddress = (*env)->GetMethodID(env, address_class, "getAddress", "()[B");
  jbyteArray addr;
  jbyte *addr_bytes;

  memset(sock_addr, 0, sizeof(*sock_addr));

  assert(NULL != getAddress);

  addr = (*env)->CallObjectMethod(env, address, getAddress);

  addr_bytes = (*env)->GetByteArrayElements(env, addr, NULL);

  /* uhh... this works because:
     1: ints are 32 bytes.
     2: the address is returned in network order from getAddress
     3: we're lucky. :)
     */
  memcpy(&sock_addr->sin_addr.s_addr, addr_bytes, 4); 

#ifdef HAVE_SIN_LEN
  sock_addr->sin_len = sizeof(struct sockaddr_in);
#endif
  /* XXX Should probably get family from InetAddress.family */
  sock_addr->sin_family = AF_INET;
  sock_addr->sin_port = htons(port);

  (*env)->ReleaseByteArrayElements(env, addr, addr_bytes, 0);

  JAVARLOG2(MYLOG, 2, "making address bytes=0x%08x "
	    "port=0x%04x network order\n",
	    sock_addr->sin_addr.s_addr, sock_addr->sin_port);
}

/* Return local port */
static jint
socket_bind(JNIEnv *env,
	    int sock_fd,
	    jobject address,
	    jint port)
{
  struct sockaddr_in sock_addr;
  int res = -1;

  fill_in_sockaddr(env, address, port, &sock_addr);

  res = bind(sock_fd, (struct sockaddr*)&sock_addr, sizeof(sock_addr));
  JAVARLOG5(MYLOG, 2, "%d = bind(%d, addr{addr=0x%08x port=0x%04x}, %d)\n",
	    res, sock_fd,
	    sock_addr.sin_addr.s_addr, sock_addr.sin_port, sizeof(sock_addr));

  log_sockname(sock_fd);

  if (res == -1)
    {
      char buffer[100];
      snprintf(buffer, 99, "bind(%d, (sin_addr.s_addr=0x%x, sin_port=%d),..",
	       sock_fd, sock_addr.sin_addr.s_addr, port);
      switch (errno) {
      case ENOTSOCK:
      case EBADF:
	strcat(buffer, " failed, bad sockfd.");
	throw_Exception(env, JAVA_NET_BINDEXCEPTION, buffer);
	return 0;
	break;
	/* case EACCESS: /* Unknown on Linux */
      case EINVAL:
	strcat(buffer, " failed, invalid addr.");
	throw_Exception(env, JAVA_NET_BINDEXCEPTION, buffer);
	return 0;
	break;
      case ENOENT:
      default:
	JAVARLOG1(MYLOG, 2, "bind() failed, errno=0x%x\n",errno);
	strcat(buffer, " failed, unknown error");
	throw_Exception(env, JAVA_NET_BINDEXCEPTION, buffer);
	return 0;
	break;
      }
    }
  {
    int addr_len = sizeof(sock_addr);
    getsockname(sock_fd, &sock_addr, &addr_len);
  }
  return (jint)ntohs(sock_addr.sin_port);
}

/*
 * Return address_object(InetAddress).address in host order
 */
static jint
get_inetaddress(JNIEnv *env, jobject address_obj)
{
  jclass address_class = (*env)->FindClass(env, "java/net/InetAddress");
  jfieldID add_field = (*env)->GetFieldID(env, address_class, "address", "I");

  return (*env)->GetIntField(env, address_obj, add_field);
}

/***  java.net.PlainDatagramSocketImpl  ***/

/*
 * The PlainDatagramSocketImpl native stub implementation was mostly done
 * by Petter with almost no previous socket layer programming experience.
 * There are probably lots of improvments and optimizing to do.
 */

JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_bind(JNIEnv *env,
					   jobject obj,
					   jint lport,
					   jobject laddr)
{
  int sock_fd = (int)get_object_fd(env, obj, "java/net/DatagramSocketImpl");

  jint port = socket_bind(env, sock_fd, laddr, lport);
  set_object_int(env, obj, "java/net/DatagramSocketImpl", "localPort", port);
}

JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_send(JNIEnv *env,
					   jobject obj,
					   jobject packet)
{
  struct sockaddr_in sock_addr;
  size_t addr_len = sizeof(sock_addr);
  int sock_fd = get_object_fd(env, obj, "java/net/DatagramSocketImpl");
  int retval = -1;
  jobject packet_class = (*env)->GetObjectClass(env, packet);
  jfieldID length_field = (*env)->GetFieldID(env, packet_class, "length", "I");
  jfieldID buf_field = (*env)->GetFieldID(env, packet_class, "buf", "[B");
  jbyte *buf = NULL;
  jint length;

  assert(NULL != packet_class);
  assert(NULL != length_field);
  assert(NULL != buf_field);

  length = (*env)->GetIntField(env, packet, length_field);

  { /* set addr from packet->address->address and port */
    jint port = get_object_int(env, packet, "java/net/DatagramPacket", "port");
    jobject address_obj = get_object_inetaddress(env, packet,
						 "java/net/DatagramPacket");

    fill_in_sockaddr(env, address_obj, (int)port, &sock_addr);
  }

  { /* Dig out buffer from packet */
    jbyteArray buf_array = (*env)->GetObjectField(env, packet, buf_field);
    assert(NULL != buf_array);
    buf = (*env)->GetByteArrayElements(env, buf_array, NULL);
    assert(NULL != buf);

    JAVARLOG6(MYLOG, 2, "sendto(%d, 0x%x, %d, 0, addr{addr=0x%08x port=0x%04x}, %d);\n",
	      sock_fd, buf, length,
	      sock_addr.sin_addr.s_addr, sock_addr.sin_port, addr_len);

    retval = sendto(sock_fd, (const char*)buf, (int)length, 0,
		    (const struct sockaddr *)&sock_addr, addr_len);

    JAVARLOG7(MYLOG, 2, "%d = sendto(%d, 0x%x, %d, 0, addr{addr=0x%08x port=0x%04x}, %d);\n",
	      retval, sock_fd, buf, length,
	      sock_addr.sin_addr.s_addr, sock_addr.sin_port, addr_len);

    log_sockname(sock_fd);

    /* Nothing should have been written to the buffer, so I throw it away. */
    (*env)->ReleaseByteArrayElements(env, buf_array, buf, JNI_ABORT);
  }

  if (0 > retval) {
    JAVARLOG2(MYLOG, 2, "Error in sendto, errno=%d (%s)\n",
	      errno, errorstr(errno));
    throw_Exception(env, JAVA_NET_SOCKETEXCEPTION, "sendto failed");
    return;
  }

}

JNIEXPORT jint JNICALL
Java_java_net_PlainDatagramSocketImpl_peek(JNIEnv *env,
					   jobject obj,
					   jobject address)
{
  int sock_fd = get_object_fd(env, obj, "java/net/DatagramSocketImpl");
  struct sockaddr_in sock_addr;
  size_t addr_len = sizeof(sock_addr);
  int retval = -1;

  retval = recvfrom(sock_fd, 0, 0, MSG_PEEK,
		    (struct sockaddr*)&sock_addr, &addr_len);
  if (0 > retval)
    {
      throw_Exception(env, JAVA_NET_SOCKETEXCEPTION,
		      "recvfrom(.., MSG_PEEK) failed");
      return 0;
    }

  set_inetaddress(env, address, ntohl(sock_addr.sin_addr.s_addr));

  return retval;
}

JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_receive(JNIEnv *env,
					      jobject obj,
					      jobject packet)
{
  int sock_fd = (int)get_object_fd(env, obj, "java/net/DatagramSocketImpl");
  struct sockaddr_in sock_addr;
  size_t addr_len = sizeof(sock_addr);
  jclass packet_class = (*env)->FindClass(env, "java/net/DatagramPacket");
  jfieldID buf_field = (*env)->GetFieldID(env, packet_class, "buf", "[B");
  jbyte *buf = NULL;
  int retval = -1;

  assert(NULL != packet_class);

  /* Dig out buffer from packet */
  {
    jint length =
      get_object_int(env, packet, "java/net/DatagramPacket", "length");
    jbyteArray buf_array = (*env)->GetObjectField(env, packet, buf_field);
    buf = (*env)->GetByteArrayElements(env, buf_array, NULL);

    memset(&sock_addr, 0, sizeof(sock_addr));

    JAVARLOG6(MYLOG, 3,
	      "recvfrom(fd=%d, buf=0x%d, len=%d, 0, addr{addr=0x%08x port=0x%04x}, %d);\n",
	      sock_fd, buf, length,
	      sock_addr.sin_addr.s_addr, sock_addr.sin_port, addr_len);
  
    retval = recvfrom(sock_fd, (char*)buf, (int)length, 0, &sock_addr, &addr_len);
    
    JAVARLOG7(MYLOG, 3,
	      "%d = recvfrom(fd=%d, buf=0x%d, len=%d, 0, addr{addr=0x%08x port=0x%04x}, %d);\n",
	      retval, sock_fd, buf, length,
	      sock_addr.sin_addr.s_addr, (int)sock_addr.sin_port, addr_len);

    (*env)->ReleaseByteArrayElements(env, buf_array, buf, 0);
  }

  if (0 > retval)
    {
      throw_Exception(env, JAVA_NET_SOCKETEXCEPTION, "recvfrom failed");
      return;
    }

  set_object_int(env, packet, "java/net/DatagramPacket", "length", (jint)retval);
  set_object_int(env, packet, "java/net/DatagramPacket", "port",
		 (jint)ntohs((int)sock_addr.sin_port));

  { /* set packet->address->address from addr.sin_addr.s_addr */
    jobject address =
      get_object_inetaddress(env, packet, "java/net/DatagramPacket");
    set_inetaddress(env, address, ntohl(sock_addr.sin_addr.s_addr));
  }
}

JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_setTTL(JNIEnv *env,
					     jobject obj,
					     jbyte ttl)
{
  int sock_fd = (int)get_object_fd(env, obj, "java/net/DatagramSocketImpl");
  int ittl = (int)ttl;

  setsockopt(sock_fd, IPPROTO_IP,
	     IP_TTL, &ittl, sizeof(int));
}

JNIEXPORT jbyte JNICALL
Java_java_net_PlainDatagramSocketImpl_getTTL(JNIEnv *env,
					     jobject obj)
{
  int sock_fd = (int)get_object_fd(env, obj, "java/net/DatagramSocketImpl");
  int ttl = 0;
  /*
   * Solaris 2.5, Irix 6.2 and Linux RedHat 5.0 have int* as 5. param
   * to getsockopt(), not size_t*.
   */
  int len = 0;

  getsockopt(sock_fd, IPPROTO_IP,
	     IP_TTL, &ttl, &len);

  return (jbyte)ttl;
}

JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_join(JNIEnv *env,
					   jobject obj,
					   jobject multicast_addr)
{
  int sock_fd = (int)get_object_fd(env, obj, "java/net/DatagramSocketImpl");
  struct ip_mreq mreq;

  mreq.imr_multiaddr.s_addr = htonl(get_inetaddress(env, multicast_addr));
  mreq.imr_interface.s_addr = htonl(INADDR_ANY);

  setsockopt(sock_fd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
	     &mreq, sizeof(mreq));
}

JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_leave(JNIEnv *env,
					    jobject obj,
					    jobject multicast_addr)
{
  int sock_fd = (int)get_object_fd(env, obj, "java/net/DatagramSocketImpl");
  struct ip_mreq mreq;

  mreq.imr_multiaddr.s_addr = htonl(get_inetaddress(env, multicast_addr));
  mreq.imr_interface.s_addr = htonl(INADDR_ANY);

  setsockopt(sock_fd, IPPROTO_IP, IP_DROP_MEMBERSHIP,
	     &mreq, sizeof(mreq));
}

JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_datagramSocketCreate(JNIEnv *env,
							   jobject obj)
{
  int sock_fd;

  sock_fd = socket(PF_INET, SOCK_DGRAM, 0);

  if (sock_fd == -1)
    {
      throw_Exception(env, JAVA_IO_IOEXCEPTION,
		      "datagram socket creation failed.");
    }

  /* we need to set all the system fd's to non blocking. */
  /* toshok - actually, we do want them blocking, since the thread
     library is the one that's supposed to take care of blocking
     threads instead of process.  We're not supposed to care about this. */
  /*  fcntl(sock_fd, F_SETFL, O_NONBLOCK);*/

  set_object_fd(env, obj, sock_fd, "java/net/DatagramSocketImpl");
}

JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_datagramSocketClose(JNIEnv *env,
							  jobject obj)
{
  int sock_fd = (int)get_object_fd(env, obj, "java/net/DatagramSocketImpl");

  close(sock_fd);
}

JNIEXPORT void JNICALL
Java_java_net_PlainDatagramSocketImpl_socketSetOption(JNIEnv *env,
						      jobject obj,
						      jint opt,
						      jobject value)
{
  JAVARLOG2(MYLOG, 2, "socketSetOption(.., opt=%d, value=0x%x)\n", opt, value);
  (*env)->FatalError(env, "java.net.PlainDatagramSocketImpl.socketSetOption not implemented");
}

JNIEXPORT jint JNICALL
Java_java_net_PlainDatagramSocketImpl_socketGetOption(JNIEnv *env,
						      jobject obj,
						      jint opt)
{
  int sock_fd = (int)get_object_fd(env, obj, "java/net/DatagramSocketImpl");
  int ret_val;
  
  switch (opt)
    {
    case 0x000F: /* SO_BINDADDR */
      {
	struct sockaddr_in name;
	int namelen;
	int foo = getsockname(sock_fd, (struct sockaddr*)&name, &namelen);

	if (foo == -1 || namelen != sizeof(struct sockaddr_in))
	  {
	    throw_Exception(env, JAVA_NET_SOCKETEXCEPTION,
			    "Error getting bind address for socket.");
	    return -1;
	  }

	/* XXX What is the byte order of this value? [pere] */
	return ntohl(name.sin_addr.s_addr);
      }
    break;
    case 0x0004: /* SO_REUSEADDR */
      {
	(*env)->FatalError(env, "socketGetOption don't handle SO_REUSEADDR yet\n");
	return -1;
      }
    break;
    case 0x1006: /* SO_TIMEOUT */
      {
#if defined( SO_RCVTIMEO )
	/* choose either the read timeout or the write timeout.  we
	   assume they're both set to the same thing. */
	jint timeout;
	int timeout_length;

	ret_val = getsockopt(sock_fd, SOL_SOCKET, SO_RCVTIMEO, &timeout, &timeout_length);

	return timeout;
#else
#warning "Missing SO_RCVTIMEO!"
	(*env)->FatalError(env, "socketGetOption failed.  Missing SO_RCVTIMEO\n");
	return -1;
#endif
      }
    break;
    case 0x0080: /* SO_LINGER */
      {
	struct linger linger;
	int linger_length;
		
	ret_val = getsockopt(sock_fd, SOL_SOCKET, SO_LINGER, &linger, &linger_length);
		
	return linger.l_linger;
      }
    break;
    default:
      {
	throw_Exception(env, JAVA_NET_SOCKETEXCEPTION,
			"Invalid option sent to socketGetOption\n");
	return -1;
      }
    }
}


/***  java.net.PlainSocketImpl  ***/

JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_socketCreate(JNIEnv *env,
					   jobject obj,
					   jboolean isServer /* what's this do? */)
{
  int sock_fd;

  sock_fd = socket(PF_INET, SOCK_STREAM, 0);

  if (sock_fd == -1)
    {
      throw_Exception(env, JAVA_IO_IOEXCEPTION, "socket creation failed.");
    }

  /* we need to set all the system fd's to non blocking. */
  /* toshok - actually, we do want them blocking, since the thread
     library is the one that's supposed to take care of blocking
     threads instead of process.  We're not supposed to care about this. */
  /*  fcntl(sock_fd, F_SETFL, O_NONBLOCK);*/

  set_object_fd(env, obj, sock_fd, "java/net/SocketImpl");
}

JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_socketConnect(JNIEnv *env,
					    jobject obj,
					    jobject address,
					    jint port)
{
  struct sockaddr_in sock_addr;
  int sock_fd = get_object_fd(env, obj, "java/net/SocketImpl");

  fill_in_sockaddr(env, address, port, &sock_addr);

  if (connect(sock_fd, (struct sockaddr*)&sock_addr, sizeof(struct sockaddr_in)) == -1)
    {
      JAVARLOG2(MYLOG, 3, "connect failed, errno=%d (%s)\n",
		errno, errorstr(errno));
      throw_Exception(env, JAVA_IO_IOEXCEPTION, "socket connect failed.");
    }
}

JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_socketBind(JNIEnv *env,
					 jobject obj,
					 jobject address,
					 jint lport)
{
  int sock_fd = get_object_fd(env, obj, "java/net/SocketImpl");

  jint port = socket_bind(env, sock_fd, address, lport);
  set_object_int(env, obj, "java/net/PlainSocketImpl", "localport", port);
}

JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_socketListen(JNIEnv *env,
					   jobject obj,
					   jint count)
{
  int sock_fd = get_object_fd(env, obj, "java/net/SocketImpl");

  if (listen(sock_fd, count) == -1)
    {
      throw_Exception(env, JAVA_IO_IOEXCEPTION, "socket listen failed.");
    }
}

JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_socketAccept(JNIEnv *env,
					   jobject obj)
{
  int sock_fd = get_object_fd(env, obj, "java/net/SocketImpl");
  struct sockaddr_in addr;
  size_t addr_len = sizeof(addr);
  int retval = -1;

  retval = accept(sock_fd, (struct sockaddr*)&addr, &addr_len);
  if (0 > retval)
    {
      JAVARLOG2(MYLOG, 3, "connect failed, errno=%d (%s)\n",
		errno, errorstr(errno));
      throw_Exception(env, JAVA_IO_IOEXCEPTION, "accept() failed");
      return;
    }
  set_object_fd(env, obj, retval, "java/net/SocketImpl");
  JAVARLOG5(MYLOG, 3, "%d = accept(fd=%d, port=0x%04x addr=0x%08x, len=%d\n",
	    retval, sock_fd, (int)addr.sin_port, addr.sin_addr.s_addr, addr_len);

  /* Fill socket with info */
  set_object_int(env, obj, "java/net/SocketImpl", "port",
		 ntohs((int)addr.sin_port));
  {
    jobject addr_obj = get_object_inetaddress(env, obj,
					      "java/net/SocketImpl");
    assert(sizeof(jint) == sizeof(addr.sin_addr.s_addr));
    set_inetaddress(env, addr_obj, ntohl(addr.sin_addr.s_addr));
  }
}

JNIEXPORT jint JNICALL
Java_java_net_PlainSocketImpl_socketAvailable(JNIEnv *env,
					      jobject obj)
{
  int sock_fd = get_object_fd(env, obj, "java/net/SocketImpl");
  int retval = -1;
  int length = 0;
#if defined( HAVE_IOCTL ) && defined( FIONREAD )
  retval = ioctl(sock_fd, FIONREAD, &len);
  if (retval < 0) {
    throw_Exception(env, JAVA_IO_IOEXCEPTION, "ioctl(,FIONREAD,) failed");
  }
#else
  static struct timeval tm = {0,0};
  fd_set rd;
  FD_ZERO(&rd);
  FD_SET(sock_fd, &rd);
  retval = select(sock_fd+1, &rd, NULL, NULL, &tm);
  if (1 == retval)
    length = 1;
#endif
  return length;
}

JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_socketClose(JNIEnv *env,
					  jobject obj)
{
  int sock_fd = get_object_fd(env, obj, "java/net/SocketImpl");

  close(sock_fd);
}

JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_initProto(JNIEnv *env,
					jobject obj)
{
  /* this doesn't do anything on unix.. at least, not that I *know* of */
}

JNIEXPORT void JNICALL
Java_java_net_PlainSocketImpl_socketSetOption(JNIEnv *env,
					      jobject obj,
					      jint cmd,
					      jboolean on,
					      jobject value)
{
  int sock_fd = get_object_fd(env, obj, "java/net/SocketImpl");

  switch (cmd)
    {
    case 0x0001: /* TCP_NODELAY */
      /* hmmm... */
      break;
    case 0x0080: /* SO_LINGER */
      {
	struct linger linger;

	if (on)
	  {
	    jclass cls = (*env)->GetObjectClass(env, value);
	    jmethodID intValue = (*env)->GetMethodID(env, cls, "intValue", "()I");
		
	    linger.l_linger = (*env)->CallIntMethod(env, value, intValue);

	    setsockopt(sock_fd, SOL_SOCKET, SO_LINGER, &linger, sizeof(struct linger));
	  }
	else
	  {
	    /* how do we disable it? */
	  }
	break;
      }
    case 0x1006: /* SO_TIMEOUT */
      {
#if defined( SO_RCVTIMEO )
	jclass cls = (*env)->GetObjectClass(env, value);
	jmethodID intValue = (*env)->GetMethodID(env, cls, "intValue", "()I");
	jint timeout = (*env)->CallIntMethod(env, value, intValue);
		
	setsockopt(sock_fd, SOL_SOCKET, SO_RCVTIMEO, &timeout, sizeof(timeout));
#else
#warning "Missing SO_RCVTIMEO!"
	(*env)->FatalError(env, "socketSetOption failed.  Missing SO_RCVTIMEO\n");
#endif
      }
    default:
      {
	throw_Exception(env, JAVA_NET_SOCKETEXCEPTION,
			"Invalid option sent to socketSetOption\n");
	break;
      }
    }
}

JNIEXPORT jint JNICALL
Java_java_net_PlainSocketImpl_socketGetOption(JNIEnv *env,
					      jobject obj,
					      jint opt)
{
  int sock_fd = get_object_fd(env, obj, "java/net/SocketImpl");
  int ret_val;
  
  switch (opt)
    {
    case 0x0001: /* TCP_NODELAY */
      /* hmmm... */
      return -1;
      break;
    case 0x000F: /* SO_BINDADDR */
      {
	struct sockaddr_in name;
	int namelen;
	int foo = getsockname(sock_fd, (struct sockaddr*)&name, &namelen);

	if (foo == -1 || namelen != sizeof(struct sockaddr_in))
	  {
	    throw_Exception(env, JAVA_NET_SOCKETEXCEPTION,
			    "Error getting bind address for socket.");
	    return -1;
	  }

	return ntohl(name.sin_addr.s_addr);
      }
    break;
    case 0x0080: /* SO_LINGER */
      {
	struct linger linger;
	int linger_length;
		
	ret_val = getsockopt(sock_fd, SOL_SOCKET, SO_LINGER, &linger, &linger_length);
		
	return linger.l_linger;
      }
    break;
    default:
      {
	throw_Exception(env, JAVA_NET_SOCKETEXCEPTION,
			"Invalid option sent to socketGetOption\n");
	return -1;
      }
    }
}
