/* -*- Mode: C; c-file-style: "gnu" -*-
   clparse.c -- class file parsing.
   Created: Chris Toshok <toshok@hungry.com>, 23-Jul-97
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "config.h"
#include "ClazzFile.h"
#include "objects.h"
#include "class-repository.h"
#include "native-threads.h"
#include "log.h"
#include "gc.h"
#include "jniint.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netinet/in.h>
#include <assert.h>

#ifdef HAVE_MMAN_H
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#endif

#ifndef MAP_FAILED
#define MAP_FAILED ((char*)(-1))
#endif

#define MYLOG "ParseClass"

#define EXTRACT_U1(dest, buf, pos) \
do { \
       assert(sizeof(u1) == sizeof(*dest)); \
					      (*(dest)) = (buf)[ (*(pos))++ ]; \
										 } while (0);

#define EXTRACT_U2(dest, buf, pos) \
do { \
       assert(sizeof(u2) == sizeof(*dest)); \
					      memcpy((void*) (dest), (void*) ((buf) + (*(pos))), 2); \
												       (*(dest)) = ntohs(*(dest)); \
																     (*(pos)) += 2; \
																		      } while (0);

#define EXTRACT_U4(dest, buf, pos) \
do { \
       assert(sizeof(u4) == sizeof(*dest)); \
					      memcpy ((void *) (dest), (void *) ((buf)+(*(pos))), 4); \
													(*(dest)) = ntohl (*(dest)); \
																       (*(pos)) += 4; \
																			} while (0);

static GC_obj gc_obj_template = {
  NULL, NULL,

  GC_FLAG_CLASS,

  sizeof(ClazzFile),

  class_finalize
};

char*
slashes_to_dots(char *str)
{
  int i;

  for (i = 0; i < strlen(str); i ++)
    if (str[i] == '/')
      str[i] = '.';

  return str;
}

char*
dots_to_slashes(char *str)
{
  int i;

  for (i = 0; i < strlen(str); i ++)
    if (str[i] == '.')
      str[i] = '/';

  return str;
}

char*
slashes_to_underscores(char *str)
{
  int i;

  for (i = 0; i < strlen(str); i ++)
    if (str[i] == '/')
      str[i] = '_';

  return str;
}

char*
dots_to_underscores(char *str)
{
  int i;

  for (i = 0; i < strlen(str); i ++)
    if (str[i] == '.')
      str[i] = '_';

  return str;
}

static char *
load_class_file(char *file_name, int *length_return)
{
  char *buf;
  FILE *fp;
  int file_size;
  struct stat stbuf;

  file_size = stat(file_name, &stbuf);

  if (file_size == -1)
    {
      JAVARLOG1(MYLOG, 2, 
                "load_file_to_memory failed for file '%s'\n", file_name);
      return NULL;
    }

  file_size = stbuf.st_size;

  fp = fopen(file_name, "r");

  if (fp == NULL)
    {
      JAVARLOG1(MYLOG, 2, "couldn't open '%s'\n", file_name);
      return NULL;
    }

#ifdef HAVE_MMAN_H
  {
    JAVARLOG1(MYLOG, 2, "load_class_file reading from '%s'\n", file_name);

    buf = (char*)mmap(0, file_size, PROT_READ, MAP_SHARED, fileno(fp), 0);

    assert(MAP_FAILED != buf); /* XXX should be handled more gracefully */
  }
#else
  {
    int retval;
    int feofmark;

    buf = (char *) malloc ( (size_t) file_size );

    JAVARLOG1(MYLOG, 2, "load_class_file reading from '%s'\n", file_name);

    retval = fread ((void *) buf, (size_t) 1, (size_t) file_size, fp);
    feofmark = feof(fp);
    clearerr(fp);

    if ((retval < file_size && retval >= 0) && feofmark == 0)
      {
	JAVARLOG2(MYLOG, 2, "load_file_to_memory: only got %d out of %d bytes\n",
		  retval, file_size);
	JAVARLOG0(MYLOG, 2, "fread failed.\n");
	return NULL;
      }

    if (retval < 0)
      {
	JAVARLOG0(MYLOG, 2, "fread failed.\n");
	return NULL;
      }
  }
#endif

  fclose(fp);

  *length_return = file_size;
  return buf;
}

static ConstantPoolEntry *
get_constant(ClazzFile *clazz, int index)
{
  assert(NULL != clazz);
  assert(index > 0);
  assert(index < clazz->constant_pool_count);

  return &clazz->constants[index];
}

static void
parse_field_attribute(JNIEnv *env, char *buf, int *buf_pos, FieldStruct *field)
{
  u2 attribute_name;
  u4 attribute_length;
  char *att_name;

  EXTRACT_U2(&attribute_name, buf, buf_pos);
  EXTRACT_U4(&attribute_length, buf, buf_pos);

  att_name = ResolveUtf8(env, field->clazz,
			 get_constant(field->clazz, attribute_name));

  if (!strcmp("ConstantValue", att_name))
    {
      u2 index;

      EXTRACT_U2(&index, buf, buf_pos);

      field->has_constant_value = 1;
      field->constant_value_index = index;
      field->constant_value.l = (void*)VALUE_UNRESOLVED;
    }
  else if (!strcmp("Deprecated", att_name))
    {
      field->flags |= FLAG_DEPRECATED;
    }
  else if (!strcmp("Synthetic", att_name))
    field->flags |= FLAG_SYNTHETIC;
  else
    {
#ifdef DEBUG
      printf ("Unhandled field attribute type '%s'\n", att_name);
#endif
      *buf_pos += attribute_length;
    }
}

static FieldStruct *
parse_field(JNIEnv *env, char *buf, int *buf_pos, ClazzFile *clazz)
{
  FieldStruct *new_field = (FieldStruct*)calloc(1, sizeof(FieldStruct));
  u2 tmp_u2;

  new_field->clazz = clazz;
  new_field->field_index = VALUE_UNRESOLVED;
  new_field->java_type = VALUE_UNRESOLVED;

  EXTRACT_U2(&new_field->access_flags, buf, buf_pos);

  if (new_field->access_flags & ACC_STATIC)
    clazz->num_static_fields ++;
  else
    clazz->num_instance_fields ++;

  /* get the name of the field */
  EXTRACT_U2(&tmp_u2, buf, buf_pos);

  new_field->name = ResolveUtf8(env, clazz, get_constant(clazz,tmp_u2));

  /* get the signature of the field */
  EXTRACT_U2(&tmp_u2, buf, buf_pos);
  new_field->sig = SIG_parseFromJavaSig(env, ResolveUtf8(env, clazz, get_constant(clazz,tmp_u2)));
  new_field->sig_size_in_bytes = SIG_sizeInBytes(env, new_field->sig);

  /* parse the attributes of this field */
  EXTRACT_U2(&tmp_u2, buf, buf_pos);
  if (tmp_u2)
    {
      int att_num;

      for (att_num = 0;
	   att_num < tmp_u2;
	   att_num ++)
	parse_field_attribute(env, buf, buf_pos, new_field);
    }

  return new_field;
}

static void
parse_method_attribute(JNIEnv *env, char *buf, int *buf_pos, MethodStruct *method)
{
  u2 attribute_name;
  u4 attribute_length;
  char *att_name;

  EXTRACT_U2(&attribute_name, buf, buf_pos);
  EXTRACT_U4(&attribute_length, buf, buf_pos);

  att_name = ResolveUtf8(env,method->clazz,
			 get_constant(method->clazz, attribute_name));

  if (!strcmp("Code", att_name))
    {
      u2 tmp_u2;

      EXTRACT_U2(&method->max_stack, buf, buf_pos);
      EXTRACT_U2(&method->max_locals, buf, buf_pos);
      EXTRACT_U4(&method->code_length, buf, buf_pos);

      if (method->code_length)
	{
	  method->code = (u1*)calloc(method->code_length,
				     sizeof(u1));

	  memcpy(method->code, buf + *buf_pos, method->code_length);
	  *buf_pos += method->code_length;
	}

      EXTRACT_U2(&method->num_exception_blocks, buf, buf_pos);
      
      if (method->num_exception_blocks)
	{
	  int i;
	  method->exceptions = (ExceptionBlock*)calloc(method->num_exception_blocks,
						       sizeof(ExceptionBlock));

	  for (i = 0;
	       i < method->num_exception_blocks;
	       i ++)
	    {
	      EXTRACT_U2(&method->exceptions[i].start_pc, buf, buf_pos);
	      EXTRACT_U2(&method->exceptions[i].end_pc, buf, buf_pos);
	      EXTRACT_U2(&method->exceptions[i].handler_pc, buf, buf_pos);
	      EXTRACT_U2(&method->exceptions[i].catch_clazz_index, buf, buf_pos);
	    }
	}

      EXTRACT_U2(&tmp_u2, buf, buf_pos);
      if (tmp_u2)
	{
	  int att_num;

	  for (att_num = 0; att_num < tmp_u2; att_num ++)
	    parse_method_attribute(env, buf, buf_pos, method);
	}
    }
  else if (!strcmp("Exceptions", att_name))
    {
      u2 tmp_u2, foo;
      int i;

      EXTRACT_U2(&tmp_u2, buf, buf_pos);

      for (i = 0; i < tmp_u2; i ++)
	EXTRACT_U2(&foo, buf, buf_pos);

      /* ?????  what the hell is it attached here for?*/
    }
  else if (!strcmp("LineNumberTable", att_name))
    {
      EXTRACT_U2(&method->num_line_number_blocks, buf, buf_pos);

      if (method->num_line_number_blocks)
	{
	  int i;

	  method->line_numbers = (LineNumberBlock*)calloc(method->num_line_number_blocks,
							  sizeof(LineNumberBlock));
	  
	  for (i = 0; i < method->num_line_number_blocks; i ++)
	    {
	      EXTRACT_U2(&method->line_numbers[i].start_pc, buf, buf_pos);
	      EXTRACT_U2(&method->line_numbers[i].line_number, buf, buf_pos);
	    }
	}
    }
  else if (!strcmp("LocalVariableTable", att_name))
    {
      EXTRACT_U2(&method->num_local_variables, buf, buf_pos);

      if (method->num_local_variables)
	{
	  int i;
	  u2 tmp_u2;
	  method->local_variables = (LocalVariableEntry*)calloc(method->num_local_variables,
								sizeof(LocalVariableEntry));

	  for (i = 0; i < method->num_local_variables; i ++)
	    {
	      EXTRACT_U2(&method->local_variables[i].start_pc, buf, buf_pos);
	      EXTRACT_U2(&tmp_u2, buf, buf_pos);
	      method->local_variables[i].end_pc = method->local_variables[i].start_pc + tmp_u2;

	      EXTRACT_U2(&tmp_u2, buf, buf_pos);
	      method->local_variables[i].name =
		ResolveUtf8(env, method->clazz,
			    get_constant(method->clazz, tmp_u2));

	      EXTRACT_U2(&tmp_u2, buf, buf_pos);
	      method->local_variables[i].sig =
		SIG_parseFromJavaSig(env,
				     ResolveUtf8(env, method->clazz, 
						 get_constant(method->clazz, tmp_u2)));

	      EXTRACT_U2(&method->local_variables[i].slot, buf, buf_pos);
	    }
	}

    }
  else if (!strcmp("Deprecated", att_name))
    {
      method->flags |= FLAG_DEPRECATED;
    }
  else if (!strcmp("Synthetic", att_name))
    {
      method->flags |= FLAG_SYNTHETIC;
    }
  else
    {
#ifdef DEBUG
      printf ("Unhandled method attribute type '%s'\n", att_name);
#endif
      *buf_pos += attribute_length;
    }
}

static MethodStruct *
parse_method(JNIEnv *env, char *buf, int *buf_pos, ClazzFile *clazz)
{
  MethodStruct *new_method = (MethodStruct*)calloc(1, sizeof(MethodStruct));
  u2 tmp_u2;

  new_method->clazz = clazz;

  EXTRACT_U2(&new_method->access_flags, buf, buf_pos);

  /* get the name of the method */
  EXTRACT_U2(&tmp_u2, buf, buf_pos);
  new_method->name = ResolveUtf8(env, clazz, get_constant(clazz,tmp_u2));

  /* get the signature of the method */
  EXTRACT_U2(&tmp_u2, buf, buf_pos);
  new_method->sig = SIG_parseFromJavaSig(env,
					 ResolveUtf8(env, clazz, get_constant(clazz,tmp_u2)));

  if (!strcmp(new_method->name, "<clinit>")
      && !strcmp(ResolveUtf8(env, clazz, get_constant(clazz, tmp_u2)), "()V")
      && new_method->access_flags & ACC_STATIC)
    {
      clazz->access_flags |= HAS_CLINIT;
    }

  JAVARLOG2(MYLOG, 2, "Parsing method %s %s\n", new_method->name,
	    new_method->access_flags & ACC_NATIVE ? "(native)" : "");

  {
    /* figure out how many words need to be there to call this method. */
    /* this should perhaps be moved off until we resolve the method.... */
    int i;

    new_method->num_param_words = 0;
    for (i = 0; i < SIG_numParams(env, new_method->sig); i ++)
      new_method->num_param_words += SIG_sizeInWords(env, new_method->sig->method.params[i]);
  }
  
  /* parse the attributes of this method */
  EXTRACT_U2(&tmp_u2, buf, buf_pos);
  if (tmp_u2)
    {
      int att_num;

      for (att_num = 0;
	   att_num < tmp_u2;
	   att_num ++)
	parse_method_attribute(env, buf, buf_pos, new_method);
    }

  JAVARLOG1(MYLOG, 2, "Done parsing method %s\n", new_method->name);

  return new_method;
}

static void
parse_class_attribute(JNIEnv *env, char *buf, int *buf_pos, ClazzFile *clazz)
{
  u2 attribute_name;
  u4 attribute_length;
  char *att_name;

  EXTRACT_U2(&attribute_name, buf, buf_pos);
  EXTRACT_U4(&attribute_length, buf, buf_pos);

  att_name = ResolveUtf8(env, clazz, get_constant(clazz,attribute_name));

  if (!strcmp("SourceFile", att_name))
    {
      u2 source_file_index;
      EXTRACT_U2(&source_file_index, buf, buf_pos);

      clazz->source_filename = ResolveUtf8(env,clazz, get_constant(clazz,source_file_index));
    }
  else if (!strcmp("InnerClasses", att_name))
    {
      u2 i;

      EXTRACT_U2(&clazz->num_innerclasses, buf, buf_pos);

      if (clazz->num_innerclasses)
	clazz->innerclass_indices = (u2*)calloc(clazz->num_innerclasses, sizeof(u2));

      for (i = 0; i < clazz->num_innerclasses; i ++)
	{
	  EXTRACT_U2(&clazz->innerclass_indices[i], buf, buf_pos);
	}
    }
  else if (!strcmp("AbsoluteSourcePath", att_name))
    {
      u2 tmp;
      /* don't do anything...  should this just override SourcePath if present? */
      EXTRACT_U2(&tmp, buf, buf_pos);
    }
  else
    {
#ifdef DEBUG
      printf ("Unknown class attribute type '%s'\n", att_name);
#endif
      *buf_pos += attribute_length;
    }
}

static void
parse_constant(char *buf, int *buf_pos, ConstantPoolEntry *entry)
{
  u1 tmp_u1;

  EXTRACT_U1(&tmp_u1, buf, buf_pos);
  entry->generic.tag = tmp_u1;

  JAVARLOG1(MYLOG, 2, " parsing constant type %d\n", tmp_u1);
  switch (entry->generic.tag)
    {
    case CONSTANT_Utf8:
      {
	/* this is just screaming for optimization... */

	EXTRACT_U2(&entry->utf8_info.length, buf, buf_pos);
	entry->utf8_info.bytes = (u1*)calloc(entry->utf8_info.length, sizeof(u1));
#if 1
	memcpy(entry->utf8_info.bytes, buf + *buf_pos, entry->utf8_info.length);
	*buf_pos += entry->utf8_info.length;
#else
	for (i = 0; i < entry->utf8_info.length; i ++)
	  EXTRACT_U1(&entry->utf8_info.bytes[i], buf, buf_pos);
#endif
	break;
      }
    case CONSTANT_Unicode:
      {
	int i;

	EXTRACT_U2(&entry->unicode_info.length, buf, buf_pos);
	entry->unicode_info.bytes = (u2*)malloc(sizeof(u2) * entry->utf8_info.length);

	for (i  =0; i < entry->utf8_info.length; i ++)
	  EXTRACT_U2(&entry->unicode_info.bytes[i], buf, buf_pos);
	break;
      }
    case CONSTANT_Integer:
      EXTRACT_U4(&entry->integer_info.bytes, buf, buf_pos);
      break;
    case CONSTANT_Float:
      EXTRACT_U4(&entry->float_info.bytes, buf, buf_pos);
      break;
    case CONSTANT_Long:
      EXTRACT_U4(&entry->long_info.high_bytes, buf, buf_pos);
      EXTRACT_U4(&entry->long_info.low_bytes, buf, buf_pos);
      break;
    case CONSTANT_Double:
      EXTRACT_U4(&entry->double_info.high_bytes, buf, buf_pos);
      EXTRACT_U4(&entry->double_info.low_bytes, buf, buf_pos);
      break;
    case CONSTANT_Class:
      EXTRACT_U2(&entry->clazz_info.name_index, buf, buf_pos);
      break;
    case CONSTANT_String:
      EXTRACT_U2(&entry->string_info.string_index, buf, buf_pos);
      break;
    case CONSTANT_Fieldref:
      EXTRACT_U2(&entry->fieldref_info.class_index, buf, buf_pos);
      EXTRACT_U2(&entry->fieldref_info.name_and_type_index, buf, buf_pos);
      break;
    case CONSTANT_Methodref:
      EXTRACT_U2(&entry->methodref_info.class_index, buf, buf_pos);
      EXTRACT_U2(&entry->methodref_info.name_and_type_index, buf, buf_pos);
      break;
    case CONSTANT_InterfaceMethodref:
      EXTRACT_U2(&entry->interfacemethodref_info.class_index, buf, buf_pos);
      EXTRACT_U2(&entry->interfacemethodref_info.name_and_type_index, buf, buf_pos);
      break;
    case CONSTANT_NameAndType:
      EXTRACT_U2(&entry->nameandtype_info.name_index, buf, buf_pos);
      EXTRACT_U2(&entry->nameandtype_info.signature_index, buf, buf_pos);
      break;
    default:
      fprintf (stderr, "invalid constant tag: %d, pos is %d\n",
	       entry->generic.tag, *buf_pos);
    }
}

ClazzFile *
define_class(JNIEnv *env, char *buf, int buf_length)
{
  int buf_pos = 0;
  ClazzFile *clazz;
  u4 tmp_u4;
  u2 tmp_u2;
  int constant_num;
  u2 thisclass_index;

  clazz = (ClazzFile*)calloc(1, sizeof(ClazzFile));

  EXTRACT_U4(&tmp_u4, buf, &buf_pos);

  if (tmp_u4 != 0xCAFEBABE)
    {
      fprintf (stderr, "magic is wrong -- not a java class..\n");
      return NULL;
    }

  EXTRACT_U2(&clazz->minor, buf, &buf_pos);
  JAVARLOG1(MYLOG, 2, "minor_version is %d\n", clazz->minor);

  EXTRACT_U2(&clazz->major, buf, &buf_pos);
  JAVARLOG1(MYLOG, 2, "major_version is %d\n", clazz->major);

  EXTRACT_U2(&clazz->constant_pool_count, buf, &buf_pos);
  JAVARLOG1(MYLOG, 2, "%d constants in constant_pool_count\n", clazz->constant_pool_count);

  clazz->constants = (ConstantPoolEntry*) calloc (clazz->constant_pool_count + 1, /* fuckers. */
						  sizeof(ConstantPoolEntry));
  
  for (constant_num = 1;
       constant_num < clazz->constant_pool_count;
       constant_num ++)
    {
      ConstantPoolEntry *constant = get_constant(clazz,constant_num);
      JAVARLOG1(MYLOG, 2, "getting constant #%d\n", constant_num);

      parse_constant(buf, &buf_pos, constant);

      if (constant->generic.tag == CONSTANT_Double ||
	  constant->generic.tag == CONSTANT_Long)
	constant_num ++;
    }

  EXTRACT_U2 (&clazz->access_flags, buf, &buf_pos);
  EXTRACT_U2 (&thisclass_index, buf, &buf_pos);
  EXTRACT_U2 (&clazz->superclass_index, buf, &buf_pos);

  /* parse through the interface indices, filling in our
     pre-resolved array. */
  EXTRACT_U2 (&clazz->num_interfaces, buf, &buf_pos);
  if (clazz->num_interfaces)
    {
      int interface_num;
      JAVARLOG1(MYLOG, 2, "%d interfaces to read\n", clazz->num_interfaces);

      clazz->interface_indices = (u2*)calloc(clazz->num_interfaces,
					     sizeof(u2));

      clazz->interfaces = (ClazzFile**)calloc(clazz->num_interfaces,
					      sizeof(ClazzFile*));

      for (interface_num = 0;
	   interface_num < clazz->num_interfaces;
	   interface_num ++)
	{
	  EXTRACT_U2(&(clazz->interface_indices[interface_num]), buf, &buf_pos);
	}
    }

  /* parse through the fields of this class. */
  EXTRACT_U2 (&clazz->num_fields,buf, &buf_pos);
  if (clazz->num_fields)
    {
      int field_index;
      
      JAVARLOG3(MYLOG, 2, "%d (0x%x) fields to parse (pos=%d)\n",
		clazz->num_fields, clazz->num_fields, buf_pos);
      clazz->fields = (FieldStruct**)calloc(clazz->num_fields,
					    sizeof(FieldStruct*));
      
      for (field_index = 0;
	   field_index < clazz->num_fields;
	   field_index ++)
	{
	  JAVARLOG1(MYLOG, 2, "parsing field #%d\n", field_index);
	  clazz->fields[field_index] = parse_field(env, buf, &buf_pos,
						   clazz);
	}
    }

  /* parse through the methods of this class. */
  EXTRACT_U2 (&clazz->num_methods,buf, &buf_pos);
  if (clazz->num_methods)
    {
      int method_num;
      
      clazz->methods = (MethodStruct**)calloc(clazz->num_methods,
					      sizeof(MethodStruct*));
      
      for (method_num = 0;
	   method_num < clazz->num_methods;
	   method_num ++)
	{
	  clazz->methods[method_num] = parse_method(env, buf, &buf_pos,
						    clazz);
	}
    }

  /* parse through the attributes of this class */
  EXTRACT_U2 (&tmp_u2, buf, &buf_pos);
  if (tmp_u2)
    {
      int att_num;
      for (att_num = 0;
	   att_num < tmp_u2;
	   att_num ++)
	{
	  parse_class_attribute(env, buf, &buf_pos, clazz);
	}
    }

  /* fill in the name of this class. */
  clazz->class_name = ResolveClassName(env, clazz, get_constant(clazz,thisclass_index));

  memcpy(&clazz->_gc_obj, &gc_obj_template, sizeof(GC_obj));

  if (env) /* japharh doesn't create it's own JNIEnv */
    GC_addReference((JavaVM*)((HungryJNIEnv*)env)->_vm, (GC_obj*)clazz);

  return clazz;
}

ClazzFile *
parse_class(JNIEnv *env, char *full_class_name)
{
  char *buf = NULL;
  int buf_length = 0;
  ClazzFile *result;
  jboolean verbose_loading = JNI_FALSE;

  if (env && ((HungryJNIEnv*)env)->_vm->_args.enableVerboseClassLoading)
    verbose_loading = JNI_TRUE;

  buf = load_class_file(full_class_name, &buf_length);

  if (buf == NULL
      || buf_length == 0)
    {
      return NULL;
    }

  if (verbose_loading)
    {
      printf("[Loading %s... ", full_class_name);
      fflush(stdout);
    }

  result = define_class(env, buf, buf_length);

  if (verbose_loading)
    printf("done]\n");

  return result;
}

void
class_finalize(GC_obj *obj)
{
  int i;
  ClazzFile *clazz = (ClazzFile*)obj;
  JNIEnv *env = THREAD_getEnv();
  printf ("Finalizing class (%s)\n", getClassName(env, clazz));

  delete_class_from_repository(env, getClassName(env, clazz));

  /* free the method structs */
  for (i = 0; i < clazz->num_methods; i ++)
    {
      free(clazz->methods[i]);
    }
  if (clazz->methods)
    free(clazz->methods);

  /* free the field structs */
  for (i = 0; i < clazz->num_fields; i ++)
    {
      free(clazz->fields[i]);
    }
  if (clazz->fields)
    free(clazz->fields);

  /* free our interfaces. */
  for (i = 0; i < clazz->num_interfaces; i ++)
    {
      free(clazz->interfaces[i]);
    }
  if (clazz->innerclass_indices)
    free(clazz->innerclass_indices);
  if (clazz->interface_indices)
    free(clazz->interface_indices);
  if (clazz->interfaces)
    free(clazz->interfaces);

  /* free our constants. */
  for (i = 0; i < clazz->constant_pool_count; i ++)
    {
    }
  free(clazz->constants);

  if (clazz->static_fields)
    free(clazz->static_fields);
  if (clazz->class_name)
    free(clazz->class_name);
  if (clazz->source_filename)
    free(clazz->source_filename);

  free(clazz);
}

ClazzFile *
ExceptionBlock_getHandlerClazz(JNIEnv *env, ClazzFile *clazz, ExceptionBlock *block)
{
  if (block->catch_clazz == NULL && block->catch_clazz_index > 0)
    {
      block->catch_clazz = ResolveClass(env, clazz, get_constant(clazz,block->catch_clazz_index));
    }

  return block->catch_clazz;
}

char *
ResolveClassName(JNIEnv *env, ClazzFile *clazz, ConstantPoolEntry *constant)
{
  assert(constant->generic.tag & CONSTANT_Class);

  if (!(constant->generic.tag & CONSTANT_RESOLVED))
    {
      u2 name_index = constant->clazz_info.name_index;
      
      constant->generic.tag |= CONSTANT_RESOLVED;
      constant->res_clazz_info.name = ResolveUtf8(env, clazz, get_constant(clazz,name_index));
      constant->res_clazz_info.clazz = NULL;
    }
  
  return constant->res_clazz_info.name;
}

ClazzFile *
ResolveClass(JNIEnv *env, ClazzFile *clazz, ConstantPoolEntry *constant)
{
  assert(constant->generic.tag & CONSTANT_Class);

  if (!(constant->generic.tag & CONSTANT_RESOLVED))
    {
      u2 name_index = constant->clazz_info.name_index;

      constant->generic.tag |= CONSTANT_RESOLVED;
      constant->res_clazz_info.name = ResolveUtf8(env, clazz, get_constant(clazz,name_index));
    }

  if (constant->res_clazz_info.clazz == NULL)
    {
      constant->res_clazz_info.clazz =
	(*env)->FindClass(env, constant->res_clazz_info.name);
    }

  return constant->res_clazz_info.clazz;
}

char *
ResolveUtf8(JNIEnv *env, ClazzFile *clazz, ConstantPoolEntry *constant)
{
#if 1
  assert(constant->generic.tag & CONSTANT_Utf8);
  if (constant->generic.tag & CONSTANT_RESOLVED)
    return ((ResolvedUtf8Constant*)constant)->value;
  else
    {
      char *value;

      value = (char*)malloc(constant->utf8_info.length + 1);
      memcpy(value, constant->utf8_info.bytes, constant->utf8_info.length);
      value[constant->utf8_info.length] = 0;
      
      free(constant->utf8_info.bytes);

      constant->res_utf8_info.tag |= CONSTANT_RESOLVED;
      constant->res_utf8_info.value = value;

      return value;
    }
#else
  char *value;

  value = (char*)malloc(constant->utf8_info.length + 1);
  memcpy(value, constant->utf8_info.bytes, constant->utf8_info.length);
  value[constant->utf8_info.length] = 0;
      
  return value;
#endif
}

MethodStruct *
ResolveNonVirtualMethodRef(JNIEnv *env, ClazzFile *clazz, ConstantPoolEntry *entry)
{
  assert(entry->generic.tag & CONSTANT_Methodref);

  if (!(entry->generic.tag & CONSTANT_RESOLVED)
      || entry->methodref_info.nonvirtual_method == NULL)
    {
      int i;
      MethodStruct *method = NULL;
      ClazzFile *method_clazz;
      char *method_name;
      Signature *sig;

      if (entry->generic.tag & CONSTANT_RESOLVED)
	{
	  method_clazz = entry->res_method_info.clazz;
	  method_name = entry->res_method_info.name;
	  sig = entry->res_method_info.sig;
	}
      else
	{
	  char *method_sig;
		  
	  method_clazz = ResolveClass(env, clazz, get_constant(clazz,entry->methodref_info.class_index));
	  method_name = ResolveUtf8(env, clazz, 
				    get_constant(clazz,
						 get_constant(clazz,
							      entry->methodref_info.name_and_type_index)->nameandtype_info.name_index));
	  method_sig = ResolveUtf8(env, clazz,
				   get_constant(clazz,
						get_constant(clazz,entry->methodref_info.name_and_type_index)->nameandtype_info.signature_index));

	  sig = SIG_parseFromJavaSig(env, method_sig);

	  entry->res_method_info.clazz = method_clazz;
	  entry->res_method_info.name = method_name;
	  entry->res_method_info.sig = sig;
	}

      for (i = 0; i < method_clazz->num_methods; i ++)
	{
	  if (!strcmp(method_clazz->methods[i]->name, method_name)
	      && SIG_equal(env, method_clazz->methods[i]->sig, sig))
	    method = method_clazz->methods[i];
	}

      assert(method != NULL);

      entry->generic.tag |= CONSTANT_RESOLVED;
      entry->res_method_info.nonvirtual_method = method;
    }

  return entry->res_method_info.nonvirtual_method;
}

MethodStruct *
ResolveStaticMethodRef(JNIEnv *env, ClazzFile *clazz, ConstantPoolEntry *entry)
{
  assert(entry->generic.tag & CONSTANT_Methodref);

  if (!(entry->generic.tag & CONSTANT_RESOLVED)
      || entry->res_method_info.static_method == NULL)
    {
      int i;
      MethodStruct *method = NULL;
      char *method_name;
      Signature *sig;
      ClazzFile *method_clazz;

      if (entry->generic.tag & CONSTANT_RESOLVED)
	{
	  method_clazz = entry->res_method_info.clazz;
	  method_name = entry->res_method_info.name;
	  sig = entry->res_method_info.sig;
	}
      else
	{
	  char *method_sig;

	  method_clazz = ResolveClass(env, clazz,
				      get_constant(clazz,entry->methodref_info.class_index));

	  method_name = ResolveUtf8(env, clazz, 
				    get_constant(clazz,
						 get_constant(clazz,
							      entry->methodref_info.name_and_type_index)->nameandtype_info.name_index));
	  method_sig = ResolveUtf8(env, clazz,
				   get_constant(clazz,
						get_constant(clazz,entry->methodref_info.name_and_type_index)->nameandtype_info.signature_index));
	  sig = SIG_parseFromJavaSig(env, method_sig);

	  entry->res_method_info.clazz = method_clazz;
	  entry->res_method_info.name = method_name;
	  entry->res_method_info.sig = sig;
	}

      for (i = 0; i < method_clazz->num_methods; i ++)
	{
	  if (!strcmp(method_clazz->methods[i]->name, method_name)
	      && SIG_equal(env, method_clazz->methods[i]->sig, sig)
	      && method_clazz->methods[i]->access_flags & ACC_STATIC)
	    method = method_clazz->methods[i];
	}

      assert(method != NULL);

      entry->generic.tag |= CONSTANT_RESOLVED;
      entry->res_method_info.static_method = method;
    }

  return entry->res_method_info.static_method;
}

MethodStruct *
GetMethodByNameAndSig(JNIEnv *env, ClazzFile *clazz, char *method_name, Signature *sig)
{
  int i;
  MethodStruct *result = NULL;

  for (i = 0; i < clazz->num_methods; i ++)
    if (!strcmp(clazz->methods[i]->name, method_name)
	&& SIG_equal(env, clazz->methods[i]->sig, sig))
      {
	result = clazz->methods[i];
	break;
      }

  return result;
}

int
GetMethodByName(JNIEnv *env, ClazzFile *clazz, char *method_name, MethodStruct ***methods)
{
  int num_found = 0;
  int i;

  for (i = 0; i < clazz->num_methods; i ++)
    if (!strcmp(clazz->methods[i]->name, method_name))
      {
	if (methods)
	  (*methods[num_found]) = clazz->methods[i];
		
	num_found ++;
      }

  return num_found;
}

void
ResolveMethod(JNIEnv *env, ClazzFile *clazz, ConstantPoolEntry *constant,
	      char **name, Signature **sig)
{
  assert(constant->generic.tag & CONSTANT_Methodref);
  if (constant->generic.tag & CONSTANT_RESOLVED)
    {
      *sig = constant->res_method_info.sig;
      *name = constant->res_method_info.name;
    }
  else
    {
      char *sig_str = ResolveUtf8(env, clazz,
				  get_constant(clazz,
					       get_constant(clazz,
							    constant->methodref_info.name_and_type_index)->nameandtype_info.signature_index));
      char *name_str = ResolveUtf8(env, clazz,
				   get_constant(clazz,
						get_constant(clazz,
							     constant->methodref_info.name_and_type_index)->nameandtype_info.name_index));

      constant->res_method_info.tag |= CONSTANT_RESOLVED;
      constant->res_method_info.sig = SIG_parseFromJavaSig(env, sig_str);
      constant->res_method_info.name = name_str;

      *sig = constant->res_method_info.sig;
      *name = constant->res_method_info.name;
    }
}

void
ResolveInterfaceMethod(JNIEnv *env, ClazzFile *clazz, ConstantPoolEntry *constant,
		       char **name, Signature **sig)
{
  assert(constant->generic.tag & CONSTANT_Methodref);
  if (constant->generic.tag & CONSTANT_RESOLVED)
    {
      *sig = constant->res_interfacemethodref_info.sig;
      *name = constant->res_interfacemethodref_info.name;
    }
  else
    {
      char *sig_str = ResolveUtf8(env, clazz,
				  get_constant(clazz,
					       get_constant(clazz,
							    constant->interfacemethodref_info.name_and_type_index)->nameandtype_info.signature_index));
      char *name_str = ResolveUtf8(env, clazz,
				   get_constant(clazz,
						get_constant(clazz,
							     constant->interfacemethodref_info.name_and_type_index)->nameandtype_info.name_index));

      constant->res_interfacemethodref_info.tag |= CONSTANT_RESOLVED;
      constant->res_interfacemethodref_info.sig = SIG_parseFromJavaSig(env, sig_str);
      constant->res_interfacemethodref_info.name = name_str;

      *sig = constant->res_interfacemethodref_info.sig;
      *name = constant->res_interfacemethodref_info.name;
    }
}

FieldStruct *
ResolveFieldRef(JNIEnv *env, ClazzFile *clazz, ConstantPoolEntry *entry)
{
  assert(entry->generic.tag & CONSTANT_Fieldref);

  if (!(entry->generic.tag & CONSTANT_RESOLVED))
    {
      int i;
      FieldStruct *field = NULL;
      ClazzFile *field_clazz = ResolveClass(env, clazz,
					    get_constant(clazz, entry->fieldref_info.class_index));
      char *field_name = ResolveUtf8(env, clazz,
				     get_constant(clazz, get_constant(clazz, entry->fieldref_info.name_and_type_index)->nameandtype_info.name_index));
      char *field_sig = ResolveUtf8(env, clazz,
				    get_constant(clazz, get_constant(clazz, entry->fieldref_info.name_and_type_index)->nameandtype_info.signature_index));
      Signature *sig = SIG_parseFromJavaSig(env, field_sig);

      for (i = 0; i < field_clazz->num_fields; i ++)
	{
	  if (!strcmp(field_clazz->fields[i]->name, field_name)
	      && SIG_equal(env, field_clazz->fields[i]->sig, sig))
	    field = field_clazz->fields[i];
	}

      SIG_free(env, sig);

      assert(field != NULL);

      entry->generic.tag |= CONSTANT_RESOLVED;
      entry->res_fieldref_info.field = field;
    }

  return entry->res_fieldref_info.field;
}

void *
ResolveString(JNIEnv *env, ClazzFile *clazz, ConstantPoolEntry *constant)
{
  assert(constant->generic.tag & CONSTANT_String);
  
  if (!(constant->generic.tag & CONSTANT_RESOLVED))
    {
      char *text;
	  
      text = ResolveUtf8(env, clazz, get_constant(clazz, constant->string_info.string_index));

      constant->generic.tag |= CONSTANT_RESOLVED;
      constant->res_string_info.string = (void*)(*env)->NewStringUTF(env, text);
      constant->res_string_info.string = (void*)(*env)->NewGlobalRef(env, constant->res_string_info.string);
    }

  return constant->res_string_info.string;
}

char *
ResolveStringAsCString(JNIEnv *env, ClazzFile *clazz, ConstantPoolEntry *constant)
{
  int tag = constant->generic.tag;

  if (tag & CONSTANT_Unicode)
    {
      fprintf (stderr, "getting unicode string as C string... do you *really* want to do this?\n");
      abort();
    }
  else if (tag & CONSTANT_Utf8)
    {
      return ResolveUtf8(env, clazz, constant);
    }
  else
    assert(0);

  return 0;
}

char *
getClassName(JNIEnv *env, ClazzFile *clazz)
{
  return clazz->class_name;
}

char *
getSuperClassName(JNIEnv *env, ClazzFile *clazz)
{
  if (clazz->superclass_index == 0)
    return NULL;
  else
    return ResolveClassName(env, clazz, get_constant(clazz, clazz->superclass_index));
}

ClazzFile *
getSuperClass(JNIEnv *env, ClazzFile *clazz)
{
  if (clazz->superclass_index == 0)
    return NULL;
  else
    return ResolveClass(env, clazz, get_constant(clazz, clazz->superclass_index));
}

ClazzFile *
getInterface(JNIEnv *env, ClazzFile *clazz, int interface_index)
{
  if (clazz->interfaces[interface_index] == NULL)
    {
      clazz->interfaces[interface_index] =
	ResolveClass(env, clazz, get_constant(clazz, clazz->interface_indices[interface_index]));
    }

  return clazz->interfaces[interface_index];
}
