/* -*- Mode: C; c-file-style: "gnu" -*-
   interpfunc.c -- opcode functions.
   Created: Chris Toshok <toshok@hungry.com>
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "config.h"
#include "arrays.h"
#include "array-class.h"
#include "objects.h"
#include "interp.h"
#include "log.h"
#include "sig.h"
#include "exceptions.h"
#include "native-threads.h"
#include "ClazzFile.h"
#include "jniint.h"

#include <assert.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#ifndef HAVE_FMODF
static float fmodf(float a, float b) {
  return fmod(a, b);
}
#endif

#define MYLOG "Interp"

int wide_vindex_set = JNI_FALSE; /* XXXX this should be HungryJNIEnv specific. */

void
op_nop (StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_nop ()\n");
}

void
op_wide (StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_wide ()\n");

  wide_vindex_set = JNI_TRUE;
}

void
op_breakpoint(StackFrame *f)
{
  /* ok.  For this we want to do something that is catchable from the
     debugger, and also halts execution in the normal interpreter.
     since this opcode is only valid as a debugger type thing anyway.
     so we raise a signal. */

  raise(SIGILL);
}

void op_bipush(StackFrame *f)
{
  byte byte1;
  jint value;

  JAVARLOG0(MYLOG, 1, "in op_bipush()\n");

  /*
   * push a one-byte signed integer.
   *
   * byte1 is terpreted as a signed 8 bit value.  This value is expanded
   * to an integer and pushed onto the operand stack.
   *
   */

  byte1 = get_next_byte(f);

  value = (jint)byte1;
  
  op_stack_push_int(ENV(f), OPSTACK(f), value);
}

void op_sipush(StackFrame *f)
{
  short value;
  jint item;

  JAVARLOG0(MYLOG, 1, "in op_sipush()\n");

  /*
   * push two byte signed integer
   *
   * byte1 and byte2 are assembled into a signed 16 bit value.  
   * This value is expanded to an integer and pushed onto the operand
   * stack.
   */

  value = get_next_byte(f);
  value <<= 8;
  value |= get_next_byte(f);

  item = (jint)value;

  op_stack_push_int(ENV(f), OPSTACK(f), item);
}


void 
push_item_from_constant_pool (StackFrame *f, 
			      int index)
{
  ConstantPoolEntry constant = METHOD(f)->clazz->constants[ index ];
  int resolved = constant.generic.tag & CONSTANT_RESOLVED;
  int tag;

  if (resolved) 
    tag = constant.generic.tag ^ CONSTANT_RESOLVED;
  else
    tag = constant.generic.tag;

  JAVARLOG0(MYLOG, 1, "in push_item_from_constant_pool()");
  JAVARLOG1(MYLOG, 2, "    pushing item %d from constant pool (", index);
  
  switch (tag)
    {
    case CONSTANT_Utf8:
      JAVARLOG0(MYLOG, 2, "Utf8 FINISH ME)\n");
      abort();
      break;
    case CONSTANT_Unicode:
      JAVARLOG0(MYLOG, 2, "Unicode FINISH ME)\n");
      abort();
      break;
    case CONSTANT_Integer:
      JAVARLOG0(MYLOG, 2, "Integer)\n");
      op_stack_push_int(ENV(f), OPSTACK(f), constant.integer_info.bytes);
      break;
    case CONSTANT_Float:
      {
	float fvalue;
	JAVARLOG0(MYLOG, 2, "Float)\n");

	fvalue = *(float*)((void*)&constant.float_info.bytes);
	op_stack_push_float(ENV(f), OPSTACK(f), fvalue);
	break;
      }
    case CONSTANT_Long:
      JAVARLOG0(MYLOG, 2, "Long)\n");
      op_stack_push_long_w(ENV(f), OPSTACK(f), constant.long_info.high_bytes, constant.long_info.low_bytes);
      break;
    case CONSTANT_Double:
      {
	JAVARLOG0(MYLOG, 2, "Double)\n");

	op_stack_push_double_w(ENV(f), OPSTACK(f), constant.double_info.high_bytes, constant.double_info.low_bytes);
	break;
      }
    case CONSTANT_String:
      JAVARLOG0(MYLOG, 2, "String)\n");
      op_stack_push_object (ENV(f), OPSTACK(f), ResolveString(ENV(f), METHOD(f)->clazz, &METHOD(f)->clazz->constants[index]));
      break;
    case CONSTANT_Class:
    case CONSTANT_Fieldref:
    case CONSTANT_Methodref:
    case CONSTANT_InterfaceMethodref:
    case CONSTANT_NameAndType:
    default:
      printf ("pushing invalid Constant Pool type: %d\n", constant.generic.tag);
      throw_Exception(ENV(f), "java/lang/RuntimeException", NULL);
      return;
    }
}

void 
op_ldc(StackFrame *f)
{
  u1 indexbyte1;
  u2 index;

  JAVARLOG0(MYLOG, 1, "in op_ldc()\n");

  indexbyte1 = (u1)get_next_byte(f);
  index = (u2)indexbyte1;

  push_item_from_constant_pool (f, index);
}

void
op_ldc_w(StackFrame *f)
{
  u2 index;

  JAVARLOG0(MYLOG, 1, "in op_ldc_w()\n");

  index = (u1)get_next_byte(f);
  index <<= 8;
  index |= (u1)get_next_byte(f);

  push_item_from_constant_pool (f, index);
}

void 
op_ldc2w(StackFrame *f)
{
  u2 index;

  JAVARLOG0(MYLOG, 1, "in op_ldc2w()\n");
  
  index = (u1)get_next_byte(f);
  index <<= 8;
  index |= (u1)get_next_byte(f);

  push_item_from_constant_pool (f, index);
}

void 
op_aconst_null(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_aconst_null()\n");

  op_stack_push_object(ENV(f), OPSTACK(f), 0);
}

void 
op_iconst_m1(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_iconst_m1()\n");

  op_stack_push_int(ENV(f), OPSTACK(f), -1);
}

void 
op_iconst_0(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_iconst_0()\n");

  op_stack_push_int(ENV(f), OPSTACK(f), 0);
}

void 
op_iconst_1(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_iconst_1()\n");

  op_stack_push_int(ENV(f), OPSTACK(f), 1);
}

void 
op_iconst_2(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_iconst_2()\n");

  op_stack_push_int(ENV(f), OPSTACK(f), 2);
}

void 
op_iconst_3(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_iconst_3()\n");

  op_stack_push_int(ENV(f), OPSTACK(f), 3);
}

void op_iconst_4(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_iconst_4()\n");

  op_stack_push_int(ENV(f), OPSTACK(f), 4);
}

void op_iconst_5(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_iconst_5()\n");

  op_stack_push_int(ENV(f), OPSTACK(f), 5);
}

void op_lconst_0(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_lconst_0()\n");

  op_stack_push_long(ENV(f), OPSTACK(f), 0);
}

void 
op_lconst_1(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_lconst_1()\n");

  op_stack_push_long(ENV(f), OPSTACK(f), 1);
}

void 
op_fconst_0(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_fconst_0()\n");

  op_stack_push_float(ENV(f), OPSTACK(f), 0.0);
}

void 
op_fconst_1(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_fconst_1()\n");

  op_stack_push_float(ENV(f), OPSTACK(f), 1.0);
}

void 
op_fconst_2(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_fconst_2()\n");

  op_stack_push_float(ENV(f), OPSTACK(f), 2.0);
}

void 
op_dconst_0(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_dconst_0()\n");

  op_stack_push_double(ENV(f), OPSTACK(f), 0.0);
}

void 
op_dconst_1(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_dconst_1()\n");

  op_stack_push_double(ENV(f), OPSTACK(f), 1.0);
}

void op_iload(StackFrame *f)
{
  u2 vindex;

  JAVARLOG1(MYLOG, 1, "in op_iload() wide=%d\n", wide_vindex_set);
  
  vindex = get_next_byte(f);
  if (wide_vindex_set)
    {
      u1 vindex2 = get_next_byte (f);
      vindex = (vindex << 8) | vindex2;
      wide_vindex_set = JNI_FALSE;
    }

  op_stack_push_int(ENV(f), OPSTACK(f), promote_to_int(ENV(f), VARS(f)[vindex]));
}

void op_iload_0(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_iload_0()\n");

  op_stack_push_int(ENV(f), OPSTACK(f), promote_to_int(ENV(f), VARS(f)[0]));
}

void op_iload_1(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_iload_1()\n");

  op_stack_push_int(ENV(f), OPSTACK(f), promote_to_int(ENV(f), VARS(f)[1]));
}

void op_iload_2(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_iload_2()\n");

  op_stack_push_int(ENV(f), OPSTACK(f), promote_to_int(ENV(f), VARS(f)[2]));
}

void op_iload_3(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_iload_3()\n");

  op_stack_push_int(ENV(f), OPSTACK(f), promote_to_int(ENV(f), VARS(f)[3]));
}

void op_lload(StackFrame *f)
{
  u2 vindex;
  jint lvalue1, lvalue2;

  JAVARLOG1(MYLOG, 1, "in op_lload() wide=%d\n", wide_vindex_set);

  vindex = get_next_byte(f);
  if (wide_vindex_set)
    {
      u1 vindex2 = get_next_byte (f);
      vindex = (vindex << 8) | vindex2;
      wide_vindex_set = JNI_FALSE;
    }
  
  lvalue1 = VARS(f)[vindex].value.i;
  lvalue2 = VARS(f)[vindex + 1].value.i;

  op_stack_push_long_w(ENV(f), OPSTACK(f), lvalue1, lvalue2);
}

void op_lload_0(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_lload_0()\n");

  op_stack_push_long_w(ENV(f), OPSTACK(f), VARS(f)[0].value.i, VARS(f)[1].value.i);
}

void op_lload_1(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_lload_1()\n");

  op_stack_push_long_w(ENV(f), OPSTACK(f), VARS(f)[1].value.i, VARS(f)[2].value.i);
}

void op_lload_2(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_lload_2()\n");

  op_stack_push_long_w(ENV(f), OPSTACK(f), VARS(f)[2].value.i, VARS(f)[3].value.i);
}

void op_lload_3(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_lload_3()\n");

  op_stack_push_long_w(ENV(f), OPSTACK(f), VARS(f)[3].value.i, VARS(f)[4].value.i);
}

void op_fload(StackFrame *f)
{
  u2 vindex;
  jfloat fvalue;

  JAVARLOG1(MYLOG, 1, "in op_fload() wide=%d\n", wide_vindex_set);

  vindex = get_next_byte(f);
  if (wide_vindex_set)
    {
      u1 vindex2 = get_next_byte (f);
      vindex = (vindex << 8) | vindex2;
      wide_vindex_set = JNI_FALSE;
    }

  fvalue = VARS(f)[vindex].value.f;

  op_stack_push_float(ENV(f), OPSTACK(f), fvalue);
}

void op_fload_0(StackFrame *f)
{
  jfloat fvalue;

  JAVARLOG0(MYLOG, 1, "in op_fload_0()\n");
  
  fvalue = VARS(f)[0].value.f;

  op_stack_push_float(ENV(f), OPSTACK(f), fvalue);
}

void op_fload_1(StackFrame *f)
{
  jfloat fvalue;
  
  JAVARLOG0(MYLOG, 1, "in op_fload_1()\n");

  fvalue = VARS(f)[1].value.f;

  op_stack_push_float(ENV(f), OPSTACK(f), fvalue);
}

void op_fload_2(StackFrame *f)
{
  jfloat fvalue;

  JAVARLOG0(MYLOG, 1, "in op_fload_2()\n");

  fvalue = VARS(f)[2].value.f;

  op_stack_push_float(ENV(f), OPSTACK(f), fvalue);
}

void op_fload_3(StackFrame *f)
{
  jfloat fvalue;

  JAVARLOG0(MYLOG, 1, "in op_fload_2()\n");

  fvalue = VARS(f)[3].value.f;

  op_stack_push_float(ENV(f), OPSTACK(f), fvalue);
}

void op_dload(StackFrame *f)
{
  u2 vindex;
  jdouble fvalue1, fvalue2;

  JAVARLOG1(MYLOG, 1, "in op_dload() wide=%d\n", wide_vindex_set);

  vindex = get_next_byte(f);
  if (wide_vindex_set)
    {
      u1 vindex2 = get_next_byte (f);
      vindex = (vindex << 8) | vindex2;
      wide_vindex_set = JNI_FALSE;
    }
  
  fvalue1 = VARS(f)[vindex].value.i;
  fvalue2 = VARS(f)[vindex + 1].value.i;

  op_stack_push_double_w(ENV(f), OPSTACK(f), fvalue1, fvalue2);
}

void op_dload_0(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_dload_0()\n");

  op_stack_push_double_w(ENV(f), OPSTACK(f), VARS(f)[0].value.i, VARS(f)[1].value.i);
}

void op_dload_1(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_dload_1()\n");

  op_stack_push_double_w(ENV(f), OPSTACK(f), VARS(f)[1].value.i, VARS(f)[2].value.i);
}

void op_dload_2(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_dload_2()\n");

  op_stack_push_double_w(ENV(f), OPSTACK(f), VARS(f)[2].value.i, VARS(f)[3].value.i);
}

void op_dload_3(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_dload_3()\n");

  op_stack_push_double_w(ENV(f), OPSTACK(f), VARS(f)[3].value.i, VARS(f)[4].value.i);
}

void op_aload(StackFrame *f)
{
  u2 vindex;

  JAVARLOG1(MYLOG, 1, "in op_aload() wide=%d\n", wide_vindex_set);

  /**
   * load object reference from local variable
   **/
  /**
   * The value of the local variable at vindex in the current Java frame
   * is pushed onto the operand stack.
   */
  vindex = get_next_byte(f);
  if (wide_vindex_set)
    {
      u1 vindex2 = get_next_byte (f);
      vindex = (vindex << 8) | vindex2;
      wide_vindex_set = JNI_FALSE;
    }

  op_stack_push_object(ENV(f), OPSTACK(f), VARS(f)[vindex].value.l);
}

void op_aload_0(StackFrame *f)
{
  /**
   * load object reference from local variable
   **/
  JAVARLOG0(MYLOG, 1, "in op_aload_0()\n");
  /**
   * The value of the local variable at 0 in the current java frame
   * is pushed onto the operand stack.
   * This instruction is the same as aload with a vindex of 0, except
   * that the operand 0 is implicit.
   */

  op_stack_push_object(ENV(f), OPSTACK(f), VARS(f)[0].value.l);
}

void op_aload_1(StackFrame *f)
{
  /**
   * load object reference from local variable
   **/
  JAVARLOG0(MYLOG, 1, "in op_aload_1()\n");
  /**
   * The value of the local variable at 1 in the current java frame
   * is pushed onto the operand stack.
   * This instruction is the same as aload with a vindex of 1, except
   * that the operand 1 is implicit.
   */

  op_stack_push_object(ENV(f), OPSTACK(f), VARS(f)[1].value.l);
}

void op_aload_2(StackFrame *f)
{
  /**
   * load object reference from local variable
   **/
  JAVARLOG0(MYLOG, 1, "in op_aload_2()\n");
  /**
   * The value of the local variable at 2 in the current java frame
   * is pushed onto the operand stack.
   * This instruction is the same as aload with a vindex of 2, except
   * that the operand 2 is implicit.
   */

  op_stack_push_object(ENV(f), OPSTACK(f), VARS(f)[2].value.l);
}

void op_aload_3(StackFrame *f)
{
  /**
   * load object reference from local variable
   **/
  JAVARLOG0(MYLOG, 1, "in op_aload_3()\n");
  /**
   * The value of the local variable at 3 in the current java frame
   * is pushed onto the operand stack.
   * This instruction is the same as aload with a vindex of 3, except
   * that the operand 3 is implicit.
   */

  op_stack_push_object(ENV(f), OPSTACK(f), VARS(f)[3].value.l);
}

void op_ishl(StackFrame *f)
{
  jint value1;
  jint value2;

  JAVARLOG0(MYLOG, 1, "in op_ishl()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int (ENV(f), OPSTACK(f), &value1);  

  value1 = value1 << (value2 & 31); /* value2 & low 5 bits */

  op_stack_push_int(ENV(f), OPSTACK(f), value1);
}

void op_ishr(StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_ishr()\n");
  
  op_stack_pop_int (ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int (ENV(f), OPSTACK(f), &value1);  

  value1 = value1 >> (value2 & 31); /* value2 & low 5 bits */

  /* sign extension ???
     leave the top bit alone, do the shift, and then put the sign back */

  op_stack_push_int(ENV(f), OPSTACK(f), value1);
}


void op_iushr(StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_iushr()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int (ENV(f), OPSTACK(f), &value1);
  
  value1 = value1 >> (value2 & 31); /* value2 & low 5 bits */

  /* sign extension ???  (no sign extension) */

  op_stack_push_int(ENV(f), OPSTACK(f), value1);
}

void op_iand(StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_iand()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int (ENV(f), OPSTACK(f), &value1);  

  op_stack_push_int(ENV(f), OPSTACK(f), value1 & value2);
}

void op_land(StackFrame *f)
{
  jlong value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_land()\n");

  op_stack_pop_long (ENV(f), OPSTACK(f), &value2);
  op_stack_pop_long (ENV(f), OPSTACK(f), &value1);

  op_stack_push_long(ENV(f), OPSTACK(f), value1 & value2);
}

void op_ior(StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_ior()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int (ENV(f), OPSTACK(f), &value1);  

  op_stack_push_int(ENV(f), OPSTACK(f), value1 | value2); 
}

void op_lor(StackFrame *f)
{
  jlong value1, value2, result;

  JAVARLOG0(MYLOG, 1, "in op_lor()\n");

  op_stack_pop_long (ENV(f), OPSTACK(f), &value1);

  op_stack_pop_long (ENV(f), OPSTACK(f), &value2);

  result = value1 | value2;

  op_stack_push_long (ENV(f), OPSTACK(f), result);
}


void op_ixor (StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_ixor()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int (ENV(f), OPSTACK(f), &value1);  

  op_stack_push_int (ENV(f), OPSTACK(f), value1 ^ value2); 
}


void op_lxor (StackFrame *f)
{
  jlong value1, value2, result;

  JAVARLOG0(MYLOG, 1, "in op_lxor()\n");

  op_stack_pop_long (ENV(f), OPSTACK(f), &value1);

  op_stack_pop_long (ENV(f), OPSTACK(f), &value2);

  result = value1 ^ value2;

  op_stack_push_long (ENV(f), OPSTACK(f), result);
}


void op_lshl (StackFrame *f)
{
  jlong value1, result;
  jint value2;

  JAVARLOG0(MYLOG, 1, "in op_ishl()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &value2);
  op_stack_pop_long (ENV(f), OPSTACK(f), &value1);

  result = value1 << (value2 & 63); /* value2 & low 6 bits */

  op_stack_push_long(ENV(f), OPSTACK(f), result);
}

void op_lshr (StackFrame *f)
{
  jlong value1, result;
  jint value2;

  JAVARLOG0(MYLOG, 1, "in op_lshr()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &value2);
  op_stack_pop_long (ENV(f), OPSTACK(f), &value1);

  result = value1 >> (value2 & 63); /* value2 & low 6 bits */

  op_stack_push_long(ENV(f), OPSTACK(f), result);
}

void op_lushr (StackFrame *f)
{
  jlong value1, result;
  jint value2;

  JAVARLOG0(MYLOG, 1, "in op_lushr()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &value2);
  op_stack_pop_long (ENV(f), OPSTACK(f), &value1);

  result = value1 >> (value2 & 63); /* value2 & low 6 bits */

  /* sign extension ???  (no sign extension) */

  op_stack_push_long(ENV(f), OPSTACK(f), result);
}

void 
op_istore (StackFrame *f)
{
  u2 vindex;
  jint value;

  JAVARLOG1(MYLOG, 1, "in op_istore () wide=%d\n", wide_vindex_set);
  
  vindex = get_next_byte (f);
  if (wide_vindex_set)
    {
      u1 vindex2 = get_next_byte (f);
      vindex = (vindex << 8) | vindex2;
      wide_vindex_set = JNI_FALSE;
    }

  op_stack_pop_int (ENV(f), OPSTACK(f), &value);
  VARS (f)[ vindex ].value.i = value;
  VARS (f)[ vindex ].tag = SIG_JINT;
}

void 
op_lstore (StackFrame *f)
{
  u2 vindex;
  jint value1, value2;

  JAVARLOG1(MYLOG, 1, "in op_lstore () wide=%d\n", wide_vindex_set);

  vindex = get_next_byte (f);
  if (wide_vindex_set)
    {
      u1 vindex2 = get_next_byte (f);
      vindex = (vindex << 8) | vindex2;
      wide_vindex_set = JNI_FALSE;
    }

  op_stack_pop_long_w (ENV(f), OPSTACK(f), &value1, &value2);

  VARS (f)[ vindex ].value.i = value1;
  VARS (f)[ vindex ].tag = SIG_JLONG1;
  VARS (f)[ vindex+1 ].value.i = value2;
  VARS (f)[ vindex+1 ].tag = SIG_JLONG2;
}

void 
op_lstore_0 (StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_lstore_0 ()\n");

  op_stack_pop_long_w (ENV(f), OPSTACK(f), &value1, &value2);

  VARS (f)[ 0 ].value.i = value1;
  VARS (f)[ 0 ].tag = SIG_JLONG1;
  VARS (f)[ 1 ].value.i = value2;
  VARS (f)[ 1 ].tag = SIG_JLONG2;
}

void 
op_lstore_1 (StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_lstore_1 ()\n");

  op_stack_pop_long_w (ENV(f), OPSTACK(f), &value1, &value2);

  VARS (f)[ 1 ].value.i = value1;
  VARS (f)[ 1 ].tag = SIG_JLONG1;
  VARS (f)[ 2 ].value.i = value2;
  VARS (f)[ 2 ].tag = SIG_JLONG2;
}

void 
op_lstore_2 (StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_lstore_2 ()\n");

  op_stack_pop_long_w (ENV(f), OPSTACK(f), &value1, &value2);

  VARS (f)[ 2 ].value.i = value1;
  VARS (f)[ 2 ].tag = SIG_JLONG1;
  VARS (f)[ 3 ].value.i = value2;
  VARS (f)[ 3 ].tag = SIG_JLONG2;
}

void 
op_lstore_3 (StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_lstore_3 ()\n");

  op_stack_pop_long_w (ENV(f), OPSTACK(f), &value1, &value2);

  VARS (f)[ 3 ].value.i = value1;
  VARS (f)[ 3 ].tag = SIG_JLONG1;
  VARS (f)[ 4 ].value.i = value2;
  VARS (f)[ 4 ].tag = SIG_JLONG2;
}

void 
op_fstore (StackFrame *f)
{
  u2 vindex;
  /*double value;*/
  jfloat value;

  JAVARLOG1(MYLOG, 1, "in op_fstore () wide=%d\n", wide_vindex_set);

  vindex = get_next_byte (f);
  if (wide_vindex_set)
    {
      u1 vindex2 = get_next_byte (f);
      vindex = (vindex << 8) | vindex2;
      wide_vindex_set = JNI_FALSE;
    }

  op_stack_pop_float (ENV(f), OPSTACK(f), &value);
  VARS (f)[ vindex ].value.f = value;
  VARS (f)[ vindex ] .tag = SIG_JFLOAT;
}

void 
op_fstore_0 (StackFrame *f)
{
  jfloat value;

  JAVARLOG0(MYLOG, 1, "in op_fstore_0 ()\n");

  op_stack_pop_float (ENV(f), OPSTACK(f), &value);

  VARS (f)[ 0 ].value.f = value;
  VARS (f)[ 0 ].tag = SIG_JFLOAT;
}

void 
op_fstore_1 (StackFrame *f)
{
  jfloat value;

  JAVARLOG0(MYLOG, 1, "in op_fstore_1 ()\n");

  op_stack_pop_float (ENV(f), OPSTACK(f), &value);

  VARS (f)[ 1 ].value.f = value;
  VARS (f)[ 1 ].tag = SIG_JFLOAT;
}

void 
op_fstore_2 (StackFrame *f)
{
  jfloat value;

  JAVARLOG0(MYLOG, 1, "in op_fstore_2 ()\n");

  op_stack_pop_float (ENV(f), OPSTACK(f), &value);

  VARS (f)[ 2 ].value.f = value;
  VARS (f)[ 2 ].tag = SIG_JFLOAT;
}

void 
op_fstore_3 (StackFrame *f)
{
  jfloat value;

  JAVARLOG0(MYLOG, 1, "in op_fstore_3 ()\n");

  op_stack_pop_float (ENV(f), OPSTACK(f), &value);

  VARS (f)[ 3 ].value.f = value;
  VARS (f)[ 3 ].tag = SIG_JFLOAT;
}

void 
op_dstore (StackFrame *f)
{
  u2 vindex;
  /* double value1, value2; */
  jint value1, value2;

  JAVARLOG1(MYLOG, 1, "in op_dstore () wide=%d\n", wide_vindex_set);

  vindex = get_next_byte (f);
  if (wide_vindex_set)
    {
      u1 vindex2 = get_next_byte (f);
      vindex = (vindex << 8) | vindex2;
      wide_vindex_set = JNI_FALSE;
    }

  op_stack_pop_double_w (ENV(f), OPSTACK(f), &value1, &value2);

  VARS (f)[ vindex ].value.i = value1;
  VARS (f)[ vindex ].tag = SIG_JDOUBLE1;
  VARS (f)[ vindex+1 ].value.i = value2;
  VARS (f)[ vindex+1 ].tag = SIG_JDOUBLE2;
}

void 
op_dstore_0 (StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_dstore_0 ()\n");

  op_stack_pop_double_w (ENV(f), OPSTACK(f), &value1, &value2);

  VARS (f)[ 0 ].value.i = value1;
  VARS (f)[ 0 ].tag = SIG_JDOUBLE1;
  VARS (f)[ 1 ].value.i = value2;
  VARS (f)[ 1 ].tag = SIG_JDOUBLE2;
}

void 
op_dstore_1 (StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_dstore_1 ()\n");

  op_stack_pop_double_w (ENV(f), OPSTACK(f), &value1, &value2);

  VARS (f)[ 1 ].value.i = value1;
  VARS (f)[ 1 ].tag = SIG_JDOUBLE1;
  VARS (f)[ 2 ].value.i = value2;
  VARS (f)[ 2 ].tag = SIG_JDOUBLE2;
}

void 
op_dstore_2 (StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_dstore_2 ()\n");

  op_stack_pop_double_w (ENV(f), OPSTACK(f), &value1, &value2);

  VARS (f)[ 2 ].value.i = value1;
  VARS (f)[ 2 ].tag = SIG_JDOUBLE1;
  VARS (f)[ 3 ].value.i = value2;
  VARS (f)[ 3 ].tag = SIG_JDOUBLE2;
}

void 
op_dstore_3 (StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_dstore_3()\n");

  op_stack_pop_double_w (ENV(f), OPSTACK(f), &value1, &value2);

  VARS (f)[ 3 ].value.i = value1;
  VARS (f)[ 3 ].tag = SIG_JDOUBLE1;
  VARS (f)[ 4 ].value.i = value2;
  VARS (f)[ 4 ].tag = SIG_JDOUBLE2;
}

void 
op_astore (StackFrame *f)
{
  u2 vindex;
  jobject value;

  JAVARLOG1(MYLOG, 1, "in op_astore() wide=%d\n", wide_vindex_set);

  vindex = get_next_byte (f);
  if (wide_vindex_set)
    {
      u1 vindex2 = get_next_byte (f);
      vindex = (vindex << 8) | vindex2;
      wide_vindex_set = JNI_FALSE;
    }

  op_stack_pop_object (ENV(f), OPSTACK(f), &value);

  VARS (f)[ vindex ].value.l = value;
  VARS (f)[ vindex ].tag = SIG_JOBJECT;
}

void 
op_istore_0 (StackFrame *f)
{
  jint value;

  JAVARLOG0(MYLOG, 1, "in op_istore_0()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &value);

  VARS (f)[ 0 ].value.i = value;
  VARS (f)[ 0 ].tag = SIG_JINT;
}

void 
op_istore_1 (StackFrame *f)
{
  jint value;

  JAVARLOG0(MYLOG, 1, "in op_istore_1 ()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &value);

  VARS (f)[ 1 ].value.i = value;
  VARS (f)[ 1 ].tag = SIG_JINT;
}

void 
op_istore_2 (StackFrame *f)
{
  jint value;

  JAVARLOG0(MYLOG, 1, "in op_istore_2 ()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &value);

  VARS (f)[ 2 ].value.i = value;
  VARS (f)[ 2 ].tag = SIG_JINT;
}

void 
op_istore_3 (StackFrame *f)
{
  jint value;

  JAVARLOG0(MYLOG, 1, "in op_istore_3 ()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &value);

  VARS (f)[ 3 ].value.i = value;
  VARS (f)[ 3 ].tag = SIG_JINT;
}

void 
op_astore_0 (StackFrame *f)
{
  u4 object_ref;

  JAVARLOG0(MYLOG, 1, "in op_astore_0 ()\n");

  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&object_ref);

  VARS (f)[ 0 ].value.i = object_ref;
  VARS (f)[ 0 ].tag = SIG_JOBJECT;
}

void 
op_astore_1 (StackFrame *f)
{
  jobject object_ref;

  JAVARLOG0(MYLOG, 1, "in op_astore_1 ()\n");

  op_stack_pop_object (ENV(f), OPSTACK(f), &object_ref);

  VARS (f)[ 1 ].value.l = object_ref;
  VARS (f)[ 1 ].tag = SIG_JOBJECT;
}


void 
op_astore_2 (StackFrame *f)
{
  jobject object_ref;

  JAVARLOG0(MYLOG, 1, "in op_astore_2 ()\n");

  op_stack_pop_object (ENV(f), OPSTACK(f), &object_ref);

  VARS (f)[ 2 ].value.l = object_ref;
  VARS (f)[ 2 ].tag = SIG_JOBJECT;
}

void 
op_astore_3 (StackFrame *f)
{
  jobject object_ref;

  JAVARLOG0(MYLOG, 1, "in op_astore_3 ()\n");

  op_stack_pop_object (ENV(f), OPSTACK(f), &object_ref);

  VARS (f)[ 3 ].value.l = object_ref;
  VARS (f)[ 3 ].tag = SIG_JOBJECT;
}

void op_putfield (StackFrame *f)
{
  u2 index;
  JavaStackItem value;
  jobject objectref;
  obj_thunk *casted_object;
  FieldStruct *field;

  JAVARLOG0(MYLOG, 1, "in op_putfield()\n");

  index = get_next_byte(f) << 8;
  index |= get_next_byte(f);
  
  field = ResolveFieldRef(ENV(f), METHOD(f)->clazz, &METHOD(f)->clazz->constants[index]);

  initialize_class(ENV(f), field->clazz);

  if (field->access_flags & ACC_STATIC)
    {
      throw_Exception(ENV(f), "java/lang/IncompatibleClassChangeError", NULL);
      return;
    }

  JAVARLOG1(MYLOG, 2, "        field_name='%s'\n",
	    field->name);

  if (UNRESOLVED ( field->field_index )
      || UNRESOLVED ( field->java_type ))
    {
      throw_Exception(ENV(f), "java/lang/RuntimeException", NULL);
      return;
    }

  switch (field->java_type)
    {
    case SIG_JDOUBLE:
      {
	op_stack_pop_double(ENV(f), OPSTACK(f), &value.value.d);

	value.tag = SIG_JDOUBLE;

	break;
      }
    case SIG_JLONG:
      {
	op_stack_pop_long(ENV(f), OPSTACK(f), &value.value.j);

	value.tag = SIG_JLONG;
	break;
      }
    default:
      op_stack_pop_any (ENV(f), OPSTACK(f), &value);
      break;
    }

  op_stack_pop_object (ENV(f), OPSTACK(f), &objectref);

  if (objectref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  casted_object = find_instance_data_for_class((obj_thunk*)objectref,
					       field->clazz);

  if (!casted_object)
    {
      throw_Exception(ENV(f), "java/lang/IncompatibleClassChangeError", NULL);
      return;
    }

  set_instance_field(casted_object, field, value);
}


void op_getfield(StackFrame *f)
{
  u2 index;
  jobject objectref;
  obj_thunk *casted_object;
  FieldStruct *field;
  JavaStackItem item;

  JAVARLOG0(MYLOG, 1, "in op_getfield()\n");

  index = get_next_byte(f) << 8;
  index |= get_next_byte(f);

  field = ResolveFieldRef(ENV(f), METHOD(f)->clazz, &METHOD(f)->clazz->constants[index]);

  initialize_class(ENV(f), field->clazz);

  JAVARLOG3(MYLOG, 2, "   getting %s.%s(%s)\n", 
	    getClassName(ENV(f), field->clazz), field->name, SIG_formatToJavaSig(ENV(f), field->sig));

  if (field->access_flags & ACC_STATIC)
    {
      throw_Exception(ENV(f), "java/lang/IncompatibleClassChangeError", NULL);
      return;
    }

  op_stack_pop_object (ENV(f), OPSTACK(f), &objectref);

  if (objectref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  casted_object = find_instance_data_for_class ((obj_thunk*) objectref, field->clazz);

  if (!casted_object)
    {
      throw_Exception(ENV(f), "java/lang/IncompatibleClassChangeError", NULL);
      return;
    }

  JAVARLOG2(MYLOG, 2, "        index=%d, objectref=%d\n", index, objectref);

  if ( UNRESOLVED ( field->field_index )
       || UNRESOLVED ( field->java_type ))
    {
      printf ("CALLING GETFIELD ON NON_NEW'ED OBJECT!!!!\n");
      throw_Exception(ENV(f), "java/lang/RuntimeException", NULL);
      return;
    }

  get_instance_field(casted_object, field, &item);
  op_stack_push_any(ENV(f), OPSTACK(f),
		    &item);
}


void op_putstatic(StackFrame *f)
{
  u2 index;
  JavaStackItem value;
  FieldStruct *field;

  JAVARLOG0(MYLOG, 1, "in op_putstatic()\n");

  index = get_next_byte(f) << 8;
  index |= get_next_byte(f);
  
  field = ResolveFieldRef(ENV(f), METHOD(f)->clazz, &METHOD(f)->clazz->constants[index]);
  initialize_class(ENV(f), field->clazz);

  if (!(field->access_flags & ACC_STATIC))
    {
      throw_Exception(ENV(f), "java/lang/IncompatibleClassChangeError", NULL);
      return;
    }

  switch (field->java_type)
    {
    case SIG_JDOUBLE:
      {
	jdouble foo;
	op_stack_pop_double(ENV(f), OPSTACK(f), &foo);

	value.value.d = foo;
	value.tag = SIG_JDOUBLE;

	set_static_field(field->clazz, field, value);
	break;
      }
    case SIG_JLONG:
      {
	jlong foo;
	op_stack_pop_long(ENV(f), OPSTACK(f), &foo);
      
	value.value.j = foo;
	value.tag = SIG_JLONG;
      
	set_static_field(field->clazz, field, value);
	break;
      }
    default:
      op_stack_pop_any (ENV(f), OPSTACK(f), &value);
      set_static_field(field->clazz, field, value);
      break;
    }
}


void 
op_getstatic(StackFrame *f)
{
  u2 index;
  FieldStruct *field;
  JavaStackItem item;

  JAVARLOG0(MYLOG, 1, "in op_getstatic()\n");

  index = get_next_byte(f) << 8;
  index |= get_next_byte(f);

  JAVARLOG1(MYLOG, 2, "        index=%d\n", index);

  field = ResolveFieldRef(ENV(f), METHOD(f)->clazz, &METHOD(f)->clazz->constants[index]);
  initialize_class(ENV(f), field->clazz);

  if (!(field->access_flags & ACC_STATIC))
    {
      throw_Exception(ENV(f), "java/lang/IncompatibleClassChangeError", NULL);
      return;
    }

  if (UNRESOLVED( field->field_index )
      || UNRESOLVED( field->java_type ))
    {
      printf ("CALLING GETSTATIC ON NON_RESOLVED CLASS!!!!\n");
      throw_Exception(ENV(f), "java/lang/RuntimeException", NULL);
      return;
    }

  get_static_field(field->clazz, field, &item);
  op_stack_push_any(ENV(f), OPSTACK(f), &item);
}


void 
op_new (StackFrame *f)
{
  u2 index;
  ClazzFile *cf;
  jobject object_reference;

  JAVARLOG0(MYLOG, 1, "in op_new()\n");

  index = get_next_byte (f) << 8;
  index |= get_next_byte (f);

  cf = ResolveClass(ENV(f), METHOD(f)->clazz, &METHOD(f)->clazz->constants[index]);
  JAVARLOG1(MYLOG, 2, 
	    "name of class to new: '%s'\n", getClassName(ENV(f), cf));
  object_reference = new_object (ENV(f), cf);
  assert(object_reference);
  op_stack_push_object (ENV(f), OPSTACK(f), (void*) object_reference);
  GC_addReference((JavaVM*)((HungryJNIEnv*)ENV(f))->_vm, (GC_obj*)object_reference);
}


void op_checkcast (StackFrame *f)
{
  u2 index;
  obj_thunk *object_ref;
  /*obj_thunk *thunk;*/
  ClazzFile *cf;

  JAVARLOG0(MYLOG, 1, "in op_checkcast()\n");

  index = get_next_byte(f) << 8;
  index |= get_next_byte(f);

  cf = ResolveClass(ENV(f), METHOD(f)->clazz, &METHOD(f)->clazz->constants[index]);

  op_stack_pop_object(ENV(f), OPSTACK(f), (jobject*)&object_ref);

#if 1
  op_stack_push_object(ENV(f), OPSTACK(f), object_ref);
#else  

  thunk = object_ref;

  /* trivial case -- thunk is NULL, which is castable to anything */
  if (!thunk)
    {
      op_stack_push_object(ENV(f), OPSTACK(f), object_ref);
      return;
    }
  else
    {
      /* we need to decrement the object reference 
	 until we hit a fence post. */

      while (!IS_POST(thunk))
	thunk --;

      /* start at the first object instance. */
      thunk ++;

      if (is_instance_of(thunk, cf))
	{
	  op_stack_push_object(ENV(f), OPSTACK(f), object_ref);
	  return;
	}
      else
	{
	  throw_Exception(ENV(f), "java/lang/ClassCastException", NULL);
	  return;
	}
    }
#endif
}

void op_instanceof (StackFrame *f)
{
  u2 index;
  jobject objectref;

  JAVARLOG0(MYLOG, 1, "in op_instanceof()\n");

  index = get_next_byte(f) << 8;
  index |= get_next_byte(f);

  op_stack_pop_object (ENV(f), OPSTACK(f), &objectref);
  if (objectref == 0)
    {
      op_stack_push_int (ENV(f), OPSTACK(f), 0);
    }
  else
    {
      ClazzFile *cf = ResolveClass(ENV(f), METHOD(f)->clazz, &METHOD(f)->clazz->constants[index]);

      op_stack_push_int (ENV(f), OPSTACK(f), 
			 is_instance_of((obj_thunk*)objectref, 
					cf));
    }
}

void 
op_pop(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_pop()\n");

  OPSTACK(f)->stack_top -= 1;
}

void 
op_pop2(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_pop2()\n");

  OPSTACK(f)->stack_top -= 2;
}

void 
op_dup(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_dup()\n");

  op_stack_push_any(ENV(f), OPSTACK(f), OPSTACK(f)->stack_top - 1);
}

void 
op_dupx1(StackFrame *f)
{
  JavaStackItem any1, any2;

  JAVARLOG0(MYLOG, 1, "in op_dupx1()\n");

  op_stack_pop_any(ENV(f), OPSTACK(f), &any1);
  op_stack_pop_any(ENV(f), OPSTACK(f), &any2);

  op_stack_push_any(ENV(f), OPSTACK(f), &any1);
  op_stack_push_any(ENV(f), OPSTACK(f), &any2);
  op_stack_push_any(ENV(f), OPSTACK(f), &any1);
}

void 
op_dupx2(StackFrame *f)
{
  JavaStackItem any1, any2, any3;

  JAVARLOG0(MYLOG, 1, "in op_dupx2()\n");

  op_stack_pop_any(ENV(f), OPSTACK(f), &any1);
  op_stack_pop_any(ENV(f), OPSTACK(f), &any2);
  op_stack_pop_any(ENV(f), OPSTACK(f), &any3);

  op_stack_push_any(ENV(f), OPSTACK(f), &any1);
  op_stack_push_any(ENV(f), OPSTACK(f), &any3);
  op_stack_push_any(ENV(f), OPSTACK(f), &any2);
  op_stack_push_any(ENV(f), OPSTACK(f), &any1);
}

void 
op_dup2(StackFrame *f)
{
  JavaStackItem any1, any2;

  JAVARLOG0(MYLOG, 1, "in op_dup2()\n");
  
  op_stack_pop_any(ENV(f), OPSTACK(f), &any1);
  op_stack_pop_any(ENV(f), OPSTACK(f), &any2);

  op_stack_push_any(ENV(f), OPSTACK(f), &any2);
  op_stack_push_any(ENV(f), OPSTACK(f), &any1);
  op_stack_push_any(ENV(f), OPSTACK(f), &any2);
  op_stack_push_any(ENV(f), OPSTACK(f), &any1);
}

void 
op_dup2x1(StackFrame *f)
{
  JavaStackItem any1, any2, any3;

  JAVARLOG0(MYLOG, 1, "in op_dup2x1()\n");

  op_stack_pop_any(ENV(f), OPSTACK(f), &any1);
  op_stack_pop_any(ENV(f), OPSTACK(f), &any2);
  op_stack_pop_any(ENV(f), OPSTACK(f), &any3);

  op_stack_push_any(ENV(f), OPSTACK(f), &any2);
  op_stack_push_any(ENV(f), OPSTACK(f), &any1);
  op_stack_push_any(ENV(f), OPSTACK(f), &any3);
  op_stack_push_any(ENV(f), OPSTACK(f), &any2);
  op_stack_push_any(ENV(f), OPSTACK(f), &any1);
}

void 
op_dup2x2(StackFrame *f)
{
  JavaStackItem any1, any2, any3, any4;

  JAVARLOG0(MYLOG, 1, "in op_dup2x2()\n");

  op_stack_pop_any(ENV(f), OPSTACK(f), &any1);
  op_stack_pop_any(ENV(f), OPSTACK(f), &any2);
  op_stack_pop_any(ENV(f), OPSTACK(f), &any3);
  op_stack_pop_any(ENV(f), OPSTACK(f), &any4);

  op_stack_push_any(ENV(f), OPSTACK(f), &any2);
  op_stack_push_any(ENV(f), OPSTACK(f), &any1);
  op_stack_push_any(ENV(f), OPSTACK(f), &any4);
  op_stack_push_any(ENV(f), OPSTACK(f), &any3);
  op_stack_push_any(ENV(f), OPSTACK(f), &any2);
  op_stack_push_any(ENV(f), OPSTACK(f), &any1);
}

void 
op_swap(StackFrame *f)
{
  JavaStackItem any1, any2;

  JAVARLOG0(MYLOG, 1, "in op_swap()\n");

  op_stack_pop_any(ENV(f), OPSTACK(f), &any1);
  op_stack_pop_any(ENV(f), OPSTACK(f), &any2);

  op_stack_push_any(ENV(f), OPSTACK(f), &any1);
  op_stack_push_any(ENV(f), OPSTACK(f), &any2);
}

void
do_method_call (StackFrame *this_frame,
		MethodStruct *method)
{
  ClazzFile *cf = method->clazz;

  initialize_class(ENV(this_frame), cf);

  JAVARLOG2("DO_METHOD_CALL", 1, "    In do_method_call, class_name=%s,\n"
	    "     method_name=%s,\n",
	    getClassName(ENV(this_frame), cf), method->name);

  if (method->access_flags & ACC_NATIVE)
    {
      JAVARLOG0(MYLOG, 1, "   access is ACC_NATIVE\n");
      JAVARLOG2(MYLOG, 1, "\tNative method name is %s_%s\n",
		getClassName(ENV(this_frame), cf),
		method->name);

      do_native_method_call(ENV(this_frame), method);
    }
  else
    {
      StackFrame *f = create_frame_for_method (THREAD_INFO(this_frame), method);

      JAVARLOG2(MYLOG, 1, "    Virtual Method is %s.%s\n", 
		getClassName(ENV(this_frame), cf),
		method->name);

      fill_local_vars_from_stack (f, method->num_param_words,
				  method->access_flags & ACC_STATIC);

      OPSTACK_TOP(f) = OPSTACK(f)->stack_top;

      maybe_enter_monitor_for_method(ENV(f), method, THIS(f));
    }
}

void 
op_invokevirtual (StackFrame *f)
{
  u4 index;
  ClazzFile *cf;
  MethodStruct *method = NULL;

  char *class_name = NULL;
  char *method_name;
  obj_thunk* objectref;
  Signature *sig;
  int sig_size;
  int i;
  JNIEnv *env = ENV(f);

  JAVARLOG0(MYLOG, 1, "in op_invokevirtual()\n");

  index = (get_next_byte(f) << 8);
  index += get_next_byte(f);

  ResolveMethod(env, METHOD(f)->clazz, &METHOD(f)->clazz->constants[index],
		&method_name, &sig);

  sig_size = 0;
  for (i = 0; i < sig->method.num_params; i ++)
    sig_size += SIG_sizeInWords(env, sig->method.params[i]);

  op_stack_check_bottomspace(ENV(f), OPSTACK(f), sig_size + 1);

  objectref = (OPSTACK(f)->stack_top - sig_size - 1)->value.l;

  if (objectref == 0)
    {
      char msg[110];
#ifndef DEBUG
      snprintf(msg, sizeof(msg), "objectref == null in call to virtual method '%s'",
	       method_name);
#else
      /* CLASSNAME and METHODNAME is only defined when DEBUG is defined */
      snprintf(msg, sizeof(msg), "objectref == null"
	       " in call to virtual method '%s' from %s.%s()",
	       method_name,
	       CLASSNAME(f),
	       METHODNAME(f));
#endif
      throw_Exception(ENV(f), "java/lang/NullPointerException", msg);
      return;
    }

  while (!IS_POST(objectref))
    {
      objectref--;
    }
  if (!objectref) return;

  objectref++;

  while (!IS_POST(objectref) 
	 && !(method = GetMethodByNameAndSig(env,
					     STATIC_DATA(objectref),
					     method_name, sig)))
    {
      objectref ++;
    }

  if (IS_POST(objectref))
    {
      fprintf(stderr, "method=%s, ref=%p\n", method_name, objectref);
      assert(0 == "IS_POST(objectref)"); /* XXX Hack to get decent assert msg */
    }
  
  cf = (ClazzFile*)STATIC_DATA(objectref);
  class_name = getClassName(env, cf);
  
  JAVARLOG2(MYLOG, 1, "    Method is %s.%s\n", class_name, method_name);
  
  do_method_call (f, 
		  method);
}

void 
op_invokenonvirtual (StackFrame *f)
{
  u4 index;
  ClazzFile *cf;
  MethodStruct *method;

  JAVARLOG0(MYLOG, 1, "in op_invokenonvirtual()\n");

  index = (get_next_byte(f) << 8);
  index += get_next_byte(f);

  method = ResolveNonVirtualMethodRef(ENV(f), METHOD(f)->clazz, &METHOD(f)->clazz->constants[index]);
  cf = method->clazz;

  if (cf == NULL)
    {
      throw_Exception(ENV(f), "java/lang/RuntimeException", NULL);
      return;
    }

  if (method)
    {
      do_method_call (f, method);
    }
  else
    {
      /* should probably raise an exception */
    }
}


void 
op_invokestatic (StackFrame *f)
{
  u4 index;
  ClazzFile *cf;
  MethodStruct *method;

  JAVARLOG0(MYLOG, 1, "in op_invokestatic()\n");

  index = (get_next_byte(f) << 8);
  index += get_next_byte(f);

  method = ResolveStaticMethodRef(ENV(f), METHOD(f)->clazz, &METHOD(f)->clazz->constants[index]);
  cf = method->clazz;
  if (cf == NULL)
    {
      throw_Exception(ENV(f), "java/lang/RuntimeException", NULL);
      return;
    }

  do_method_call(f, method);
}


void 
op_invokeinterface (StackFrame *f)
{
  u4 index;
  u4 nargs;
  u4 reserved;
  jobject objectref;
  char *interface_methodname;
  Signature *interface_methodsig;
  MethodStruct *method;
  ClazzFile *clazz;

  JAVARLOG0(MYLOG, 1, "in op_invokeinterface()\n");

  index = (get_next_byte(f) << 8);
  index += get_next_byte(f);

  nargs = get_next_byte(f);
  reserved = get_next_byte(f);

  ResolveInterfaceMethod(ENV(f), METHOD(f)->clazz, &METHOD(f)->clazz->constants[index],
			 &interface_methodname,
			 &interface_methodsig);

  op_stack_check_bottomspace(ENV(f), OPSTACK(f), nargs);

  objectref = (OPSTACK(f)->stack_top - nargs)->value.l;

  method = get_interface_method_info(ENV(f),
				     (obj_thunk*)objectref,
				     &clazz,
				     interface_methodname,
				     interface_methodsig);

  method = GetMethodByNameAndSig(ENV(f), clazz, interface_methodname, interface_methodsig);

  do_method_call(f, method);
}

void
maybe_push_return_value(StackFrame *f)
{
  StackFrame *parent = get_frame_parent(f);

  if (parent < THREAD_INFO(f)->stack_highwater)
    {
      /* restore the old stack top -- this must use OPSTACK_TOP(f)!! */
      OPSTACK(f)->stack_top = OPSTACK_TOP(f);

      /* and, if the parent frame is java (and there's actually a
         return value), push the return value. */
      if (! (parent->flags & FRAME_NATIVE)
	  && THREAD_INFO(parent)->return_value.tag != SIG_JVOID )
	op_stack_push_any(ENV(parent),
			  OPSTACK(parent),
			  &THREAD_INFO(parent)->return_value);
    }
}

void 
op_ireturn (StackFrame *f)
{
  jint v;

  JAVARLOG0(MYLOG, 1, "in op_ireturn ()\n");

  op_stack_pop_int(ENV(f), OPSTACK(f), &v);
  
  THREAD_INFO(f)->return_value.tag = SIG_JINT;
  THREAD_INFO(f)->return_value.value.i = v;

  maybe_push_return_value(f);

  pop_frame(THREAD_INFO(f));
}


void 
op_lreturn (StackFrame *f)
{
  jlong ret_val;

  JAVARLOG0(MYLOG, 1, "in op_lreturn ()\n");

  op_stack_pop_long(ENV(f), OPSTACK(f), &ret_val);
  
  THREAD_INFO(f)->return_value.tag = SIG_JLONG;
  THREAD_INFO(f)->return_value.value.j = ret_val;

  maybe_push_return_value(f);

  pop_frame (THREAD_INFO(f));
}


void 
op_freturn (StackFrame *f)
{
  jfloat v;

  JAVARLOG0(MYLOG, 1, "in op_freturn ()\n");

  op_stack_pop_float(ENV(f), OPSTACK(f), &v);
  
  THREAD_INFO(f)->return_value.tag = SIG_JFLOAT;
  THREAD_INFO(f)->return_value.value.f = v;

  maybe_push_return_value(f);

  pop_frame (THREAD_INFO(f));
}

void 
op_dreturn (StackFrame *f)
{
  jdouble ret_val;

  JAVARLOG0(MYLOG, 1, "in op_dreturn ()\n");

  op_stack_pop_double(ENV(f), OPSTACK(f), &ret_val);
  
  THREAD_INFO(f)->return_value.tag = SIG_JDOUBLE;
  THREAD_INFO(f)->return_value.value.d = ret_val;

  maybe_push_return_value(f);

  pop_frame (THREAD_INFO(f));
}

void 
op_areturn (StackFrame *f)
{
  u4 v;

  JAVARLOG0(MYLOG, 1, "in op_areturn ()\n");

  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&v);
  
  THREAD_INFO(f)->return_value.tag = SIG_JOBJECT;
  THREAD_INFO(f)->return_value.value.i = v;

  maybe_push_return_value(f);

  pop_frame (THREAD_INFO(f));
}


void 
op_return (StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_return ()\n");

  THREAD_INFO(f)->return_value.tag = SIG_JVOID;

  maybe_push_return_value(f);

  pop_frame (THREAD_INFO(f));
}

void
op_ifeq(StackFrame *f)
{
  u2 a;
  short int branch_add;
  jint value;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_ifeq()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;
  op_stack_pop_int(ENV(f), OPSTACK(f), &value);

  if (value == 0)
    PC(f) = current_pc + branch_add;
}

void
op_ifnull(StackFrame *f)
{
  u2 a;
  short int branch_add;
  long value;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_ifnull()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;
  op_stack_pop_object(ENV(f), OPSTACK(f), (jobject*)&value);

  if (value == 0)
    PC(f) = current_pc + branch_add;
}

void
op_iflt(StackFrame *f)
{
  u2 a;
  short int branch_add;
  jint value;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_iflt()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;
  op_stack_pop_int(ENV(f), OPSTACK(f), &value);

  if (value < 0)
    PC(f) = current_pc + branch_add;
}

void
op_ifle(StackFrame *f)
{
  u2 a;
  short int branch_add;
  jint value;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_ifle()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;
  op_stack_pop_int(ENV(f), OPSTACK(f), &value);

  if (value <= 0)
    PC(f) = current_pc + branch_add;
}

void
op_ifne(StackFrame *f)
{
  u2 a;
  short int branch_add;
  jint value;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_ifne()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;
  op_stack_pop_int(ENV(f), OPSTACK(f), &value);

  if (value != 0)
    PC(f) = current_pc + branch_add;
}

void
op_ifnonnull(StackFrame *f)
{
  u2 a;
  short int branch_add;
  u4 value;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_ifnonnull()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;
  op_stack_pop_object(ENV(f), OPSTACK(f), (jobject*)&value);

  if (value != 0)
    PC(f) = current_pc + branch_add;
}

void
op_ifgt(StackFrame *f)
{
  u2 a;
  short int branch_add;
  jint value;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_ifgt()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;
  op_stack_pop_int(ENV(f), OPSTACK(f), &value);

  if (value > 0)
    PC(f) = current_pc + branch_add;
}

void
op_ifge(StackFrame *f)
{
  u2 a;
  short int branch_add;
  jint value;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_ifge()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;
  op_stack_pop_int(ENV(f), OPSTACK(f), &value);

  if (value >= 0)
    PC(f) = current_pc + branch_add;
}

void
op_ificmpeq(StackFrame *f)
{
  u2 a;
  short int branch_add;
  jint value1, value2;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_ificmpeq()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;

  op_stack_pop_int(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int(ENV(f), OPSTACK(f), &value1);

  if (value1 == value2)
    PC(f) = current_pc + branch_add;
}

void
op_ificmpne(StackFrame *f)
{
  u2 a;
  short int branch_add;
  jint value1, value2;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_ificmpne()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;
  op_stack_pop_int(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int(ENV(f), OPSTACK(f), &value1);

  if (value1 != value2)
    PC(f) = current_pc + branch_add;
}

void
op_ificmplt(StackFrame *f)
{
  u2 a;
  short int branch_add;
  jint value1, value2;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_ificmplt()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;
  op_stack_pop_int(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int(ENV(f), OPSTACK(f), &value1);

  if (value1 < value2)
    PC(f) = current_pc + branch_add;
}

void
op_ificmpgt(StackFrame *f)
{
  u2 a;
  short int branch_add;
  jint value1, value2;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_ificmpgt()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;
  op_stack_pop_int(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int(ENV(f), OPSTACK(f), &value1);

  if (value1 > value2)
    PC(f) = current_pc + branch_add;
}

void
op_ificmple(StackFrame *f)
{
  u2 a;
  short int branch_add;
  jint value1, value2;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_ificmple()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;
  op_stack_pop_int(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int(ENV(f), OPSTACK(f), &value1);

  if (value1 <= value2)
    PC(f) = current_pc + branch_add;
}

void
op_ificmpge(StackFrame *f)
{
  u2 a;
  short int branch_add;
  jint value1, value2;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_ificmpge()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;
  op_stack_pop_int(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int(ENV(f), OPSTACK(f), &value1);

  if (value1 >= value2)
    PC(f) = current_pc + branch_add;
}

void
op_lcmp(StackFrame *f)
{
  jlong value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_lcmp()\n");

  op_stack_pop_long(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_long(ENV(f), OPSTACK(f), &value1);

  if (value1 > value2)
    op_stack_push_int(ENV(f), OPSTACK(f), 1);
  else if (value1 == value2)
    op_stack_push_int(ENV(f), OPSTACK(f), 0);
  else
    op_stack_push_int(ENV(f), OPSTACK(f), -1);
}

void
op_fcmpl(StackFrame *f)
{
  jfloat value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_fcmpl()\n");  

  op_stack_pop_float(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_float(ENV(f), OPSTACK(f), &value1);

  if (value1 > value2)
    op_stack_push_int(ENV(f), OPSTACK(f), 1);
  else if (value1 == value2)
    op_stack_push_int(ENV(f), OPSTACK(f), 0);
  else if (value1 < value2)
    op_stack_push_int(ENV(f), OPSTACK(f), -1);
  else /* error condition ? */
    op_stack_push_int(ENV(f), OPSTACK(f), -1);
}

void
op_fcmpg(StackFrame *f)
{
  jfloat value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_fcmpg()\n");  

  op_stack_pop_float(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_float(ENV(f), OPSTACK(f), &value1);

  if (value1 > value2)
    op_stack_push_int(ENV(f), OPSTACK(f), 1);
  else if (value1 == value2)
    op_stack_push_int(ENV(f), OPSTACK(f), 0);
  else if (value1 < value2)
    op_stack_push_int(ENV(f), OPSTACK(f), -1);
  else /* error condition ? */
    op_stack_push_int(ENV(f), OPSTACK(f), 1);
}

void
op_dcmpl(StackFrame *f)
{
  jdouble value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_dcmpl()\n");

  op_stack_pop_double(ENV(f), OPSTACK(f), &value1);
  op_stack_pop_double(ENV(f), OPSTACK(f), &value2);

  if (value1 > value2)
    op_stack_push_int(ENV(f), OPSTACK(f), 1);
  else if (value1 == value2)
    op_stack_push_int(ENV(f), OPSTACK(f), 0);
  else if (value1 < value2)
    op_stack_push_int(ENV(f), OPSTACK(f), -1);
  else /* error condition ? */
    op_stack_push_int(ENV(f), OPSTACK(f), -1);
}

void
op_dcmpg(StackFrame *f)
{
  jdouble value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_dcmpg()\n");

  op_stack_pop_double(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_double(ENV(f), OPSTACK(f), &value1);

  if (value1 > value2)
    op_stack_push_int(ENV(f), OPSTACK(f), 1);
  else if (value1 == value2)
    op_stack_push_int(ENV(f), OPSTACK(f), 0);
  else if (value1 < value2)
    op_stack_push_int(ENV(f), OPSTACK(f), -1);
  else /* error condition ? */
    op_stack_push_int(ENV(f), OPSTACK(f), -1);
}

void
op_ifacmpeq(StackFrame *f)
{
  u2 a;
  short int branch_add;
  obj_thunk *value1, *value2;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_ifacmpeq()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;
  op_stack_pop_object(ENV(f), OPSTACK(f), (jobject*)&value2);
  op_stack_pop_object(ENV(f), OPSTACK(f), (jobject*)&value1);

  while (value1 && !IS_POST(value1)) value1--;
  while (value2 && !IS_POST(value2)) value2--;

  if (value1 == value2)
    PC(f) = current_pc + branch_add;
}

void
op_ifacmpne(StackFrame *f)
{
  u2 a;
  short int branch_add;
  obj_thunk *value1, *value2;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_ifacmpne()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;
  op_stack_pop_object(ENV(f), OPSTACK(f), (jobject*)&value2);
  op_stack_pop_object(ENV(f), OPSTACK(f), (jobject*)&value1);

  while (value1 && !IS_POST(value1)) value1--;
  while (value2 && !IS_POST(value2)) value2--;

  if (value1 != value2)
    PC(f) = current_pc + branch_add;
}

void
op_goto(StackFrame *f)
{
  u2 a;
  short int branch_add;
  int current_pc = PC(f) - 1;

  JAVARLOG0(MYLOG, 1, "in op_goto()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;

  PC(f) = current_pc + branch_add;
}

void
op_goto_w(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_goto_w()\n");  

  throw_Exception(ENV(f), "java/lang/RuntimeException", "unfinished opcode:  goto_w");
}

void
op_jsr(StackFrame *f)
{
  u2 a;
  short int branch_add;
  int current_pc = PC(f) - 1;
  jobject return_address;

  JAVARLOG0(MYLOG, 1, "in op_jsr()\n");  

  a = get_next_byte (f) << 8;
  a |= get_next_byte (f);

  branch_add = (short int) a;

  assert( sizeof(jobject) >= sizeof(PC(f)) );
  /* XXX Gives warning if sizeof(jobject) != sizeof(int) */
  return_address = (jobject)PC(f);

  /* return addresses are treated by the java compiler generated code
     as object references... what the fuck is wrong with sun?  this
     completely breaks the way the garbage collector should work. */
  op_stack_push_object(ENV(f), OPSTACK(f), return_address);

  PC(f) = current_pc + branch_add;
}

void
op_jsr_w(StackFrame *f)
{
  JAVARLOG0(MYLOG, 1, "in op_jsr_w()\n");  

  throw_Exception(ENV(f), "java/lang/RuntimeException", "unfinished opcode:  jsr_w");
}

void
op_ret(StackFrame *f)
{
  u2 vindex;

  JAVARLOG1(MYLOG, 1, "in op_ret() wide=%d\n", wide_vindex_set);  

  vindex = get_next_byte(f);
  if (wide_vindex_set)
    {
      u1 vindex2 = get_next_byte (f);
      vindex = (vindex << 8) | vindex2;
      wide_vindex_set = JNI_FALSE;
    }      

  PC(f) = VARS(f)[vindex].value.i;
}

void
op_iadd(StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_iadd()\n");

  op_stack_pop_int(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int(ENV(f), OPSTACK(f), &value1);

  op_stack_push_int(ENV(f), OPSTACK(f), value1 + value2);
}

void
op_ladd(StackFrame *f)
{
  jlong value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_ladd()\n");

  op_stack_pop_long(ENV(f), OPSTACK(f),  &value1);

  op_stack_pop_long(ENV(f), OPSTACK(f), &value2);

  op_stack_push_long(ENV(f), OPSTACK(f), value1 + value2);
}

void
op_fadd(StackFrame *f)
{
  jfloat value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_fadd()\n");

  op_stack_pop_float(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_float(ENV(f), OPSTACK(f), &value1);

  op_stack_push_float(ENV(f), OPSTACK(f), value1 + value2);
}

void
op_dadd(StackFrame *f)
{
  jdouble value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_dadd()\n");

  op_stack_pop_double(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_double(ENV(f), OPSTACK(f), &value1);

  op_stack_push_double(ENV(f), OPSTACK(f), value1 + value2);
}

void
op_isub(StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_isub()\n");

  op_stack_pop_int(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int(ENV(f), OPSTACK(f), &value1);

  op_stack_push_int(ENV(f), OPSTACK(f), value1 - value2);
}

void
op_lsub(StackFrame *f)
{
  jlong value1, value2;
  
  JAVARLOG0(MYLOG, 1, "in op_lsub()\n");

  op_stack_pop_long(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_long(ENV(f), OPSTACK(f), &value1);

  op_stack_push_long(ENV(f), OPSTACK(f), value1 - value2);
}

void
op_fsub(StackFrame *f)
{
  jfloat value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_fsub()\n");

  op_stack_pop_float(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_float(ENV(f), OPSTACK(f), &value1);

  op_stack_push_float(ENV(f), OPSTACK(f), value1 - value2);
}

void
op_dsub(StackFrame *f)
{
  jdouble value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_dsub()\n");

  op_stack_pop_double(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_double(ENV(f), OPSTACK(f), &value1);

  op_stack_push_double(ENV(f), OPSTACK(f), value1 - value2);
}

void
op_imul(StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_imul()\n");

  op_stack_pop_int(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int(ENV(f), OPSTACK(f), &value1);
 
  op_stack_push_int(ENV(f), OPSTACK(f), value1 * value2);
}

void
op_lmul(StackFrame *f)
{
  jlong value1, value2;
  
  JAVARLOG0(MYLOG, 1, "in op_lmul()\n");

  op_stack_pop_long(ENV(f), OPSTACK(f),  &value2);
  op_stack_pop_long(ENV(f), OPSTACK(f), &value1);

  op_stack_push_long(ENV(f), OPSTACK(f), value1 * value2);
}

void
op_fmul(StackFrame *f)
{
  jfloat value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_fmul()\n");

  op_stack_pop_float(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_float(ENV(f), OPSTACK(f), &value1);

  op_stack_push_float(ENV(f), OPSTACK(f), value1 * value2);
}

void
op_dmul(StackFrame *f)
{ 
  jdouble value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_dmul()\n");

  op_stack_pop_double(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_double(ENV(f), OPSTACK(f), &value1);

  op_stack_push_double(ENV(f), OPSTACK(f), value1 * value2);
}

void
op_idiv(StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_idiv()\n");

  op_stack_pop_int(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int(ENV(f), OPSTACK(f), &value1);

  if (value2 == 0)
    {
      throw_Exception(ENV(f), "java/lang/ArithmeticException", "divisor is 0 in integer division.");
      return;
    }
  else
    op_stack_push_int(ENV(f), OPSTACK(f), value1 / value2);
}

void
op_ldiv(StackFrame *f)
{
  jlong value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_ldiv()\n");

  op_stack_pop_long(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_long(ENV(f), OPSTACK(f), &value1);

  if (value2 == 0)
    {
      throw_Exception(ENV(f), "java/lang/ArithmeticException", "divisor is 0 in long division.");
      return;
    }
  else
    op_stack_push_long(ENV(f), OPSTACK(f), value1 / value2);
}

void
op_fdiv(StackFrame *f)
{
  jfloat value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_fdiv()\n");

  op_stack_pop_float(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_float(ENV(f), OPSTACK(f), &value1);

  op_stack_push_float(ENV(f), OPSTACK(f), value1 / value2);
}

void
op_ddiv(StackFrame *f)
{
  jdouble value1, value2;
  
  JAVARLOG0(MYLOG, 1, "in op_ddiv()\n");

  op_stack_pop_double(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_double(ENV(f), OPSTACK(f), &value1);

  op_stack_push_double(ENV(f), OPSTACK(f), value1 / value2);
}

void
op_irem(StackFrame *f)
{
  jint value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_irem()\n");

  op_stack_pop_int(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_int(ENV(f), OPSTACK(f), &value1);

  if (value2 == 0)
    {
      throw_Exception(ENV(f), "java/lang/ArithmeticException", "divisor is 0 in integer remainder.");
      return;
    }
  else
    op_stack_push_int(ENV(f), OPSTACK(f), value1 % value2);
}

void
op_lrem(StackFrame *f)
{
  jlong value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_lrem()\n");

  op_stack_pop_long(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_long(ENV(f), OPSTACK(f), &value1);

  if (value2 == 0)
    {
      throw_Exception(ENV(f), "java/lang/ArithmeticException", "divisor is 0 in long remainder.");
      return;
    }
  else
    op_stack_push_long(ENV(f), OPSTACK(f), value1 % value2);
}

void
op_frem(StackFrame *f)
{
  jfloat value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_frem()\n");

  op_stack_pop_float(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_float(ENV(f), OPSTACK(f), &value1);

  op_stack_push_float(ENV(f), OPSTACK(f), fmodf(value1,value2));
}

void
op_drem(StackFrame *f)
{
  jdouble value1, value2;

  JAVARLOG0(MYLOG, 1, "in op_drem()\n");

  op_stack_pop_double(ENV(f), OPSTACK(f), &value2);
  op_stack_pop_double(ENV(f), OPSTACK(f), &value1);

  op_stack_push_double(ENV(f), OPSTACK(f), fmod(value1,value2));
}

void
op_ineg(StackFrame *f)
{
  jint value;

  JAVARLOG0(MYLOG, 1, "in op_ineg()\n");

  op_stack_pop_int(ENV(f), OPSTACK(f), &value);

  op_stack_push_int(ENV(f), OPSTACK(f), -1 * value);
}

void
op_lneg(StackFrame *f)
{
  jlong value;

  JAVARLOG0(MYLOG, 1, "in op_lneg()\n");

  op_stack_pop_long(ENV(f), OPSTACK(f), &value);

  op_stack_push_long(ENV(f), OPSTACK(f), -1 * value);
}

void
op_fneg(StackFrame *f)
{
  jfloat value;

  JAVARLOG0(MYLOG, 1, "in op_fneg()\n");

  op_stack_pop_float(ENV(f), OPSTACK(f), &value);
  op_stack_push_float(ENV(f), OPSTACK(f), -1.0 * value);
}

void
op_dneg(StackFrame *f)
{
  jdouble value;

  JAVARLOG0(MYLOG, 1, "in op_dneg()\n");

  op_stack_pop_double(ENV(f), OPSTACK(f), &value);
  op_stack_push_double(ENV(f), OPSTACK(f), -1.0 * value);
}

void
op_iinc (StackFrame *f)
{
  u2 vindex;

  /* should typecheck the variable... */

  JAVARLOG1(MYLOG, 1, "in op_iinc() wide=%d\n", wide_vindex_set);

  vindex = get_next_byte(f);
  if (wide_vindex_set)
    {
      u1 vindex2 = get_next_byte (f);
      vindex = (vindex << 8) | vindex2;
    }

  /* These are handlet separatly to get then signed constant correct */
  if (wide_vindex_set)
    {
      jshort c = get_next_byte(f);
      u1 c2 = get_next_byte (f);
      vindex = (c << 8) | c2;
      wide_vindex_set = JNI_FALSE;
      VARS(f)[vindex].value.i += c;
    }
  else
    {
      jbyte c = get_next_byte(f);
      VARS(f)[vindex].value.i += c;
    }
}

void op_i2l(StackFrame *f)
{
  jint ivalue;

  JAVARLOG0(MYLOG, 1, "in op_i2l()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &ivalue);
  op_stack_push_long (ENV(f), OPSTACK(f), (jlong)ivalue); 
}


void op_i2f(StackFrame *f)
{
  jint ivalue;

  JAVARLOG0(MYLOG, 1, "in op_i2f()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &ivalue);
  op_stack_push_float (ENV(f), OPSTACK(f), (jfloat)ivalue);
}


void op_i2d(StackFrame *f)
{
  jint ivalue;

  JAVARLOG0(MYLOG, 1, "in op_i2d()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &ivalue);
  op_stack_push_double (ENV(f), OPSTACK(f), (jdouble)ivalue);
}

void op_f2i(StackFrame *f)
{
  jfloat fvalue;

  JAVARLOG0(MYLOG, 1, "in op_f2i()\n");

  op_stack_pop_float(ENV(f), OPSTACK(f), &fvalue);
  op_stack_push_int(ENV(f), OPSTACK(f), (jint)fvalue);
}

void 
op_f2d(StackFrame *f)
{
  jfloat fvalue;

  JAVARLOG0(MYLOG, 1, "in op_f2d()\n");

  op_stack_pop_float(ENV(f), OPSTACK(f), &fvalue);
  op_stack_push_double(ENV(f), OPSTACK(f), (jdouble)fvalue);
}

void op_int2byte(StackFrame *f)
{
  jint ivalue;

  JAVARLOG0(MYLOG, 1, "in op_int2byte()\n");

  op_stack_pop_int(ENV(f), OPSTACK(f), &ivalue);
  op_stack_push_byte(ENV(f), OPSTACK(f), (jbyte)ivalue);
}


void op_int2char (StackFrame *f)
{
  jint ivalue;

  JAVARLOG0(MYLOG, 1, "in op_int2char()\n");

  op_stack_pop_int(ENV(f), OPSTACK(f), &ivalue);
  op_stack_push_char(ENV(f), OPSTACK(f), (jchar)(ivalue & 0xFFFF));
}

void op_l2i (StackFrame *f)
{
  jlong lvalue;

  JAVARLOG0(MYLOG, 1, "in op_l2i()\n");

  op_stack_pop_long(ENV(f), OPSTACK(f), &lvalue);
  op_stack_push_int(ENV(f), OPSTACK(f), (jint)(lvalue & 0xFFFFFFFF));
}

void op_l2f (StackFrame *f)
{
  jlong lvalue;

  JAVARLOG0(MYLOG, 1, "in op_l2f()\n");

  op_stack_pop_long(ENV(f), OPSTACK(f), &lvalue);
  op_stack_push_float(ENV(f), OPSTACK(f), (jfloat)lvalue);
}

void op_l2d (StackFrame *f)
{
  jlong lvalue;

  JAVARLOG0(MYLOG, 1, "in op_l2d()\n");

  op_stack_pop_long(ENV(f), OPSTACK(f), &lvalue);
  op_stack_push_double(ENV(f), OPSTACK(f), (jdouble)lvalue);
}

void op_f2l (StackFrame *f)
{
  jfloat fvalue;

  JAVARLOG0(MYLOG, 1, "in op_f2l()\n");

  op_stack_pop_float(ENV(f), OPSTACK(f), &fvalue);
  op_stack_push_long(ENV(f), OPSTACK(f), (jlong)fvalue);
}

void op_d2i (StackFrame *f)
{
  jdouble dvalue;

  JAVARLOG0(MYLOG, 1, "in op_d2i()\n");

  op_stack_pop_double(ENV(f), OPSTACK(f), &dvalue);
  op_stack_push_int(ENV(f), OPSTACK(f), (jint)dvalue);
}

void op_d2l (StackFrame *f)
{
  jdouble dvalue;

  JAVARLOG0(MYLOG, 1, "in op_d2l()\n");

  op_stack_pop_double(ENV(f), OPSTACK(f), &dvalue);
  op_stack_push_long(ENV(f), OPSTACK(f), (jlong)dvalue);
}

void op_d2f (StackFrame *f)
{
  jdouble dvalue;

  JAVARLOG0(MYLOG, 1, "in op_d2f()\n");

  op_stack_pop_double(ENV(f), OPSTACK(f), &dvalue);
  op_stack_push_float(ENV(f), OPSTACK(f), (jfloat)dvalue);
}

void op_int2short (StackFrame *f)
{
  jint ivalue;

  JAVARLOG0(MYLOG, 1, "in op_i2l()\n");

  op_stack_pop_int(ENV(f), OPSTACK(f), &ivalue);
  op_stack_push_short(ENV(f), OPSTACK(f), (jshort)ivalue);
}

void 
op_newarray (StackFrame *f)
{
  u1 atype;
  jint size;
  obj_thunk *array_reference;
  char *array_classname = NULL;

  JAVARLOG0(MYLOG, 1, "in op_newarray()\n");

  atype = get_next_byte (f);
  op_stack_pop_int (ENV(f), OPSTACK(f), &size);

  switch (atype)
    {
    case 4: /*T_BOOLEAN*/
      array_classname = "[Z";
      break;
    case 5: /*T_CHAR*/
      array_classname = "[C";
      break;
    case 6: /*T_FLOAT*/
      array_classname = "[F";
      break;
    case 7: /*T_DOUBLE*/
      array_classname = "[D";
      break;
    case 8: /*T_BYTE*/
      array_classname = "[B";
      break;
    case 9: /*T_SHORT*/
      array_classname = "[S";
      break;
    case 10: /*T_INT*/
      array_classname = "[I";
      break;
    case 11: /*T_LONG*/
      array_classname = "[J";
      break;
    default:
      assert(0);
    }
  
  array_reference = new_array (ENV(f), size, createFakeArrayClass(ENV(f), array_classname));
  GC_addReference((JavaVM*)((HungryJNIEnv*)ENV(f))->_vm, (GC_obj*)array_reference);

  op_stack_push_object (ENV(f), OPSTACK(f), array_reference);
}


void 
op_anewarray (StackFrame *f)
{
  u2 ob_type;
  jint size;
  obj_thunk *array_reference;
  char *element_classname;
  char *array_classname;
  
  JAVARLOG0(MYLOG, 1, "op_anewarray... \n");

  ob_type = get_next_byte (f) << 8;
  ob_type |= get_next_byte (f);

  element_classname = ResolveClassName(ENV(f), METHOD(f)->clazz, &METHOD(f)->clazz->constants[ob_type]);

  array_classname = (char*)malloc(strlen(element_classname) + 4);
  sprintf (array_classname, "[L%s;", element_classname);

  op_stack_pop_int (ENV(f), OPSTACK(f), &size);

  array_reference = new_array (ENV(f), size, createFakeArrayClass(ENV(f), array_classname));

  free(array_classname);

  GC_addReference((JavaVM*)((HungryJNIEnv*)ENV(f))->_vm, (GC_obj*)array_reference);
  op_stack_push_object (ENV(f), OPSTACK(f), array_reference);
}

void 
op_multianewarray (StackFrame *f)
{
  static jint *sizes = NULL;
  static int sizes_length;
  int i;
  u2 index;
  u1 dimensions;
  obj_thunk *array_reference;
  char *array_classname;

  JAVARLOG0(MYLOG, 1, "in op_multianewarray()\n");

  index = get_next_byte (f) << 8;
  index |= get_next_byte (f);

  dimensions = get_next_byte (f);

  /* get space to store the list of array dimensions */

  if (sizes == NULL)
    {
      sizes = (jint *) malloc (sizeof (u4) * (dimensions + 5));
      sizes_length = dimensions + 5;
    }

  if (dimensions >= sizes_length)
    {
      sizes = (jint *) realloc (sizes, sizeof (u4) * (dimensions + 5));
      sizes_length = dimensions + 5;
    }

  /* pop the array sizes from the operand stack */

  for (i=0; i<dimensions; i++)
    op_stack_pop_int (ENV(f), OPSTACK(f), sizes + dimensions - i - 1);

  array_classname = ResolveClassName(ENV(f), METHOD(f)->clazz, &METHOD(f)->clazz->constants[index]);

  array_reference = multi_new_array (ENV(f),
				     sizes,
				     dimensions,
				     createFakeArrayClass(ENV(f), array_classname));

  GC_addReference((JavaVM*)((HungryJNIEnv*)ENV(f))->_vm, (GC_obj*)array_reference);
  op_stack_push_object (ENV(f), OPSTACK(f), array_reference);
}


void 
op_arraylength (StackFrame *f)
{
  jobject arrayref;
  JavaStackItem item;

  JAVARLOG0(MYLOG, 1, "in op_arraylength ()\n");

  op_stack_pop_object(ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &item);

  assert(item.tag == SIG_JINT);

  JAVARLOG1(MYLOG, 2, "   length=%d\n", item.value.i);

  op_stack_push_any (ENV(f), OPSTACK(f), &item);
}


void 
op_iaload (StackFrame *f)
{
  jobject arrayref;
  jint index;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_iaload ()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", "op_iaload");
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException",
		      "op_iaload");
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  op_stack_push_int(ENV(f), OPSTACK(f),
		    ((jint*)body_item.value.l)[ index ]);
}

void 
op_laload (StackFrame *f)
{
  jobject arrayref;
  jint index;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_laload ()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  op_stack_push_long(ENV(f), OPSTACK(f),
		     ((jlong*)body_item.value.l)[ index ]);
}


void 
op_faload (StackFrame *f)
{
  jobject arrayref;
  jint index;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_faload ()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  op_stack_push_float(ENV(f), OPSTACK(f),
		      ((jfloat*)body_item.value.l)[ index ]);
}


void 
op_daload (StackFrame *f)
{
  jobject arrayref;
  jint index;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_daload ()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  op_stack_push_double(ENV(f), OPSTACK(f),
		       ((jdouble*)body_item.value.l)[ index ]);
}


void 
op_aaload (StackFrame *f)
{
  jobject arrayref;
  jint index;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_aaload ()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  op_stack_push_object(ENV(f), OPSTACK(f),
		       ((jobject*)body_item.value.l)[ index ]);
}


void 
op_baload (StackFrame *f)
{
  jobject arrayref;
  jint index;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_baload ()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  op_stack_push_byte(ENV(f), OPSTACK(f),
		     ((jbyte*)body_item.value.l)[ index ]);
}


void 
op_caload (StackFrame *f)
{
  jobject arrayref;
  jint index;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_caload ()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  op_stack_push_char(ENV(f), OPSTACK(f),
		     ((jchar*)body_item.value.l)[ index ]);
}


void 
op_saload (StackFrame *f)
{
  jobject arrayref;
  jint index;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_saload ()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  op_stack_push_short(ENV(f), OPSTACK(f),
		      ((jshort*)body_item.value.l)[ index ]);
}


void 
op_iastore (StackFrame *f)
{
  jobject arrayref;
  jint index;
  jint value;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_iastore ()\n");

  op_stack_pop_int (ENV(f), OPSTACK(f), &value);
  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  ((jint*)body_item.value.l)[ index ] = value;
}

void 
op_lastore (StackFrame *f)
{
  jobject arrayref;
  jint index;
  jlong value;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_lastore ()\n");

  op_stack_pop_long (ENV(f), OPSTACK(f), &value);
  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  ((jlong*)body_item.value.l)[ index ] = value;
}

void 
op_fastore (StackFrame *f)
{
  jobject arrayref;
  jint index;
  jfloat value;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_fastore ()\n");

  op_stack_pop_float (ENV(f), OPSTACK(f), &value);
  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  ((jfloat*)body_item.value.l)[ index ] = value;
}


void 
op_dastore (StackFrame *f)
{
  jobject arrayref;
  jint index;
  jdouble value;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_dastore ()\n");

  op_stack_pop_double (ENV(f), OPSTACK(f), &value);
  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  ((jdouble*)body_item.value.l)[ index ] = value;
}

void 
op_aastore (StackFrame *f)
{
  jobject arrayref;
  jint index;
  jobject value;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_aastore ()\n");

  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&value);
  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  ((jobject*)body_item.value.l)[ index ] = value;
}


void 
op_bastore (StackFrame *f)
{
  jobject arrayref;
  jint index;
  jbyte value;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_bastore ()\n");

  op_stack_pop_byte (ENV(f), OPSTACK(f), &value);
  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  ((jbyte*)body_item.value.l)[ index ] = value;
}


void 
op_castore (StackFrame *f)
{
  jobject arrayref;
  jint index;
  jchar value;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_castore ()\n");

  op_stack_pop_char (ENV(f), OPSTACK(f), &value);
  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  ((jchar*)body_item.value.l)[ index ] = value;
}


void 
op_sastore (StackFrame *f)
{
  jobject arrayref;
  jint index;
  jshort value;
  JavaStackItem length_item;
  JavaStackItem body_item;

  JAVARLOG0(MYLOG, 1, "in op_sastore ()\n");

  op_stack_pop_short (ENV(f), OPSTACK(f), &value);
  op_stack_pop_int (ENV(f), OPSTACK(f), &index);
  op_stack_pop_object (ENV(f), OPSTACK(f), (jobject*)&arrayref);

  if (arrayref == 0)
    {
      throw_Exception(ENV(f), "java/lang/NullPointerException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_LENGTH_INDEX],
		     &length_item);

  if (index >= length_item.value.i)
    {
      throw_Exception(ENV(f), "java/lang/ArrayIndexOutOfBoundsException", NULL);
      return;
    }

  get_instance_field(arrayref,
		     STATIC_DATA(arrayref)->fields[ARRAY_BODY_INDEX],
		     &body_item);

  ((jshort*)body_item.value.l)[ index ] = value;
}

void 
op_monitorenter (StackFrame *f)
{
  jobject objectref;

  op_stack_pop_object(ENV(f), OPSTACK(f), (void **)&objectref);

  MONITOR_enter(MONITOR_getForObject(objectref));
}

void 
op_monitorexit (StackFrame *f)
{
  jobject objectref;

  op_stack_pop_object(ENV(f), OPSTACK(f), (jobject*)&objectref);

  MONITOR_exit(MONITOR_getForObject(objectref));
}

long
get_4byte(StackFrame *f)
{
  long ret_val;

  ret_val = (u1)get_next_byte(f); 
  ret_val <<= 8; 
  ret_val |= (u1)get_next_byte(f); 
  ret_val <<= 8; 
  ret_val |= (u1)get_next_byte(f); 
  ret_val <<= 8; 
  ret_val |= (u1)get_next_byte(f); 

  return ret_val;
}
void 
op_lookupswitch(StackFrame *f)
{
  jint address_of_this_instruction = PC(f) - 1 ;
  jint numpairs;
  int padding;
  jint the_offset;
  jboolean found;
  jint i;
  jint key;

  JAVARLOG0(MYLOG, 1, "in op_lookupswitch()\n");

  /* eat up the padding -- the interesting stuff is aligned
     to a 4 byte boundary. */
  padding = 3 - (address_of_this_instruction % 4);

  JAVARLOG1(MYLOG, 2, "  lookupswitch:  data padded with %d 0 bytes\n", padding);
  
  for (i = 0; i < padding; i ++)
    {
      (void)get_next_byte(f);
    }

  op_stack_pop_int(ENV(f), OPSTACK(f), &key);

  /* get the default offset. */
  the_offset = get_4byte(f);

  JAVARLOG1(MYLOG, 2, "  lookupswitch:  default offset is %ld\n", the_offset);

  /* get the number of pairs. */
  numpairs = get_4byte(f);

  JAVARLOG1(MYLOG, 2, "  lookupswitch:  there are %ld pairs\n", numpairs);

  found = JNI_FALSE;
  for (i = 0; i < numpairs; i ++)
    {
      jint match = get_4byte(f);
      jint offset = get_4byte(f);

      if (!found && key == match)
	{
	  the_offset = offset;
	  found = JNI_TRUE;
	}
    }

  PC(f) = address_of_this_instruction + the_offset;
}

void 
op_tableswitch(StackFrame *f)
{
  jint index;
  jint high, low, default_off;
  jint offset_to_use, i;
  jint address_of_this_instruction = PC(f) - 1 ;
  jint padding;

  JAVARLOG0(MYLOG, 1, "in op_tableswitch()\n");

  /* eat up the padding -- the interesting stuff is aligned
     to a 4 byte boundary. */
  padding = 3 - (address_of_this_instruction % 4);

  JAVARLOG1(MYLOG, 1, "  tableswitch:  data padded with %d 0 bytes\n", padding);

  for (i = 0; i < padding; i ++)
    {
      (void)get_next_byte(f);
    }

  op_stack_pop_int(ENV(f), OPSTACK(f), &index);

  /* get the default offset. */
  default_off = get_4byte(f);

  JAVARLOG1(MYLOG, 2, "  tableswitch:  default offset is %ld\n", default_off);

  /* get the low index. */
  low = get_4byte(f);

  JAVARLOG1(MYLOG, 2, "  tableswitch:  low index is %ld\n", low);

  /* get the high index. */
  high = get_4byte(f);

  JAVARLOG1(MYLOG, 2, "  tableswitch:  high index is %ld\n", high);

  offset_to_use = default_off;

  for (i = 0 ; i < high - low + 1; i ++)
    {
      jint jump_offset = get_4byte(f);

      /* just consume all the jump_offsets */
      if (index > high || index < low) continue;
      
      if (index - low == i)
	offset_to_use = jump_offset;
    }

  JAVARLOG2(MYLOG, 2, 
	    "  tableswitch:  index = %ld, offset = %ld\n", index, offset_to_use);

  PC(f) = address_of_this_instruction + offset_to_use;
}

void op_athrow(StackFrame *f)
{
  jthrowable throwable_ref;
  JNIEnv *env = ENV(f);

  JAVARLOG0(MYLOG, 1, "in interp_athrow()\n");  

    
  /* get the object reference */
  op_stack_pop_object(env, OPSTACK(f), (jobject*)&throwable_ref);

  JAVARLOG1("Exceptions", 1, "throwing exception '%s' from java code\n",
	    getClassName(env, (ClazzFile*)STATIC_DATA(throwable_ref)));

  (*env)->Throw(env, throwable_ref);
}
