/* -*- Mode: C; c-file-style: "gnu" -*-
   interploop.c -- the main interpreter loop.
   Created: Chris Toshok <toshok@hungry.com>, 13-Aug-1997
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "config.h"
#include "jniint.h"
#include "objects.h"
#include "log.h"
#include "interp.h"
#include "resolve.h"
#include "exceptions.h"
#include "native-threads.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>

static void op_notimplemented();

#define MYLOG "InterpLoop"

/*
 * All the opcode methods are decleared in interp.h
 * (r) behind the opcode number means the opcode is reserved
 */
opcode opcode_table[] = {
  { "nop",		0,	0,	0,	op_nop  }/* 0 */,
  { "aconst_null",	0,	0,	1,	op_aconst_null }/* 1 */,
  { "iconst_m1",	0,	0,	1,	op_iconst_m1 }/* 2 */,
  { "iconst_0",		0,	0,	1,	op_iconst_0 }/* 3 */,
  { "iconst_1",		0,	0,	1,	op_iconst_1 }/* 4 */,
  { "iconst_2",		0,	0,	1,	op_iconst_2 }/* 5 */,
  { "iconst_3",		0,	0,	1,	op_iconst_3 }/* 6 */,
  { "iconst_4",		0,	0,	1,	op_iconst_4 }/* 7 */,
  { "iconst_5",		0,	0,	1,	op_iconst_5 }/* 8 */,
  { "lconst_0",		0,	0,	1,	op_lconst_0 }/* 9 */,
  { "lconst_1",		0,	0,	1,	op_lconst_1 }/* 10 */,
  { "fconst_0",		0,	0,	1,	op_fconst_0 }/* 11 */,
  { "fconst_1",		0,	0,	1,	op_fconst_1 }/* 12 */,
  { "fconst_2",		0,	0,	1,	op_fconst_2 }/* 13 */,
  { "dconst_0",		0,	0,	2,	op_dconst_0 }/* 14 */,
  { "dconst_1",		0,	0,	2,	op_dconst_1 }/* 15 */,
  { "bipush",		1,	0,	1,	op_bipush }/* 16 */,
  { "sipush",		2,	0,	1,	op_sipush }/* 17 */,
  { "ldc",		1,	0,	1,	op_ldc }/* 18 */,
  { "ldc_w",		2,	0,	1,	op_ldc_w }/* 19 */,
  { "ldc2_w",		2,	0,	2,	op_ldc2w }/* 20 */,
  { "iload",		1,	0,	1,	op_iload }/* 21 */,
  { "lload",		1,	0,	2,	op_lload }/* 22 */,
  { "fload",		1,	0,	1,	op_fload }/* 23 */,
  { "dload",		1,	0,	2,	op_dload }/* 24 */,
  { "aload",		1,	0,	1,	op_aload }/* 25 */,
  { "iload_0",		0,	0,	1,	op_iload_0 }/* 26 */,
  { "iload_1",		0,	0,	1,	op_iload_1 }/* 27 */,
  { "iload_2",		0,	0,	1,	op_iload_2 }/* 28 */,
  { "iload_3",		0,	0,	1,	op_iload_3 } /* 29 */,
  { "lload_0",		0,	0,	2,	op_lload_0 }/* 30 */,
  { "lload_1",		0,	0,	2,	op_lload_1 }/* 31 */,
  { "lload_2",		0,	0,	2,	op_lload_2 }/* 32 */,
  { "lload_3",		0,	0,	2,	op_lload_3 }/* 33 */,
  { "fload_0",		0,	0,	1,	op_fload_0 }/* 34 */,
  { "fload_1",		0,	0,	1,	op_fload_1 }/* 35 */,
  { "fload_2",		0,	0,	1,	op_fload_2 }/* 36 */,
  { "float_3",		0,	0,	1,	op_fload_3 }/* 37 */,
  { "dload_0",		0,	0,	2,	op_dload_0 }/* 38 */,
  { "dload_1",		0,	0,	2,	op_dload_1 }/* 39 */,
  { "dload_2",		0,	0,	2,	op_dload_2 }/* 40 */,
  { "dload_3",		0,	0,	2,	op_dload_3 }/* 41 */,
  { "aload_0",		0,	0,	1,	op_aload_0 }/* 42 */,
  { "aload_1",		0,	0,	1,	op_aload_1 }/* 43 */,
  { "aload_2",		0,	0,	1,	op_aload_2 }/* 44 */,
  { "aload_3",		0,	0,	1,	op_aload_3 }/* 45 */,
  { "iaload",		0,	2,	1,	op_iaload } /* 46 */,
  { "laload",		0,	2,	2,	op_laload } /* 47 */,
  { "faload",		0,	2,	1,	op_faload } /* 48 */,
  { "daload",		0,	2,	2,	op_daload } /* 49 */,
  { "aaload",		0,	2,	1,	op_aaload } /* 50 */,
  { "baload",		0,	2,	1,	op_baload } /* 51 */,
  { "caload",		0,	2,	1,	op_caload } /* 52 */,
  { "saload",		0,	2,	1,	op_saload } /* 53 */,
  { "istore",		1,	0,	3,	op_istore } /* 54 */,
  { "lstore",		1,	0,	4,	op_lstore } /* 55 */,
  { "fstore",		1,	0,	3,	op_fstore } /* 56 */,
  { "dstore",		1,	0,	4,	op_dstore } /* 57 */,
  { "astore",		1,	0,	3,	op_astore } /* 58 */,
  { "istore_0",		0,	1,	0,	op_istore_0 } /* 59 */,
  { "istore_1",		0,	1,	0,	op_istore_1 } /* 60 */,
  { "istore_2",		0,	1,	0,	op_istore_2 } /* 61 */,
  { "istore_3",		0,	1,	0,	op_istore_3 } /* 62 */,
  { "lstore_0",		0,	2,	0,	op_lstore_0 } /* 63 */,
  { "lstore_1",		0,	2,	0,	op_lstore_1 } /* 64 */,
  { "lstore_2",		0,	2,	0,	op_lstore_2 } /* 65 */,
  { "lstore_3",		0,	2,	0,	op_lstore_3 } /* 66 */,
  { "fstore_0",		0,	1,	0,	op_fstore_0 } /* 67 */,
  { "fstore_1",		0,	1,	0,	op_fstore_1 } /* 68 */,
  { "fstore_2",		0,	1,	0,	op_fstore_2 } /* 69 */,
  { "fstore_3",		0,	1,	0,	op_fstore_3 } /* 70 */,
  { "dstore_0",		0,	2,	0,	op_dstore_0 } /* 71 */,
  { "dstore_1",		0,	2,	0,	op_dstore_1 } /* 72 */,
  { "dstore_2",		0,	2,	0,	op_dstore_2 } /* 73 */,
  { "dstore_3",		0,	2,	0,	op_dstore_3 } /* 74 */,
  { "astore_0",		0,	1,	0,	op_astore_0 } /* 75 */,
  { "astore_1",		0,	1,	0,	op_astore_1 } /* 76 */,
  { "astore_2",		0,	1,	0,	op_astore_2 } /* 77 */,
  { "astore_3",		0,	1,	0,	op_astore_3 } /* 78 */,
  { "iastore",		0,	3,	0,	op_iastore } /* 79 */,
  { "lastore",		0,	4,	0,	op_lastore } /* 80 */,
  { "fastore",		0,	3,	0,	op_fastore } /* 81 */,
  { "dastore",		0,	4,	0,	op_dastore } /* 82 */,
  { "aastore",		0,	3,	0,	op_aastore } /* 83 */,
  { "bastore",		0,	3,	0,	op_bastore } /* 84 */,
  { "castore",		0,	3,	0,	op_castore } /* 85 */,
  { "sastore",		0,	3,	0,	op_sastore } /* 86 */,
  { "pop",		0,	1,	0,	op_pop }/* 87 */,
  { "pop2",		0,	2,	0,	op_pop2 } /* 88 */,
  { "dup",		0,	1,	1,	op_dup } /* 89 */,
  { "dupx1",		0,	2,	3,	op_dupx1 } /* 90 */,
  { "dupx2",		0,	3,	4,	op_dupx2 } /* 91 */,
  { "dup2",		0,	2,	4,	op_dup2 } /* 92 */,
  { "dup2x1",		0,	3,	5,	op_dup2x1 } /* 93 */,
  { "dup2x2",		0,	4,	6,	op_dup2x2 } /* 94 */,
  { "swap",		0,	2,	2,	op_swap } /* 95 */,
  { "iadd",		0,	2,	1,	op_iadd } /* 96 */,
  { "ladd",		0,	4,	2,	op_ladd } /* 97 */,
  { "fadd",		0,	2,	1,	op_fadd } /* 98 */,
  { "dadd",		0,	4,	2,	op_dadd } /* 99 */,
  { "isub",		0,	2,	1,	op_isub } /* 100 */,
  { "lsub",		0,	4,	2,	op_lsub } /* 101 */,
  { "fsub",		0,	2,	1,	op_fsub } /* 102 */,
  { "dsub",		0,	4,	2,	op_dsub } /* 103 */,
  { "imul",		0,	2,	1,	op_imul } /* 104 */,
  { "lmul",		0,	4,	2,	op_lmul } /* 105 */,
  { "fmul",		0,	2,	1,	op_fmul } /* 106 */,
  { "dmul",		0,	4,	2,	op_dmul } /* 107 */,
  { "idiv",		0,	2,	1,	op_idiv } /* 108 */,
  { "ldiv",		0,	4,	2,	op_ldiv } /* 109 */,
  { "fdiv",		0,	2,	1,	op_fdiv } /* 110 */,
  { "ddiv",		0,	4,	2,	op_ddiv } /* 111 */,
  { "irem",		0,	2,	1,	op_irem } /* 112 */,
  { "lrem",		0,	4,	2,	op_lrem } /* 113 */,
  { "frem",		0,	2,	1,	op_frem } /* 114 */,
  { "drem",		0,	4,	2,	op_drem } /* 115 */,
  { "ineg",		0,	1,	1,	op_ineg } /* 116 */,
  { "lneg",		0,	2,	2,	op_lneg } /* 117 */,
  { "fneg",		0,	1,	1,	op_fneg } /* 118 */,
  { "dneg",		0,	2,	2,	op_dneg } /* 119 */,
  { "ishl",		0,	2,	1,	op_ishl } /* 120 */,
  { "lshl",		0,	3,	2,	op_lshl } /* 121 */,
  { "ishr",		0,	2,	1,	op_ishr } /* 122 */,
  { "lshr",		0,	3,	2,	op_lshr } /* 123 */,
  { "iushr",		0,	2,	1,	op_iushr } /* 124 */,
  { "lushr",		0,	3,	2,	op_lushr } /* 125 */,
  { "iand",		0,	2,	1,	op_iand } /* 126 */,
  { "land",		0,	4,	2,	op_land } /* 127 */,
  { "ior",		0,	2,	1,	op_ior } /* 128 */,
  { "lor",		0,	4,	2,	op_lor } /* 129 */,
  { "ixor",		0,	2,	1,	op_ixor } /* 130 */,
  { "lxor",		0,	4,	2,	op_lxor } /* 131 */,
  { "iinc",		0,	0,	0,	op_iinc } /* 132 */,
  { "i2l",		0,	1,	2,	op_i2l } /* 133 */,
  { "i2f",		0,	1,	1,	op_i2f } /* 134 */,
  { "i2d",		0,	1,	2,	op_i2d } /* 135 */,
  { "l2i",		0,	2,	1,	op_l2i } /* 136 */,
  { "l2f",		0,	2,	1,	op_l2f }  /* 137 */,
  { "l2d",		0,	2,	2,	op_l2d } /* 138 */,
  { "f2i",		0,	1,	1,	op_f2i } /* 139 */,
  { "f2l",		0,	1,	2,	op_f2l } /* 140 */,
  { "f2d",		0,	1,	2,	op_f2d } /* 141 */,
  { "d2i",		0,	2,	1,	op_d2i } /* 142 */,
  { "d2l",		0,	2,	2,	op_d2l } /* 143 */,
  { "d2f",		0,	2,	1,	op_d2f } /* 144 */,
  { "int2byte",		0,	1,	1,	op_int2byte } /* 145 */,
  { "int2char",		0,	1,	1,	op_int2char } /* 146 */,
  { "int2short",	0,	1,	1,	op_int2short } /* 147 */,
  { "lcmp",		0,	4,	1,	op_lcmp } /* 148 */,
  { "fcmpl",		0,	2,	1,	op_fcmpl } /* 149 */,
  { "fcmpg",		0,	2,	1,	op_fcmpg } /* 150 */,
  { "dcmpl",		0,	4,	1,	op_dcmpl } /* 151 */,
  { "dcmpg",		0,	4,	1,	op_dcmpg } /* 152 */,
  { "ifeq",		2,	1,	0,	op_ifeq } /* 153 */,
  { "ifne",		2,	1,	0,	op_ifne } /* 154 */,
  { "iflt",		2,	1,	0,	op_iflt } /* 155 */,
  { "ifge",		2,	1,	0,	op_ifge } /* 156 */,
  { "ifgt",		2,	1,	0,	op_ifgt } /* 157 */,
  { "ifle",		2,	1,	0,	op_ifle } /* 158 */,
  { "if_icmpeq",	2,	2,	0,	op_ificmpeq } /* 159 */,
  { "if_icmpne",	2,	2,	0,	op_ificmpne } /* 160 */,
  { "if_icmplt",	2,	2,	0,	op_ificmplt } /* 161 */,
  { "if_icmpge",	2,	2,	0,	op_ificmpge } /* 162 */,
  { "if_icmpgt",	2,	2,	0,	op_ificmpgt } /* 163 */,
  { "if_icmple",	2,	2,	0,	op_ificmple } /* 164 */,
  { "if_acmpeq",	2,	2,	0,	op_ifacmpeq } /* 165 */,
  { "if_acmpne",	2,	2,	0,	op_ifacmpne } /* 166 */,
  { "goto",		2,	0,	0,	op_goto } /* 167 */,
  { "jsr",		2,	0,	1,	op_jsr } /* 168 */,
  { "ret",		1,	0,	0,	op_ret } /* 169 */,
  { "tableswitch",	-1,	1,	0,	op_tableswitch } /* 170 */,
  { "lookupswitch",	-1,	1,	0,	op_lookupswitch } /* 171 */,
  { "ireturn",		0,	1,	0,	op_ireturn } /* 172 */,
  { "lreturn",		0,	2,	0,	op_lreturn } /* 173 */,
  { "freturn",		0,	1,	0,	op_freturn } /* 174 */,
  { "dreturn",		0,	2,	0,	op_dreturn } /* 175 */,
  { "areturn",		0,	1,	0,	op_areturn } /* 176 */,
  { "return",		0,	0,	0,	op_return } /* 177 */,
  { "getstatic",	2,	0,	-1,	op_getstatic } /* 178 */,
  { "putstatic",	2,	-1,	0,	op_putstatic } /* 179 */,
  { "getfield",		2,	1,	-1,	op_getfield } /* 180 */,
  { "putfield",		2,	-1,	0,	op_putfield } /* 181 */,
  { "invokevirtual",	2,	-1,	0,	op_invokevirtual } /* 182 */,
  { "invokenonvirtual", 2,	-1,	0,	op_invokenonvirtual } /* 183 */,
  { "invokestatic",	2,	-1,	0,	op_invokestatic } /* 184 */,
  { "invokeinterface",	4,	-1,	0,	op_invokeinterface } /* 185 */,
  { "notimpl",		0,	0,	0,	op_notimplemented } /* 186 */,
  { "new",		1,	0,	1,	op_new } /* 187 */,
  { "newarray",		1,	1,	1,	op_newarray } /* 188 */,
  { "anewarray",	2,	1,	1,	op_anewarray } /* 189 */,
  { "arraylength",	0,	1,	1,	op_arraylength } /* 190 */,
  { "athrow",		0,	1,	1,	op_athrow } /* 191 */,
  { "checkcast",	2,	1,	1,	op_checkcast } /* 192 */,
  { "instanceof",	2,	1,	1,	op_instanceof } /* 193 */,
  { "monitorenter", 	0,	1,	0,	op_monitorenter } /* 194 */,
  { "monitorexit",	0,	1,	0,	op_monitorexit } /* 195 */,
  { "wide",		-1,	-1,	-1,	op_wide } /* 196 */,
  { "multianewarray",	3,	-1,	1,	op_multianewarray } /* 197 */,
  { "ifnull",		2,	1,	1,	op_ifnull } /* 198 */,
  { "ifnonnull",	2,	1,	1,	op_ifnonnull } /* 199 */,
  { "goto_w",		4,	0,	0,	op_goto_w } /* 200 */,
  { "jsr_w",		4,	0,	1,	op_jsr_w } /* 201 */,
  { "breakpoint",	0,	0,	0,	op_breakpoint } /* 202 (r) */,
  /*
   * The _quick methods should probably not be implemented.  Doing so
   * might give us problems with some of JavaSoft's patents.  Chapter
   * 9 of the Java Virtual Machine Specification informs that the
   * _quick optimization technique is covered by U.S. Patent
   * 5,367,685.
   */
  { "ldc_quick",	0,	0,	0,	op_notimplemented } /* 203 */,
  { "ldc_w_quick",	0,	0,	0,	op_notimplemented } /* 204 */,
  { "ldc2_w_quick",	0,	0,	0,	op_notimplemented } /* 205 */,
  { "getfield_quick",	0,	0,	0,	op_notimplemented } /* 206 */,
  { "putfield_quick",	0,	0,	0,	op_notimplemented } /* 207 */,
  { "getfield2_quick",	0,	0,	0,	op_notimplemented } /* 208 */,
  { "putfield2_quick",	0,	0,	0,	op_notimplemented } /* 209 */,
  { "getstatic_quick",	0,	0,	0,	op_notimplemented } /* 210 */,
  { "putstatic_quick",	0,	0,	0,	op_notimplemented } /* 211 */,
  { "getstatic2_quick",	0,	0,	0,	op_notimplemented } /* 212 */,
  { "putstatic2_quick",	0,	0,	0,	op_notimplemented } /* 213 */,
  { "invokevirtual_quick", 0,	0,	0,	op_notimplemented } /* 214 */,
  { "invokenonvirtual_quick", 0, 0,	0,	op_notimplemented } /* 215 */,
  { "invokesuper_quick", 0,	0,	0,	op_notimplemented } /* 216 */,
  { "invokestatic_quick", 0,	0,	0,	op_notimplemented } /* 217 */,
  { "invokeinterface_quick", 0,	0,	0,	op_notimplemented } /* 218 */,
  { "invokevirtualobject_quick", 0, 0,	0,	op_notimplemented } /* 219 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 220 */,
  { "new_quick",	0,	0,	0,	op_notimplemented } /* 221 */,
  { "anewarray_quick",	0,	0,	0,	op_notimplemented } /* 222 */,
  { "multianewarray_quick", 0,	0,	0,	op_notimplemented } /* 223 */,
  { "checkcast_quick",	0,	0,	0,	op_notimplemented } /* 224 */,
  { "instanceof_quick",	0,	0,	0,	op_notimplemented } /* 225 */,
  { "invokevirtual_quick_w", 0,	0,	0,	op_notimplemented } /* 226 */,
  { "getfield_quick_w", 0,	0,	0,	op_notimplemented } /* 227 */,
  { "putfield_quick_w",	0,	0,	0,	op_notimplemented } /* 228 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 229 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 230 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 231 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 232 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 233 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 234 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 235 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 236 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 237 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 238 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 239 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 240 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 241 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 242 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 243 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 244 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 245 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 246 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 247 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 248 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 249 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 250 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 251 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 252 */,
  { "unknown",		0,	0,	0,	op_notimplemented } /* 253 */,
  { "impdep1",		0,	0,	0,	op_notimplemented } /* 254 (r) */,
  { "impdep2",		0,	0,	0,	op_notimplemented } /* 255 (r) */,
};

StackFrame *
push_frame(JThreadInfo *thr_info, int num_vars)
{
  int new_frame_depth;
  int new_frame_size;
  StackFrame *new_stack_frame;
  JNIEnv *env;

  if (thr_info->top_frame == thr_info->stack_highwater)
    {
      new_frame_depth = 1;
      env = THREAD_getEnv();
    }
  else
    {
      new_frame_depth = TOPFRAME(thr_info)->depth + 1;
      env = TOPFRAME(thr_info)->jni_env;
    }

  new_frame_size = sizeof(StackFrame) + sizeof(JavaStackItem) * num_vars;
  new_stack_frame = (StackFrame*)((char*)TOPFRAME(thr_info) - new_frame_size);

  if (new_stack_frame < thr_info->stack_lowwater)
    {
      /* do something special... throw a runtime exception or something. */
    }
  else
    {
      new_stack_frame->depth = new_frame_depth;
      new_stack_frame->frame_size = new_frame_size;
      new_stack_frame->thread_info = thr_info;

      new_stack_frame->flags = 0;

      ENV(new_stack_frame) = env;
	  
      TOPFRAME(thr_info) = new_stack_frame;
    }

  JAVARLOG1(MYLOG, 1, "Pushed new frame: %p\n", new_stack_frame);
  JAVARLOG1(MYLOG, 1, "\t  now at depth %d\n", STACKDEPTH(TOPFRAME(thr_info)));

  return new_stack_frame;
}

StackFrame *
get_frame_parent(StackFrame *frame)
{
  return (StackFrame*)((char*)frame + frame->frame_size);
}

void
maybe_enter_monitor_for_method(JNIEnv *env,
			       MethodStruct *method,
			       obj_thunk *this_obj)
{
  if (method->access_flags & ACC_SYNCHRONIZED)
    {
      HMonitor mon;

      if (this_obj == NULL)
	this_obj = new_classobject(env, method->clazz);
      
      mon = MONITOR_getForObject(this_obj);
      MONITOR_enter(mon);

      JAVARLOG3(MYLOG, 4, "enter monitor for method %s.%s (this_obj = %p)\n",
		getClassName(env, method->clazz),
		method->name, this_obj);
    }
}

void
maybe_exit_monitor_for_method(JNIEnv *env,
			      MethodStruct *method,
			      obj_thunk *this_obj)
{
  if (method->access_flags & ACC_SYNCHRONIZED)
    {
      HMonitor mon;

      if (this_obj == NULL)
	this_obj = new_classobject(env, method->clazz);
      
      mon = MONITOR_getForObject(this_obj);
      MONITOR_exit(mon);

      JAVARLOG3(MYLOG, 4, "exit  monitor for method %s.%s (this_obj = %p)\n",
		getClassName(env, method->clazz),
		method->name, this_obj);
    }
}

void
pop_frame(JThreadInfo *thr_info)
{
#ifdef DEBUG
  /* XXX TOPFRAME(thr_info) used to be StackFrame *.  Correct fix? [pere] */
  if (TOPFRAME(thr_info)->flags & FRAME_NATIVE)
    JAVARLOG1(MYLOG, 1, "\tPopping native frame: %p\n", TOPFRAME(thr_info));
  else
    JAVARLOG1(MYLOG, 1, "\tPopping java frame: %p\n", TOPFRAME(thr_info));
#endif

  maybe_exit_monitor_for_method(ENV(TOPFRAME(thr_info)),
				METHOD(TOPFRAME(thr_info)),
				THIS(TOPFRAME(thr_info)));

  TOPFRAME(thr_info) = get_frame_parent(TOPFRAME(thr_info));
  
  if (TOPFRAME(thr_info))
    JAVARLOG1(MYLOG, 1, "\t  now at depth %d\n", STACKDEPTH(TOPFRAME(thr_info)));
}

void execute_opcode(StackFrame *f, u1 oc)
{
  opcode *op;

  op = &(opcode_table[ oc ]);

  /* check the stack constraints for the opcode. */

  if (op->num_stack_items_before != -1
      && op->num_stack_items_after != -1)
    {
      op_stack_check_bottomspace(ENV(f),
				 OPSTACK(f),
				 op->num_stack_items_before);

      if (op->num_stack_items_before != op->num_stack_items_after)
	op_stack_check_topspace(ENV(f),
				OPSTACK(f),
				op->num_stack_items_after - op->num_stack_items_before);
    }

  (*op->func)(f);
}

StackFrame *
create_frame_for_method (JThreadInfo *thr_info, MethodStruct *method)
{
  StackFrame *new_frame = push_frame(thr_info, method->max_locals + (method->access_flags & ACC_STATIC ? 0 : 1));

  METHOD(new_frame) = method;
  METHODNAME(new_frame) = method->name;
  CLASSNAME(new_frame) = getClassName(new_frame->jni_env, method->clazz);

  PC(new_frame) = 0;

  return new_frame;
}

void
interp_loop(StackFrame *f)
{
  u1 opcode;
  HungryJNIEnv *henv = (HungryJNIEnv*)ENV(f);
  JThreadInfo *thread_info = THREAD_INFO(f);
  int save_depth = f->depth;
  JAVARLOG1(MYLOG, 3, "in interp_loop(0x%x)\n", f);

  do {
    jthrowable exception;

    opcode = get_next_byte(TOPFRAME(thread_info));

    execute_opcode(TOPFRAME(thread_info), opcode);

    if ((exception = henv->_exception /* optimized from (*env)->ExceptionOccurred(env)*/))
      {
	throw_exception(ENV(f), exception, TOPFRAME(thread_info));
		  
	/* the only way the exception is still set is if we
	   need to throw it from java to native code.  Code
	   that calls interp_loop (or sub_mainloop for that 
	   matter) should handle this.  We just return and
	   let them deal with it. */
	if (henv->_exception /* optimized from (*env)->ExceptionOccurred(env) */)
	  {
	    JAVARLOG0("Exceptions", 1, "Exception thrown from java code to native code.");
	    return;
	  }
      }
  }
  while (TOPFRAME(thread_info) != thread_info->stack_highwater
	 && TOPFRAME(thread_info)->depth >= save_depth);
}

static void 
op_notimplemented(StackFrame *f)
{
  JNIEnv *env = ENV(f);

  printf ("Instruction %d not implemented yet :)\n", 
	  CODE(f)[ PC(f)-1 ]);

  (*env)->FatalError(env, "Instruction not implemented\n");
}

/***/

void
fill_local_vars_from_stack (StackFrame *df, /* destination frame */
                            int num_vars_needed, /* number of parameter words from the signature */
			    int static_p) /* is this a static call ? */
{
  int j;

  if (!static_p)
    num_vars_needed ++;

  for (j = num_vars_needed - 1; j >=0; j --)
    {
      op_stack_pop_any (ENV(df), OPSTACK (df), &(VARS(df)[j]));
    }

  /* the object reference is implicit, unless static. */
  if (!static_p)
    {
      assert(VARS(df)[0].tag == SIG_JOBJECT);

      THIS(df) = VARS(df)[0].value.l;
    }
  else
    {
      THIS(df) = NULL;
    }
  return;
}

void fill_local_vars (StackFrame *df, /* destination frame */
		      MethodStruct *method,
		      jvalue *args, /* must match the number in the signature. */
		      jobject obj) /* NULL for static calls */
{
  int i;

  if (obj)
    {
      VARS(df)[0].tag = SIG_JOBJECT;
      VARS(df)[0].value.l = obj;
    }
  
  for (i = method->num_param_words - 1; i >= 0; i --)
    {
      SigPrimType sig;

      if (method->sig->method.params[i]->any.tag == SIG_PRIM)
	sig = method->sig->method.params[i]->prim.type;
      else
	sig = SIG_JOBJECT;

      VARS(df)[i + (obj ? 1 : 0)].tag = sig;
      VARS(df)[i + (obj ? 1 : 0)].value.i = args[i].i;
    }
  
  THIS(df) = obj;
  
  return;
}

void
initialize_system_threads(JNIEnv *env)
{
  jclass ThreadGroup;
  jclass Thread;
  jobject system_thread_group;
  jobject system_thread;
  StackFrame *initial_native_frame;
  JThreadInfo *initial_native_thread_info;
  jfieldID field;
  jmethodID method;
  jstring thread_name;
  jint norm_priority;

  initial_native_thread_info = (JThreadInfo*)calloc(1, sizeof(JThreadInfo));
  initial_native_thread_info->name = strdup("main");
  initial_native_thread_info->op_stack = op_stack_allocate(OPSTACK_SIZE);

  initial_native_thread_info->stack_lowwater = (StackFrame*)malloc(sizeof(StackFrame) * 1000);
  assert(initial_native_thread_info->stack_lowwater);
  if (!initial_native_thread_info)
    {
      throw_Exception(env, "java/lang/VirtualMachineError",
		      "Cannot allocate thread-local stack");
      return;
    }

  initial_native_thread_info->stack_highwater =	initial_native_thread_info->stack_lowwater;
  initial_native_thread_info->stack_highwater += 1000;
  initial_native_thread_info->top_frame = initial_native_thread_info->stack_highwater;

  THREAD_setJavaInfo(initial_native_thread_info);

  Thread = find_class(env, "java/lang/Thread");
  ThreadGroup = find_class(env, "java/lang/ThreadGroup");

  if ((!Thread) || (!ThreadGroup))
    {
      fprintf(stderr,
	      "Unable to initialize threads: cannot find class java/lang/%s\n",
	      ((!Thread) ? "Thread" : "ThreadGroup"));
      exit(1);
    }

  system_thread_group = new_object(env, ThreadGroup);
  system_thread = new_object(env, Thread);

  /* have to put this off until here. */
  initial_native_thread_info->java_thread = system_thread;

  GC_addReference((JavaVM*)((HungryJNIEnv*)env)->_vm, (GC_obj*)system_thread_group);
  GC_addReference((JavaVM*)((HungryJNIEnv*)env)->_vm, (GC_obj*)system_thread);

#if norm_priority_bogus
  field = (*env)->GetStaticFieldID(env, Thread, "NORM_PRIORITY", "I");
  norm_priority = (*env)->GetStaticIntField(env, Thread, field);
#else
  norm_priority = 10;
#endif
  field = (*env)->GetFieldID(env, Thread, "priority", "I");
  (*env)->SetIntField(env, system_thread, field, norm_priority);
  field = (*env)->GetFieldID(env, Thread, "PrivateInfo", "I");

  /* XXX pointer will be truncated on 64bits systems */
  assert(sizeof(jint)>=sizeof(HThread));
  (*env)->SetIntField(env, system_thread, field, (jint)THREAD_getCurrent());

  initial_native_frame = push_frame(initial_native_thread_info,
				    0);

  OPSTACK_TOP(initial_native_frame) = OPSTACK(initial_native_frame)->stack_top;

  initial_native_frame->flags |= FRAME_NATIVE;

  method = (*env)->GetMethodID(env, ThreadGroup, "<init>", "()V");
  initial_native_frame->method = method;
  (*env)->CallVoidMethod(env, system_thread_group, method);

  thread_name = (*env)->NewStringUTF(env, "main");
  (*env)->NewGlobalRef(env, thread_name);

  method = (*env)->GetMethodID(env, Thread, "<init>", "(Ljava/lang/ThreadGroup;Ljava/lang/String;)V");
  initial_native_frame->method = method;
  (*env)->CallVoidMethod(env, system_thread, method, system_thread_group, thread_name);

  pop_frame(initial_native_thread_info);
}

void
set_instance_field(obj_thunk *obj,
		   FieldStruct *field,
		   JavaStackItem value)
{
  assert((field->access_flags & ACC_STATIC) == 0
	 && field->clazz == STATIC_DATA(obj));

  /*  assert(value.tag == field->java_type);*/

  obj->_instance->fields[ field->field_index ] = value;
}

void
set_static_field(ClazzFile *clazz,
		 FieldStruct *field,
		 JavaStackItem value)
{
  assert(field->access_flags & ACC_STATIC
	 && field->clazz == clazz);

  /*  assert(value.tag == field->java_type);*/

  clazz->static_fields[ field->field_index ] = value;
}

void
get_instance_field(obj_thunk *obj,
		   FieldStruct *field,
		   JavaStackItem *value)
{
  *value = obj->_instance->fields[ field->field_index ];
}

void
get_static_field(ClazzFile *clazz,
		 FieldStruct *field,
		 JavaStackItem *value)
{
  *value = clazz->static_fields[ field->field_index ];
}
