/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "config.h"
#include "qop.h"
#include "jniint.h"
#include "log.h"
#include "dynamic_loading.h"
#include "exceptions.h"
#include "interp.h"
#include "sig.h"
#include "ClazzFile.h"

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <assert.h>

#define MYLOG "NativeGlue"

/*
 * Convert (any/class/Name, method) to Java_any_class_Name_method.
 * The returned char * is allocated off the heap using strdup and must
 * be free'ed.
 */
static char*
create_native_func_name(char *classname, char *method_name)
{
  char name_buf[1024];
  char *bar_classname = strdup(classname);
  int i;
  int retval;

  slashes_to_underscores(bar_classname);

  retval = snprintf(name_buf, sizeof(name_buf),
		    "Java_%s_%s_stub",
		    bar_classname, method_name);
  free(bar_classname);

  if (-1 == retval)
    { /* XXX Output was truncated.  What should we do? */
      fprintf(stderr, "Error: Class method %s.%s truncated.\n",
	      classname, method_name);
    }

  return strdup(name_buf);
}

static void
resolve_native_func(JNIEnv *env, MethodStruct *method)
{
  char *class_name = getClassName(env, method->clazz);
  char *native_func_name = create_native_func_name(class_name, method->name);
  
  JAVARLOG1(MYLOG, 1, "   native stub name is %s()\n", native_func_name);
  
  method->native_func = (NativeFunc)DLL_findFunction(native_func_name);
  
  if (method->native_func == NULL) /* what do we do here? */
    {
      char *msg;
      
      msg = (char*)malloc(sizeof(char) 
			  * (strlen("Unresolved Function: ")
			     + strlen(native_func_name) + 1));
      sprintf(msg, "Unresolved Function: %s", native_func_name);
      
      (*env)->FatalError(env, msg);
    }
  
  free(native_func_name);
}

static jvalue
do_native_method_call_with_args(JNIEnv *env,
				MethodStruct *method,
				jvalue *args)
{
  HungryJNIEnv *henv = (HungryJNIEnv*)env;
  GC_obj *head_obj;
  char *class_name = getClassName(env, method->clazz);
  StackFrame *new_frame = push_frame(THREAD_getJavaInfo(),
				     0);

  new_frame->flags |= FRAME_NATIVE;
  METHOD(new_frame) = method;
  CLASSNAME(new_frame) = class_name;
  METHODNAME(new_frame) = (char *) malloc (strlen (METHOD(new_frame)->name)+1);
  strcpy (METHODNAME(new_frame), METHOD(new_frame)->name);

  /* get the function pointer for this native method, if we don't already
     know it. */
  if (!method->native_func)
    resolve_native_func(env, method);

  OPSTACK_TOP(new_frame) = OPSTACK(new_frame)->stack_top;

  if (method->access_flags & ACC_STATIC)
    args[1].l = method->clazz;
  else
    {
      /* make sure the "this" pointer to a native method is of the correct type. */
      args[1].l = find_instance_data_for_class(args[1].l, method->clazz);
    }

  if (method->access_flags & ACC_STATIC)
    THIS(new_frame) = NULL;
  else
    THIS(new_frame) = args[1].l;

  /* save off the stopping point on thread-local deletions */
  head_obj = henv->_local_refs;

  maybe_enter_monitor_for_method(ENV(new_frame), METHOD(new_frame), THIS(new_frame));

  (*METHOD(new_frame)->native_func)(env, args);

#ifdef garbage_collect_working
  /* collect all the objects that were allocated during this native call */
  {
    while (henv->_local_refs
	   && henv->_local_refs != head_obj)
      {
	GC_obj *foo = henv->_local_refs;
	GC_obj *next = foo->next;

	UNQUEUE(foo, henv->_local_refs);

	GC_deleteThreadLocal((JavaVM*)henv->_vm, foo);
	henv->_local_refs = next;
      }
  }
#endif

  if (!SIG_isVoid(env, METHOD(new_frame)->sig->method.return_type))
    switch(METHOD(new_frame)->sig->method.return_type->any.tag)
      {
      case SIG_PRIM:
	THREAD_INFO(new_frame)->return_value.tag = METHOD(new_frame)->sig->method.return_type->prim.type;
	break;
      case SIG_ARRAY:
      case SIG_CLASS:
	THREAD_INFO(new_frame)->return_value.tag = SIG_JOBJECT;
	break;
      case SIG_METHOD:
	assert(0); /* this shouldn't ever happen. */
	break;
      }

  THREAD_INFO(new_frame)->return_value.value = args[0];

  maybe_push_return_value(new_frame);
  
  /* then pop the frame off */
  pop_frame(THREAD_INFO(new_frame));

  return THREAD_INFO(new_frame)->return_value.value;
}

void
do_native_method_call(JNIEnv *env, MethodStruct *method)
{
  int num_arguments = 0;
  jvalue args[100];
  int i;
  char *class_name = getClassName(env, method->clazz);
  StackFrame *current_frame = TOPFRAME((JThreadInfo*)THREAD_getJavaInfo());

  JAVARLOG2(MYLOG, 1, "Inside do_native_method_call(%s.%s)\n",
	    class_name, method->name);

  num_arguments = SIG_numParams(env, method->sig) + 1;

  for (i = num_arguments;
       i >= ((method->access_flags & ACC_STATIC) ? 2 : 1);
       i --)
    {
      op_stack_pop_value(env, OPSTACK(current_frame), &args[ i ]);
    }

  do_native_method_call_with_args(env, method, args);
}

/* This is the generic method used by the two specific below. */
static jvalue
call_java_method(MethodStruct *method,
		 jobject obj, /* NULL for static calls */
		 jvalue *args)
{
  StackFrame *new_frame;

  JAVARLOG2(MYLOG, 4, "calling %s.%s()\n",
	    method->clazz->class_name,
	    method->name);

  new_frame = create_frame_for_method(THREAD_getJavaInfo(), method);

  fill_local_vars(new_frame, method, args, obj);

  OPSTACK_TOP(new_frame) = OPSTACK(new_frame)->stack_top;

  THREAD_INFO(new_frame)->return_value.value.j = 0;

  maybe_enter_monitor_for_method(ENV(new_frame), method, obj);

  interp_loop(new_frame);

  OPSTACK(new_frame)->stack_top = OPSTACK_TOP(new_frame);

  return THREAD_INFO(new_frame)->return_value.value;
}
  
jvalue
CallJavaMethod (MethodStruct *method,
		jobject obj,
		jvalue *args)
{
  if (method->access_flags & ACC_NATIVE)
    {
      jvalue new_args[100];
      int num_arguments = SIG_numParams(THREAD_getEnv(), method->sig) + 1;
      int i;

      for (i = num_arguments;
	   i >= 1;
	   i --)
	{
	  new_args[i] = args[i - 1];
	}

      new_args[1].l = obj;

      return do_native_method_call_with_args(THREAD_getEnv(),
					     method, new_args);
    }
  else
    {
      return call_java_method(method, obj, args);
    }
}

jvalue
CallStaticJavaMethod (MethodStruct *method,
		      jclass clazz, /* XXX Why is this here?  It is unused. */
		      jvalue *args)
{
  if (method->access_flags & ACC_NATIVE)
    {
      jvalue new_args[100];
      int num_arguments = SIG_numParams(THREAD_getEnv(), method->sig) + 1;
      int i;

      for (i = num_arguments;
	   i >=  2;
	   i --)
	{
	  new_args[i] = args[i - 1];
	}

      new_args[1].l = 0;

      return do_native_method_call_with_args(THREAD_getEnv(),
					     method, new_args);
    }
  else
    {
      return call_java_method(method, 0, args);
    }
}
