/* -*- Mode: C; c-file-style: "gnu" -*-
   objects.c -- runtime code for allocating objects.
   Created: Chris Toshok <toshok@hungry.com>, 10-Aug-1997
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "config.h"
#include "ClazzFile.h"
#include "class-repository.h"
#include "resolve.h"
#include "log.h"
#include "sig.h"
#include "objects.h"
#include "native-threads.h"
#include "exceptions.h" /* for throw_Exception() */

#include <stdlib.h>
#include <string.h>

#define MYLOG "ObjectRef"

static GC_obj gc_template = {
  NULL, NULL,
  GC_FLAG_OBJECT,

  sizeof(obj_thunk*),

  object_finalize
};

static void
index_fields(ClazzFile *cf)
{
  int i;
  int static_field_index = 0;
  int instance_field_index = 0;

  for (i = 0; i < cf->num_fields; i ++)
    {
      FieldStruct *field = cf->fields[i];
	  
      if (field->field_index != VALUE_UNRESOLVED)
	{
	  /* if we've resolved any fields, we've resolved them already. */
	  return;
	}
     
      if (field->access_flags & ACC_STATIC)
	field->field_index = static_field_index++;
      else
	field->field_index = instance_field_index++;

      if (field->sig->any.tag == SIG_CLASS || field->sig->any.tag == SIG_ARRAY)
        field->java_type = SIG_JOBJECT;
      else
        field->java_type = field->sig->prim.type;
    }
}

static void
copy_field_types(ClazzFile *cf,
		 JavaStackItem *field_items,
		 jboolean instance)
{
  int i;

  for (i = 0; i < cf->num_fields; i ++)
    {
      FieldStruct *field = cf->fields[i];

      if ((instance && !(field->access_flags & ACC_STATIC)) /* we're looking for instance fields,
							       and it is one. */
	  || (!instance && (field->access_flags & ACC_STATIC))) /* we're looking for static fields,
								   and it is one. */
	field_items[ field->field_index ].tag = field->java_type;
    }
}

obj_thunk*
clone_object(JNIEnv *env, obj_thunk *old_object)
{
  obj_thunk *new_obj, *obj_ref, *new_obj_ref;

  obj_ref = old_object;

  while (!IS_POST(obj_ref))
    obj_ref --;

  obj_ref ++;

  new_obj = new_object(env, STATIC_DATA(obj_ref));

  new_obj_ref = new_obj;

  while (!IS_POST(new_obj_ref))
    {
      ClazzFile *current_class = STATIC_DATA(new_obj_ref);
      int i;

      for (i = 0; i < current_class->num_instance_fields; i++)
	{
	  INSTANCE_DATA(new_obj_ref)->fields[i] = INSTANCE_DATA(obj_ref)->fields[i];
	}

      new_obj_ref ++;
      obj_ref ++;
    }

  return new_obj;
}

obj_thunk*
new_object(JNIEnv *env, ClazzFile *cf)
{
  int num_classes_in_heir;
  ClazzFile *current_class = cf;
  obj_thunk* new_obj, *obj_ref;

  static int num_objects = 0;

  JAVARLOG1(MYLOG, 1, "in new_object(%d)\n", ++num_objects);

  num_classes_in_heir = 0;

  /* first we initialize the class. */
  initialize_class(env, current_class);
  if ((*env)->ExceptionOccurred(env))
    {
      return NULL;
    }

  /* first we resolve the fields of all the classes,
     starting at this one and going back up to Object. */
  while (current_class != NULL)
    {
      index_fields(current_class);
      num_classes_in_heir ++;

      if (current_class->superclass_index == 0)
	current_class = NULL;
      else
	current_class = getSuperClass(env, current_class);
    };

  /* now we allocate space enough for two posts, 
     all the classes in the heir. */
  new_obj = (obj_thunk*)calloc(num_classes_in_heir + 2,
			       sizeof(obj_thunk));

  /* make the fence posts. */
  MAKE_POST(new_obj + 0);
  MAKE_POST(new_obj + num_classes_in_heir + 1);

  /* now we step through and setup up all the instance and static
     data parts for the class portion of the object */
  current_class = cf;
  obj_ref = (obj_thunk*)new_obj + 1;

  memcpy(&obj_ref->_gc_obj, &gc_template, sizeof(GC_obj));

  while (current_class != NULL)
    {
      INSTANCE_DATA(obj_ref) = (instance_data*)calloc(1, sizeof(instance_data));
      if (current_class->num_instance_fields > 0)
	{
	  INSTANCE_DATA(obj_ref)->fields = (JavaStackItem*)calloc(current_class->num_instance_fields,
								  sizeof(JavaStackItem));

	  copy_field_types(current_class, INSTANCE_DATA(obj_ref)->fields, JNI_TRUE /* do instance fields */);
	}

      STATIC_DATA(obj_ref) = current_class;

      JAVARLOG3(MYLOG, 2, "object ref %p = { %p , %p }\n", 
		obj_ref,
		INSTANCE_DATA(obj_ref),
		STATIC_DATA(obj_ref));

      if (current_class->superclass_index == 0)
	current_class = NULL;
      else
	current_class = getSuperClass(env, current_class);
      obj_ref ++;
    };

  /* is this correct? */
  obj_ref ++;

  JAVARLOG1(MYLOG, 1, "leaving new_object(), returning %p\n",
	    (new_obj + 1));

  return (new_obj + 1);
}

obj_thunk *
new_classobject(JNIEnv *env, ClazzFile *cf)
{
  obj_thunk *cached_descriptor;

  cached_descriptor = find_descriptor(env, getClassName(env, cf));

  if (cached_descriptor)
    return cached_descriptor;
  else
    {
      /* This is sort of hardcoded.  We're going to create an object
	 which looks like this: 
	 
	 <POST>
	 -------  java.lang.Object
	 ID
	 --
	 SD
	 -------  java.lang.Class
	 ID
	 --
	 SD
	 -------
	 <POST>
	 
	 The instance data used for java.lang.Class is assumed to not
	 contain anything from the sun implementation (I mean, we assume
	 that they don't keep any data there.)  I've hardcoded our
	 instance for java.lang.Class to contain nothing but a ClazzFile
	 pointer, so the instance of java.lang.Class can get at the
	 class that it represents.
	 
	 This is all really, really gross -- such is the life of a 
	 runtime for a dynamic language...
	 */
      ClazzFile *object_cf, *class_cf;
      obj_thunk *new_obj = (obj_thunk*)calloc(2 /* for Class and Object */ 
					      + 2 /* for the posts. */,
					      sizeof(obj_thunk));
      initialize_class(env, cf);
      if ((*env)->ExceptionOccurred(env))
	{
	  free(new_obj);
	  return NULL;
	}

      /* make our posts. */
      MAKE_POST(new_obj + 0);
      MAKE_POST(new_obj + 3);
      
      /* make the object thunk for java.lang.Class */
      class_cf = find_class(env, "java/lang/Class");
      STATIC_DATA(new_obj + 2) = class_cf;
      INSTANCE_DATA(new_obj + 2) = (instance_data*)calloc(1, sizeof(instance_data));
      INSTANCE_DATA(new_obj + 2)->fields = (JavaStackItem*)calloc(1, sizeof(JavaStackItem));
      INSTANCE_DATA(new_obj + 2)->fields[0].tag = SIG_JOBJECT;
      INSTANCE_DATA(new_obj + 2)->fields[0].value.l = cf;

      /* make the object thunk for java.lang.object */
      object_cf = find_class(env, "java/lang/Object");
      STATIC_DATA(new_obj + 1) = object_cf;
      INSTANCE_DATA(new_obj + 1) = (instance_data*)calloc(1, sizeof(instance_data));
      if (object_cf->num_instance_fields > 0)
	INSTANCE_DATA(new_obj + 1)->fields = (JavaStackItem*)calloc(cf->num_instance_fields, sizeof(JavaStackItem));

      store_descriptor_in_repository(env, cf, new_obj + 2);
      return new_obj + 2;
    }
}

void
object_finalize(GC_obj *gc_obj)
{
  obj_thunk *obj = (obj_thunk*)gc_obj;
  JNIEnv *env = THREAD_getEnv();
  jmethodID finalize;
  obj_thunk *cur;
  
  cur = obj;
  while (!IS_POST(cur))
    cur --;

  cur++;

  JAVARLOG1(MYLOG, 2, "Finalizing object %p\n", cur);

  finalize = (*env)->GetMethodID(env,
				 STATIC_DATA(obj),
				 "finalize", "()V");

  if (finalize && !(*env)->ExceptionOccurred(env))
    {

      (*env)->CallVoidMethod(env, obj, finalize);
      if ((*env)->ExceptionOccurred(env))
	{
	  /* XXX do something... */
	  return;
	}
    }
  if ((*env)->ExceptionOccurred(env))
    (*env)->ExceptionClear(env);

  JAVARLOG0(MYLOG, 3, "   freeing memory.\n");

  cur = obj;
  while (!IS_POST(cur))
    {
      free(cur->_instance->fields);
      free(cur->_instance);
      cur++;
    }
  free(obj - 1);
}

void
create_static_fields(ClazzFile *cf)
{
  if (cf->num_static_fields == 0)
    cf->static_fields = NULL;
  else
    cf->static_fields = (void*)calloc(cf->num_static_fields, sizeof(JavaStackItem));

  copy_field_types(cf, cf->static_fields, JNI_FALSE /* do static fields */);
}

int
is_instance_of(obj_thunk *obj, ClazzFile *cf)
{
  while (!IS_POST(obj))
    {
      if (STATIC_DATA(obj) == cf)
	return 1;

      obj++;
    }

  return 0;
}

/* this is sort of a cross-bred function... it might belong here,
   or it might belong in the same file as get_method_info.  The
   only reason I put it here is because it uses the heap representation
   to do its job. */
MethodStruct *
get_interface_method_info(JNIEnv *env,
			  obj_thunk *obj,
			  ClazzFile **cf,
			  char *method_name,
			  Signature *sig)
{
  
  /* this is _HORRIBLY_ inefficient, since we can't store the information we need
     in the ClazzFile structure, since it's reused across all instances of the
     class.  fuck. */

  while (!IS_POST(obj))
    {
      MethodStruct *method = GetMethodByNameAndSig(env,
						   STATIC_DATA(obj),
						   method_name, sig);

      if (method)
	{
	  *cf = STATIC_DATA(obj);
	  return method;
	}
      else
	{
	  obj ++;
	}
    }

  throw_Exception(THREAD_getEnv(), "java/lang/RuntimeException", NULL);
  return NULL;
}

/* the guts of the below function comes from
   "The Java Language Specification" by Gosling et. al */
void
initialize_class(JNIEnv *env, ClazzFile *cf)
{
  HMonitor clazz_monitor;

  if (cf == NULL)
    return;

  if (!cf->monitor)
    cf->monitor = MONITOR_create();

  clazz_monitor = cf->monitor;

  MONITOR_enter(clazz_monitor);

try_and_initialize:
  switch (cf->initialization_state)
    {
    case CLASS_INIT_FAILED:
      MONITOR_exit(clazz_monitor);
      throw_Exception(env, "java/lang/NoClassDefFoundError", getClassName(env, cf));
      return;
    case CLASS_INIT_FINISHED:
      MONITOR_exit(clazz_monitor);
      return;
    case CLASS_INIT_IN_PROGRESS:
      if (cf->initializing_thread == THREAD_getCurrent())
	{
	  /* we're in the process of initializing it already,
	     so just return. */
	  MONITOR_exit(clazz_monitor);
	  return;
	}
      else
	{
	  /* someone else is initializing it, so wait until they're
	     done and try again. */
	  MONITOR_wait(clazz_monitor);
	  goto try_and_initialize;
	}
    case CLASS_NOT_INITIALIZED:
      {
	cf->initialization_state = CLASS_INIT_IN_PROGRESS;
	cf->initializing_thread = THREAD_getCurrent();

	/* we can index our fields before we start anything else,
	   as it depends purely on the access flag (ACC_STATIC or
	   not). */
	index_fields(cf);
	create_static_fields(cf);

	MONITOR_exit(clazz_monitor);

	/* initialize the superclass */
	load_parent_classes(env, cf);

	if (getSuperClass(env, cf))
	  initialize_class(env, getSuperClass(env, cf));
	if ((*env)->ExceptionOccurred(env))
	  {
	    MONITOR_enter(clazz_monitor);
	    cf->initialization_state = CLASS_INIT_FAILED;
	    cf->initializing_thread = NULL;

	    MONITOR_notifyAll(clazz_monitor);

	    MONITOR_exit(clazz_monitor);
	    /* don't clear the exception */
	    return;
	  }

	if (cf->access_flags & HAS_CLINIT)
	  {
	    jmethodID clinit;

	    /* call our static initializers */
	    clinit = (*env)->GetStaticMethodID(env, cf,
					       "<clinit>", "()V");
			
	    /* if the method isn't defined, clear the exception. */
	    if ((*env)->ExceptionOccurred(env))
	      {
		/* XXX there are some we should be legitimately concerned with,
		   like OutOfMemoryError, etc. */
		(*env)->ExceptionClear(env);
	      }
			
	    if (clinit)
	      (*env)->CallStaticVoidMethod(env, cf, clinit);
	  }

	/* the logic here isn't quite right... We're supposed to throw
	   an ExceptionInInitializerError.... */
	if ((*env)->ExceptionOccurred(env))
	  {
	    MONITOR_enter(clazz_monitor);
	    cf->initialization_state = CLASS_INIT_FAILED;
	    cf->initializing_thread = NULL;
	    
	    MONITOR_notifyAll(clazz_monitor);
	    
	    MONITOR_exit(clazz_monitor);
	    /* don't clear the exception */
	    return;
	  }

	/* if we've made it here, things worked. */
	MONITOR_enter(clazz_monitor);

	cf->initialization_state = CLASS_INIT_FINISHED;
	cf->initializing_thread = NULL;

	MONITOR_notifyAll(clazz_monitor);
	MONITOR_exit(clazz_monitor);
	return;
      }
    }
}

obj_thunk *
find_instance_data_for_class(obj_thunk *object_ref,
			     ClazzFile *cf)
{
  if (object_ref == NULL)
    {
      throw_Exception(THREAD_getEnv(), "java/lang/NullPointerException", NULL);
      return NULL;
    }

  /* shortcut */
  if (STATIC_DATA(object_ref) == cf) return object_ref;

  /* just to be safe, disregard the below comment.  We'll loop
     up to the top and then start down again. */
  /* object references always refer to the first thunk, so we
     just loop down until we find the right instance data. */

  while (!IS_POST(object_ref))
    object_ref --;

  object_ref ++;

  while (!IS_POST(object_ref))
    {
      if (STATIC_DATA(object_ref) == cf)
	return object_ref;
      else
	object_ref ++;
    }

  return NULL;
}
