/* -*- Mode: C; c-file-style: "gnu" -*-
   op_stack.c -- routines dealing with the operand stack.
   Created: Chris Toshok <toshok@hungry.com>, 10-Aug-1997
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "config.h"
#include "log.h"
#include "op_stack.h"
#include "native-threads.h"
#include "exceptions.h" /* for throw_Exception() */

#include <assert.h>
#include <stdlib.h>

#define OPSTACK_LOG "OpStack"

OpStack *
op_stack_allocate(int size)
{
  OpStack *new_stack = (OpStack*)malloc(sizeof(OpStack*));

  new_stack->stack_top = (JavaStackItem*)malloc(sizeof(JavaStackItem) * size);
  new_stack->stack_bottom = new_stack->stack_top;
  new_stack->high_water = new_stack->stack_bottom + size;

  return new_stack;
}

void
op_stack_deallocate(OpStack *stack)
{
  free(stack->stack_top);
  free(stack);
}

void
op_stack_check_bottomspace(JNIEnv *env, OpStack *s,
			   int space_needed)
{
  if (s->stack_top - space_needed < s->stack_bottom)
    {
      throw_Exception(env, "java/lang/VirtualMachineError", "Unexpected stack underflow.");
      return;
    }
}

void
op_stack_check_topspace(JNIEnv *env, OpStack *s,
			int space_needed)
{
  if (s->stack_top + space_needed >= s->high_water)
    {
      throw_Exception(env, "java/lang/StackOverflowError", NULL);
      return;
    }
}

inline int
promote_to_int(JNIEnv *env, JavaStackItem i)
{
  switch (i.tag)
    {
    case SIG_JINT:
      return i.value.i;
    case SIG_JCHAR:
      return (int)i.value.c;
    case SIG_JBYTE:
      return (int)i.value.b;
    case SIG_JSHORT:
      return (int)i.value.s;
    case SIG_JBOOLEAN:
      return (int)i.value.b;
    default:
      (*env)->FatalError(env, "illegal type attempting to convert to integer.");
      return 0;
    }
}

void
op_stack_push_any(JNIEnv *env, OpStack *s,
		  JavaStackItem *any)
{
  assert(s != NULL);

  if (any->tag == SIG_JDOUBLE)
    {
      op_stack_push_double(env, s, any->value.d);
    }
  else if (any->tag == SIG_JLONG)
    {
      op_stack_push_long(env, s, any->value.j);
    }
  else
    *(s->stack_top++) = *any;
}

void
op_stack_push_value(JNIEnv *env, OpStack *s,
		    SigPrimType tag,
		    jvalue *value)
{
  assert(s != NULL);

  if (tag == SIG_JDOUBLE)
    {
      op_stack_push_double(env, s, value->d);
    }
  else if (tag == SIG_JLONG)
    {
      op_stack_push_long(env, s, value->j);
    }
  else
    {
      s->stack_top->tag = tag;
      s->stack_top->value = *value;
      s->stack_top++;
    }
}

void
op_stack_push_byte(JNIEnv *env, OpStack *s,
		   jbyte value)
{
  JavaStackItem newvalue;

  JAVARLOG2(OPSTACK_LOG, 1, "    Pushing byte %d onto operand stack %p\n", value, s);

  assert(s != NULL);

  newvalue.tag = SIG_JBYTE;
  newvalue.value.b = value;

  *(s->stack_top++) = newvalue;
}

void
op_stack_push_char(JNIEnv *env, OpStack *s,
		   jchar value)
{
  JavaStackItem newvalue;

  JAVARLOG3(OPSTACK_LOG, 1, "    Pushing char %d (%c) onto operand stack %p\n", 
	    value, (char)(value & 0xff), s);

  assert(s != NULL);

  newvalue.tag = SIG_JCHAR;
  newvalue.value.c = value;

  *(s->stack_top++) = newvalue;
}

void
op_stack_push_short(JNIEnv *env, OpStack *s,
		    jshort value)
{
  JavaStackItem newvalue;

  JAVARLOG2(OPSTACK_LOG, 1, "    Pushing short %d onto operand stack %p\n", value, s);

  assert(s != NULL);

  newvalue.tag = SIG_JSHORT;
  newvalue.value.i = value;

  *(s->stack_top++) = newvalue;
}

void
op_stack_push_boolean(JNIEnv *env, OpStack *s,
		      jboolean value)
{
  JavaStackItem newvalue;

  JAVARLOG2(OPSTACK_LOG, 1, "    Pushing boolean %s onto operand stack %p\n", 
	    value == 0 ? "False" : "True", s);

  assert(s != NULL);

  newvalue.tag = SIG_JBOOLEAN;
  newvalue.value.z = value;

  *(s->stack_top++) = newvalue;
}

void
op_stack_push_int(JNIEnv *env, OpStack *s,
		  jint value)
{
  JavaStackItem newvalue;

  JAVARLOG2(OPSTACK_LOG, 1, "    Pushing int %ld onto operand stack %p\n", value, s);

  assert(s != NULL);

  newvalue.tag = SIG_JINT;
  newvalue.value.i = value;

  *(s->stack_top++) = newvalue;
}

void
op_stack_push_long(JNIEnv *env, OpStack *s,
		   jlong value)
{
  JavaStackItem newvalue;

  JAVARLOG2(OPSTACK_LOG, 1, "    Pushing long %ld onto operand stack %p\n", value, s);

  assert(s != NULL);

  newvalue.tag = SIG_JLONG1;
  newvalue.value.i = (value >> 32) & 0xffffffff;

  *(s->stack_top++) = newvalue;

  newvalue.tag = SIG_JLONG2;
  newvalue.value.i = value & 0xffffffff;

  *(s->stack_top++) = newvalue;
}

void
op_stack_push_long_w(JNIEnv *env, OpStack *s,
		     jint high_word,
		     jint low_word)
{
  JavaStackItem newvalue;

  JAVARLOG3(OPSTACK_LOG, 1, "    Pushing long %ld:%ld onto operand stack %p\n",
	    high_word, low_word, s);

  assert(s != NULL);

  newvalue.tag = SIG_JLONG1;
  newvalue.value.i = high_word;

  *(s->stack_top++) = newvalue;

  newvalue.tag = SIG_JLONG2;
  newvalue.value.i = low_word;

  *(s->stack_top++) = newvalue;
}

void
op_stack_push_float(JNIEnv *env, OpStack *s,
		    jfloat value)
{
  JavaStackItem newvalue;

  JAVARLOG2(OPSTACK_LOG, 1, "    Pushing float %f onto operand stack %p\n", value, s);

  assert(s != NULL);

  newvalue.tag = SIG_JFLOAT;
  newvalue.value.f = value;

  *(s->stack_top++) = newvalue;
}

void
op_stack_push_double(JNIEnv *env, OpStack *s,
		     jdouble value)
{
  JavaStackItem newvalue;
  jvalue val;

  JAVARLOG2(OPSTACK_LOG, 1, "    Pushing double %lf onto operand stack %p\n", value, s);

  assert(s != NULL);

  val.d = value;

  newvalue.tag = SIG_JDOUBLE1;
  newvalue.value.i = (long)((val.j >> 32) & 0xffffffff);

  *(s->stack_top++) = newvalue;

  newvalue.tag = SIG_JDOUBLE2;
  newvalue.value.i = (long)(val.j & 0xffffffff);

  *(s->stack_top++) = newvalue;
}

void
op_stack_push_double_w(JNIEnv *env, OpStack *s,
		       jint word1,
		       jint word2)
{
  JavaStackItem newvalue;

  JAVARLOG3(OPSTACK_LOG, 1, "    Pushing double %ld:%ld onto operand stack %p\n", word1, word2, s);

  assert(s != NULL);

  newvalue.tag = SIG_JDOUBLE1;
  newvalue.value.i = word1;

  *(s->stack_top++) = newvalue;

  newvalue.tag = SIG_JDOUBLE2;
  newvalue.value.i = word2;

  *(s->stack_top++) = newvalue;
}

void
op_stack_push_object(JNIEnv *env, OpStack *s,
		     jobject o)
{
  JavaStackItem newvalue;

  JAVARLOG2(OPSTACK_LOG, 1, "    Pushing objectref %p onto operand stack %p\n", 
	    o, s);

  assert(s != NULL);

  newvalue.tag = SIG_JOBJECT;
  newvalue.value.l = o;

  *(s->stack_top++) = newvalue;
}


void
op_stack_pop_any(JNIEnv *env, OpStack *s,
		 JavaStackItem *any)
{
  assert (s != NULL);

  *any = *(--s->stack_top);
}

void
op_stack_pop_value(JNIEnv *env, OpStack *s,
		   jvalue *value)
{
  JavaStackItem item;

  assert (s != NULL);

  item = *(s->stack_top - 1);
  
  if (item.tag == SIG_JDOUBLE2)
    {
      op_stack_pop_double(env, s, &(value->d));
    }
  else if (item.tag == SIG_JLONG2)
    {
      op_stack_pop_long(env, s, &(value->j));
    }
  else
    {
      s->stack_top--;
      *value = item.value;
    }
}

void
op_stack_pop_byte(JNIEnv *env, OpStack *s,
		  jbyte *value)
{
  JavaStackItem item;

  assert (s != NULL);

  item = *(--s->stack_top);

  /*  assert(item.tag == SIG_JBYTE);*/

  *value = item.value.b;

  JAVARLOG2(OPSTACK_LOG, 1, "    Popping byte off operand stack %p, value is %d\n", s, *value);
}

void
op_stack_pop_char(JNIEnv *env, OpStack *s,
		  jchar *value)
{
  JavaStackItem item;

  assert (s != NULL);

  item = *(--s->stack_top);

  /*  assert(item.tag == SIG_JCHAR);*/

  *value = item.value.c;

  JAVARLOG3(OPSTACK_LOG, 1, "    Popping char off operand stack %p, value is %d (%c)\n",
	    s,
	    *value, (char)(*value & 0xff));
}

void
op_stack_pop_short(JNIEnv *env, OpStack *s,
		   jshort *value)
{
  JavaStackItem item;

  assert (s != NULL);

  item = *(--s->stack_top);

  /*  assert(item.tag == SIG_JSHORT);*/

  *value = item.value.s;

  JAVARLOG2(OPSTACK_LOG, 1, "    Popping short off operand stack %p, value is %d\n", s, *value);
}

void
op_stack_pop_boolean(JNIEnv *env, OpStack *s,
		     jboolean *value)
{
  JavaStackItem item;

  assert (s != NULL);

  item = *(--s->stack_top);

  /*  assert(item.tag == SIG_JBOOLEAN);*/

  *value = item.value.z;

  JAVARLOG2(OPSTACK_LOG, 1, "    Popping boolean off operand stack %p, value is %s\n", 
	    s, *value == 0 ? "False" : "True");
}

void
op_stack_pop_int(JNIEnv *env, OpStack *s,
		 jint *value)
{
  JavaStackItem item;

  assert (s != NULL);

  item = *(--s->stack_top);

  /*  assert(item.tag == SIG_JINT);*/

  *value = promote_to_int(env, item);

  JAVARLOG2(OPSTACK_LOG, 1, "    Popping int off operand stack %p, value is %ld\n", s, *value);
}

void
op_stack_pop_long(JNIEnv *env, OpStack *s,
		  jlong *value)
{
  JavaStackItem item;
  jvalue val1, val2;

  assert (s != NULL);

  item = *(--s->stack_top);

  /*  assert(item.tag == SIG_JLONG2);*/

  val2.i = item.value.i;

  item = *(--s->stack_top);

  /*  assert(item.tag == SIG_JLONG1);*/

  val1.i = item.value.i;

  *value = (jlong)val1.i << 32;
  *value |= val2.j & 0xffffffff;

  JAVARLOG2(OPSTACK_LOG, 1, "    Popping long off operand stack %p, value is %ld\n", s, *value);
}

void
op_stack_pop_long_w(JNIEnv *env, OpStack *s,
		    jint *word1,
		    jint *word2)
{
  JavaStackItem item;

  assert (s != NULL);

  item = *(--s->stack_top);

  /*  assert(item.tag == SIG_JLONG2);*/

  *word2 = item.value.i;

  item = *(--s->stack_top);

  /*  assert(item.tag == SIG_JLONG1);*/

  *word1 = item.value.i;

  JAVARLOG3(OPSTACK_LOG, 1, "    Popping long off operand stack %p, value is %ld:%ld\n",
	    s, *word1, *word2);;
}

void
op_stack_pop_float(JNIEnv *env, OpStack *s,
		   jfloat *value)
{
  JavaStackItem item;

  assert (s != NULL);

  item = *(--s->stack_top);

  /*  assert(item.tag == SIG_JFLOAT);*/

  *value = item.value.f;

  JAVARLOG2(OPSTACK_LOG, 1, "    Popping float off operand stack %p, value is %f\n", s, *value);
}

void
op_stack_pop_double(JNIEnv *env, OpStack *s,
		    jdouble *value)
{
  JavaStackItem item;
  jvalue val;
  jvalue val1, val2;

  assert (s != NULL);

  val.j = 0;

  item = *(--s->stack_top);

  /*  assert(item.tag == SIG_JDOUBLE2);*/

  val2.i = item.value.i;

  item = *(--s->stack_top);

  /*  assert(item.tag == SIG_JDOUBLE1);*/

  val1.i = item.value.i;

  val.j = (jlong)val1.i << 32;
  val.j |= val2.j & 0xffffffff;

  *value = val.d;

  JAVARLOG2(OPSTACK_LOG, 1, "    Popping double off operand stack %p, value is %lf\n", s, *value);
}

void
op_stack_pop_double_w(JNIEnv *env, OpStack *s,
		      jint *word1,
		      jint *word2)
{
  JavaStackItem item;

  assert (s != NULL);

  item = *(--s->stack_top);

  /*  assert(item.tag == SIG_JDOUBLE2);*/

  *word2 = item.value.i;

  item = *(--s->stack_top);

  /*  assert(item.tag == SIG_JDOUBLE1);*/

  *word1 = item.value.i;

  JAVARLOG3(OPSTACK_LOG, 1, "    Popping double off operand stack %p, value is %f:%f\n",
	    s, *word1, *word2);
}

void
op_stack_pop_object(JNIEnv *env, OpStack *s,
		    jobject *obj)
{
  JavaStackItem item;

  assert (s != NULL);

  item = *(--s->stack_top);

  /*  assert(item.tag == SIG_JOBJECT);*/

  *obj = item.value.l;

  JAVARLOG2(OPSTACK_LOG, 1, "  Popping objectref off operand stack %p, value is %p\n",
	    s, *obj);
}
