/* -*- Mode: C; c-file-style: "gnu" -*-
   resolve.c -- load in classes that another depends on.
   Created: Chris Toshok <toshok@hungry.com>, 12-Aug-1997.
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
#include "config.h"
#include "log.h"
#include "resolve.h"
#include "ClazzFile.h"
#include "class-repository.h"
#include "array-class.h"
#include "jniint.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>

#define MYLOG "Resolve"

jclass
find_class_on_classpath(JNIEnv *env,
			const char *class_name, ClasspathEntry *entries, int num_entries)
{
  char *full_class_name;
  ClazzFile *cf = NULL;
  int i;

  for (i = 0; i < num_entries; i ++)
    {
      if (entries[i].type == CP_ZIP ||
	  entries[i].type == CP_JAR)
	{
	  ZipDirectory *dir;
	  int j;

	  /* if there was an error reading the archive, pass over it. */
	  if (entries[i].zip.fd == -1)
	    continue;

	  full_class_name = (char*)malloc(strlen(class_name) 
					  + strlen(".class") + 2);

	  strcpy(full_class_name, class_name);
	  strcat(full_class_name, ".class");

	  dir = (ZipDirectory*)entries[i].zip.central_directory;

	  for (j = 0; j < entries[i].zip.count; j ++, dir = ZIPDIR_NEXT(dir))
	    {
	      if (!strcmp(full_class_name, ZIPDIR_FILENAME(dir)))
		{
		  char *buf;
		  int length;

		  if (get_zipfile_entry(&entries[i].zip,
					dir,
					&buf, &length) != -1)
		    {
		      cf = define_class(env, buf, length);
		    }

		  free(buf);

		  break;
		}
	    }

	  free(full_class_name);
	}
      else if (entries[i].type == CP_DIR)
	{
	  full_class_name = (char*)malloc(strlen(class_name) 
					  + strlen(entries[i].path)
					  + strlen(".class") + 2);
	  
	  strcpy(full_class_name, entries[i].path);
	  strcat(full_class_name, "/");
	  strcat(full_class_name, class_name);
	  strcat(full_class_name, ".class");
	  
	  /*      sprintf (full_class_name, "%s/%s.class", entries[i].path, class_name);*/
	  
	  cf = parse_class (env, full_class_name);
	  
	  free(full_class_name);
	  
	}

      if (cf != NULL)
	{
	  break;
	}
    }

  if (cf != NULL)
    {
      add_class_to_repository(env, cf, (char*)class_name);
    }

  return cf;
}

jclass
find_class(JNIEnv *env, const char *class_name)
{
  assert(NULL != class_name);

  if (class_name[0] == '[')
    {
      /* for array classes, we just create a fake entry for them. */
      return createFakeArrayClass(env, (char*)class_name);
    }
  else
    {
      ClazzFile *result;
      char *real_class_name;
	  
      /* dots to slashes. */
      if (strchr(class_name, '.'))
	{
	  int i;
		  
	  real_class_name = strdup(class_name);
	  for (i = 0; i < strlen(real_class_name); i ++)
	    if (real_class_name[i] == '.')
	      real_class_name[i] = '/';
	}
      else
	{
	  real_class_name = strdup(class_name);
	}
	  
      result = find_class_in_repository(env, (char*)real_class_name);
	  
      if (!result)
	{
	  HungryJNIEnv *henv = (HungryJNIEnv*)env;
	  HungryJavaVM *hvm = henv->_vm;
	  ClasspathEntry *entries = hvm->_cp_entries;
	  int num_entries = hvm->_num_cp_entries;
		  
	  result = find_class_on_classpath(env, real_class_name, entries, num_entries);
	}

      free (real_class_name);

      return result;
    }
}

jobject
find_descriptor(JNIEnv *env, const char *class_name)
{
  ClazzFile *cf = find_class(env, class_name);

  return find_descriptor_in_repository(env, cf);
}

void
load_parent_classes(JNIEnv *env, jclass clazz)
{
  char *parent_class_name;
  ClazzFile *cur_cf = (ClazzFile*)clazz;

  JAVARLOG1(MYLOG, 1, "Inside load_parent_classes(%s)\n", getClassName(env, cur_cf));

  while(cur_cf)
    {
      if (cur_cf->superclass_index == 0)
	cur_cf = NULL;
      else
	{
	  parent_class_name = getSuperClassName(env, cur_cf);
	  cur_cf = find_class(env, parent_class_name);
	}
    }
}
