/* -*- Mode: C; c-file-style: "gnu" -*-
   japhar.c -- Java interpreter.
   Created: Chris Toshok <toshok@hungry.com>, 26-Jul-1997
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdlib.h>
#include <unistd.h>
#include "getopt.h"
#include "jni.h"
#include "jniint.h"
#include "gc.h"

#ifdef WITH_MOTIF_AWT
/* we need the following to force particular pieces of motif to be
   linked in if it's an archive (.a).  We don't need this if it's a
   shared library, but since at least one person doesn't have a motif
   shared library (me), we can't depend on it.  Besides, this won't
   break in the case of shared libraries... */
#include <Xm/Xm.h>
extern WidgetClass xmDrawingAreaWidgetClass;
extern WidgetClass xmCascadeButtonWidgetClass;
extern WidgetClass xmMainWindowWidgetClass;
extern WidgetClass xmFormWidgetClass;
static WidgetClass *foo = &xmDrawingAreaWidgetClass;
static WidgetClass *bar = &xmCascadeButtonWidgetClass;
static WidgetClass *baz = &xmMainWindowWidgetClass;
static WidgetClass *baz2 = &xmFormWidgetClass;
#endif

static char *clazzname = NULL;
static char *classpath = NULL;

static jclass string_cls;
static jint param_start;
static jvalue cmd_params[1];

static
void
usage (char *name)
{
  printf ("Usage: %s [OPTION]... CLASS [ARG]...\n", name);
  printf ("Execute Java bytecode in CLASS, starting with the"
          " method\n\"static public void main(String[])\"."
          "  Each ARG is passed\ninto \"main\" via the String array.\n\n"
	  "  --classpath=PATH  use PATH to lookup classes\n"
          "  --help            display this help and exit\n"
          "  --version         output version information and exit\n");
}

static
void
parse_command_line (int argc, char **argv)
{
  while (1)
    {
      int c;
      static int version_flag = 0;
      static int help_flag = 0;

      static struct option long_options[] = 
      {
	{ "version", no_argument, &version_flag, 1 },
	{ "help", no_argument, &help_flag, 1 },
	{ "classpath", required_argument, 0, 'p' },

	/* options to be later supported */
	/* { "define", 0, 0, 0 }, */
	/* { "verbose-classload", 0, 0, 0 }, */
	/* { "verbose-gc", 0, 0, 0 }, */
	/* { "verbose-jni", 0, 0, 0 }, */
	/* { "jar", 0, 0, 0 }, */

	{ 0, 0, 0, 0 }
      };

      int option_index = 0;
      
      c = getopt_long (argc, argv, "+D:", /* the `+' sets POSIX semantics */
                       long_options, &option_index);
      
      /* Detect the end of the options. */
      if (c == -1)
        break;
      
      switch (c)
        {
        case 'D':
          fprintf (stderr, "warning: setting system properties is not supported\n");
          /* addUserProperty(&argv[option_index][2]); */
	  break;
	case 'p':
	  classpath = optarg;
        }

      if (version_flag) 
        {
          printf ("GNU Japhar %s\n"
                  "Copyright (C) 1998 The Hungry Programmers\n"
                  "Japhar comes with ABSOLUTELY NO WARRANTY.\n"
                  "You may redistribute copies of Japhar\n"
                  "under the terms of the GNU General Public License.\n"
                  "For more information about these matters, see the "
                  "files named COPYING.\n", VERSION);
          exit (0);
        }

      if (help_flag)
        {
          usage (argv[0]);
          exit (0);
        }
    }

  clazzname = argv[ optind ];
  if (!clazzname)
    { 
      usage (argv[0]);
      exit (0);
    }

  param_start = ++optind;
}

void
create_java_command_line(int argc, char **argv,
                         JNIEnv *env, JavaVM *jvm)
{
  jobjectArray params;
  int argi = param_start;

  string_cls = (*env)->FindClass(env, "java/lang/String");
  if ((*env)->ExceptionOccurred(env))
    return;

  params = (*env)->NewObjectArray(env, argc - argi, string_cls, NULL);
  params = (*env)->NewGlobalRef(env, params);

  while (argi < argc)
    {
      jstring str = (*env)->NewStringUTF(env, argv[ argi ]);
      str = (*env)->NewGlobalRef(env, str);

      (*env)->SetObjectArrayElement(env, params, argi - param_start,
                                    str);
      argi++;
    }
  
  cmd_params[0].l = params;
}

static void
handleException(JNIEnv *env,
                JavaVM *jvm)
{
  (*env)->ExceptionDescribe(env);
 
  (*env)->ExceptionClear(env);

  (*jvm)->DestroyJavaVM(jvm);

  exit(0);
}

int
main(int argc,
     char **argv)
{
  JavaVM *jvm;
  JNIEnv *env;
  JDK1_1InitArgs vm_args;

  jclass cls;
  jmethodID mid;

  /*
  ** Get the default initialization arguments and set the class path.
  */
  vm_args.version = 0x10001; /* JDK 1.1 */
  JNI_GetDefaultJavaVMInitArgs(&vm_args);

#if 0
  /* turn on class loading spew. */
  vm_args.enableVerboseClassLoading = JNI_TRUE;
#endif

  /* parse the command line arguments, extracting the clazzname 
     (and parameters) */
  parse_command_line(argc, argv);

  /*
  ** modify the classpath, if requested from the command line
  */
  if (classpath)
    vm_args.classpath = classpath;

  /*
  ** load and initialize a JavaVM, return a JNI interface pointer in
  ** env.
  */
  JNI_CreateJavaVM(&jvm, &env, &vm_args);

  create_java_command_line(argc, argv, env, jvm);
  if ((*env)->ExceptionOccurred(env))
    handleException(env, jvm);

  /* invoke the main method of the class specified on the command line */
  cls = (*env)->FindClass(env, clazzname);
  if ((*env)->ExceptionOccurred(env))
    handleException(env, jvm);

  mid = (*env)->GetStaticMethodID(env, cls, "main", "([Ljava/lang/String;)V");
  if ((*env)->ExceptionOccurred(env))
    handleException(env, jvm);

  (*env)->CallStaticVoidMethodA(env, cls, mid, cmd_params);
  if ((*env)->ExceptionOccurred(env))
    handleException(env, jvm);

  /* we now enter into this screwed loop, waiting until all the other
     user threads (if any) die. */
  while ((*jvm)->DestroyJavaVM(jvm) == -1)
    {
      sleep(1);
    }

  /* keep 'gcc -Wall -pedantic' happy */
  exit(0);
}
