/* -*- Mode: C; c-file-style: "gnu" -*-
   header.c -- special code for outputting headers for japharh.
   Created: Chris Toshok <toshok@hungry.com>
 */
/*
  This file is part of Japhar, the GNU Virtual Machine for Java Bytecodes.
  Japhar is a project of The Hungry Programmers, GNU, and OryxSoft.

  Copyright (C) 1997, 1998 The Hungry Programmers

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "ClazzFile.h"
#include <resolve.h>
#include <sig.h>
#include <log.h>
#include <japharh.h>

#define MYLOG "Javah"

static char predefs[100][100];
static int num_predefs = 0;

extern char *javah_dest_dir;

char *
type_to_predef(char *c_type)
{
  static char predef[1024];

  /* we only deal with things that have a * at
     the very end of the string -- because pointers
     to primitive types aren't returned, we can be
     reasonable comfortable in the assumption that
     we're dealing with an object reference of some
     sort. */
  if (c_type[strlen(c_type) - 1] == '*'
      && !strncmp(c_type, "struct", strlen("struct")))
    {
      strcpy(predef, c_type);
      
      /* now get rid of the '*' */
      predef[strlen(predef) - 1] = 0;
      
      return predef;
    }
  else
    return NULL;
}

void
add_predef(char *predef)
{
  int foo;
  int found = 0;
  
  for (foo = 0; foo < num_predefs; foo ++)
    if (!strcmp(predefs[foo], predef))
      found = 1;
  
  if (!found)
    strcpy(predefs [ num_predefs ++ ], predef);
}

void
collect_obj_predefs(ClazzFile *cf)
{
  int field_num;
  ClazzFile *superclass = NULL;
  JNIEnv *env = NULL;

  JAVARLOG0(MYLOG, 1, "in collect_obj_predefs()\n");

  num_predefs = 0;

  for (field_num = 0; field_num < cf->num_fields; field_num ++)
    {
      char *predef;

      if (cf->fields[field_num]->access_flags & ACC_STATIC)
	continue; /* we don't have to worry about static fields. */

      predef = type_to_predef(SIG_formatToC(env, cf->fields[field_num]->sig));

      if (predef != NULL)
	add_predef(predef);
    }

  superclass = cf->superclass;

  if (superclass
      && superclass->access_flags && ACC_ABSTRACT)
    {
      for (field_num = 0; field_num < superclass->num_fields; field_num ++)
	{
	  char *predef;
	  
	  if (superclass->fields[field_num]->access_flags & ACC_STATIC)
	    continue; /* we don't have to worry about static fields. */
	  
	  predef = type_to_predef(SIG_formatToC(env, superclass->fields[field_num]->sig));
	  
	  if (predef != NULL)
	    add_predef(predef);
	}
    }
}

void
output_predefs(FILE *fp)
{
  int i;

  JAVARLOG0(MYLOG, 1, "in output_predefs()\n");

  for (i = 0; i < num_predefs; i ++)
    {
      fprintf(fp, "%s;\n", predefs[i]);
    }
}

char *
format_constant_value(ClazzFile *cf, int index)
{
  static char working_str[1024];

  JAVARLOG0(MYLOG, 1, "in format_constant_value()\n");

  switch(cf->constants[index].generic.tag)
    {
    case CONSTANT_Integer:
      sprintf(working_str, "%ldL",
	      *(long*)&(cf->constants[index].integer_info.bytes));
      break;
    case CONSTANT_Double:
      sprintf(working_str, "%f",
	      *(double*)&(cf->constants[index].double_info.high_bytes));
      break;
    case CONSTANT_Float:
      sprintf(working_str, "%f",
	      *(float*)&(cf->constants[index].float_info.bytes));
      break;
    case CONSTANT_Long:
      sprintf(working_str, "(int64_t)0x%08lX%08lX",
	      *(long*)&(cf->constants[index].long_info.high_bytes),
	      *(long*)&(cf->constants[index].long_info.low_bytes));
      break;
    default:
      fprintf(stderr, "Go yell at toshok.  japharh doesn't work\n");
      exit(1);
      break;
    }

  return working_str;
}

char *
format_constant(ClazzFile *cf, FieldStruct *field)
{
  JAVARLOG0(MYLOG, 1, "in format_constant()\n");

  if (field->has_constant_value)
    return format_constant_value(cf, field->constant_value_index);
  else
    return NULL;
}

void
output_fields(ClazzFile *cf, FILE *fp, char *bar_name)
{
  int i;
  int num_outputted_fields = 0; /* so we can output a pad C field if 
				   there aren't any non-static java 
				   fields. */
  ClazzFile *superclass = NULL;
  ClazzFile *cur_class;
  JNIEnv *env = NULL;

  JAVARLOG0(MYLOG, 1, "in output_fields()\n");

  for (i = 0; i < cf->num_fields; i ++)
    {
      JAVARLOG1(MYLOG, 1, "Working on field '%s'", 
		cf->fields[i]->name);
	  
      if (cf->fields[i]->access_flags & ACC_FINAL)
	{
	  char *constant = format_constant(cf, cf->fields[i]);
	  if (constant)
	    fprintf (fp, "#define %s_%s %s\n", 
		     bar_name,
		     cf->fields[i]->name,
		     constant);
	  else
	    fprintf (fp, "/* Inaccessible static: %s */\n", 
		     cf->fields[i]->name);
	}
      else if (cf->fields[i]->access_flags & ACC_STATIC)
	{
	  fprintf (fp, "/* Inaccessible static: %s */\n", 
		   cf->fields[i]->name);
	}
      else
	{
	  fprintf (fp, "   %s %s;\n", 
		   SIG_formatToC(env, cf->fields[i]->sig),
		   cf->fields[i]->name);
		  
	  num_outputted_fields ++;
	}
    }
  
  
  /* if the superclass is abstract, we also need to collect the
     predefs for it's fields. */
  cur_class = cf;

  while (cur_class->superclass != 0)
    {
      superclass = cur_class->superclass;
	  
      if (superclass 
	  && superclass->access_flags && ACC_ABSTRACT)
	{
	  for (i = 0; i < superclass->num_fields; i ++)
	    {
	      JAVARLOG1(MYLOG, 1, "Working on field '%s'", 
			superclass->fields[i]->name);
			  
	      if (superclass->fields[i]->access_flags & ACC_FINAL)
		{
		  char *constant = format_constant(superclass, 
						   superclass->fields[i]);
		  if (constant)
		    fprintf (fp, "#define %s_%s %s\n", 
			     bar_name,
			     superclass->fields[i]->name,
			     constant);
		  else
		    fprintf (fp, "/* Inaccessible static: %s */\n", 
			     superclass->fields[i]->name);
		}
	      else if (superclass->fields[i]->access_flags & ACC_STATIC)
		{
		  fprintf (fp, "/* Inaccessible static: %s */\n", 
			   superclass->fields[i]->name);
		}
	      else
		{
		  fprintf (fp, "   %s %s;\n", 
			   SIG_formatToC(env, superclass->fields[i]->sig),
			   superclass->fields[i]->name);
				  
		  num_outputted_fields ++;
		}
	    }
	}
	  
      cur_class = superclass;
    }

  if (num_outputted_fields == 0)
    fprintf (fp, "    char PAD;	/* ANSI C requires structures to have a least one member */");
}

void
collect_method_predefs(Signature *sig,
		       FILE *fp)
{
  int i;
  JNIEnv *env = NULL;

  JAVARLOG0(MYLOG, 1, "in collect_method_predefs()\n");
  
  num_predefs = 0;
  
  for (i = 0; i < sig->method.num_params; i ++)
    {
      char *predef = type_to_predef(SIG_formatToC(env, sig->method.params[i]));
	  
      if (predef)
	add_predef(predef);
    }
}

void
output_method_prototypes(ClazzFile *cf,
			 FILE *fp)
{
  Signature *sig;
  char *class_name = slashes_to_underscores(strdup(cf->class_name));
  int i;
  JNIEnv *env = NULL;

  JAVARLOG0(MYLOG, 1, "in output_method_prototypes()\n");

  fprintf (fp, "#ifdef __cplusplus\n");
  fprintf (fp, "extern \"C\" {\n");
  fprintf (fp, "#endif\n\n");

  for (i = 0; i < cf->num_methods; i ++)
    {
      if (cf->methods[i]->access_flags & ACC_NATIVE)
	{
	  sig = cf->methods[i]->sig;
		  
	  /* first we output the predefs it depends on. */
	  collect_method_predefs(sig, fp);
	  output_predefs(fp);
		  
	  /* then we output the actual prototype. */
	  fprintf (fp, "extern %s Java_%s_%s(JNIEnv *env, jobject obj", 
		   SIG_formatToC(env, sig->method.return_type),
		   class_name, cf->methods[i]->name);
		  
	  if (SIG_numParams(env, sig) > 0)
	    {
	      int j;
			  
	      fprintf (fp, ", ");
	      for (j = 0; j < SIG_numParams(env, sig); j ++)
		{
		  fprintf (fp, "%s%s", SIG_formatToC(env, sig->method.params[j]),
			   j < SIG_numParams(env, sig) - 1 ? ", " : "");
		}
	    }
		  
	  fprintf (fp, ");\n");
	}
    }

  fprintf (fp, "\n#ifdef __cplusplus\n");
  fprintf (fp, "};\n");
  fprintf (fp, "#endif\n\n");

  free(class_name);
}


void
output_header(ClazzFile *cf, char *bar_name)
{
  char full_name[100];
  FILE *fp;

  JAVARLOG0(MYLOG, 1, "in output_header()\n");

  strcpy(full_name, javah_dest_dir);
  if (javah_dest_dir[strlen(javah_dest_dir) - 1] != '/')
    strcat(full_name, "/");
  strcat(full_name, bar_name);
  strcat(full_name, ".h");

  collect_obj_predefs(cf);

  fp = fopen(full_name, "w");

  if (NULL == fp)
    {
      fprintf(stderr, "Unable to open %s for writing\n", full_name);
      return;
    }

  fprintf (fp, "/* DO NOT EDIT THIS FILE - it is machine generated */\n");
  fprintf (fp, "#include \"jni.h\"\n");
  fprintf (fp, "/* Header for class %s */\n\n", bar_name);

  fprintf (fp, "#ifndef _Included_%s\n", bar_name);
  fprintf (fp, "#define _Included_%s\n", bar_name);

  output_predefs(fp);

  output_method_prototypes(cf, fp);

  fprintf (fp, "#endif /* _Included_%s */\n", bar_name);
  fclose(fp);
}
